﻿#ifndef CINEMAINTERFACEH
#define CINEMAINTERFACEH

#include <Unknwn.h>
#include "Cinema.Types.h"

  /// <summary> описание входного/выходного канала плагина, иными словами именованный канал (необходимо для дизайн-тайм систем) </summary>

struct TNamedConnection {
    wchar_t* Name;
    int Number;
};

typedef TNamedConnection TNamedOutputConnection;
typedef TNamedOutputConnection* PNamedOutputConnection;

typedef TNamedConnection TNamedInputConnection;
typedef TNamedInputConnection* PNamedInputConnection;

/// <summary> краткая информация о плагине </summary>

struct TPluginDescription {
  // имя 
  wchar_t* Name;
  // автор
  wchar_t* Autor;
  // дата создания (dd.MM.yyyy)
  wchar_t* Created;
  // версия
  wchar_t* Version;
  // краткое описание
  wchar_t* Description;
  // автор последних модификаций
  wchar_t* AutorLastModified;
  // количество именованных входных соединений; 0 - кол-во не определено, может быть сколь угодно
  int CountNamedInputConnections;
  // именованные входные соединения плагина
  wchar_t* NamedInputConnection;
  // количество именованных выходных соединений; -1 - кол-во не определено, может быть сколь угодно
  int CountNamedConnections;
  // именованные выходные соединения плагина
  wchar_t* NamedOutputConnection;
  // схема конфигурации; описывается если модуль может иметь простые настройки;
  // если nil - либо плагин не настраиваемый либо разработчик предусмотрел
  // собственный конфигуратор в виде "exe".
  // TODO: спецификации для полей ввода: текст, число, логическое, и т.д.
  wchar_t* SchemeConfig;
};

typedef TPluginDescription* PPluginDescription;

/// <summary> поток данных; тип передаваемых данных может быть любым </summary>
typedef void* TDataStream;

/// <summary> атрибут канала; кол-во не ограничено, складываются в односвязный список (через Next) </summary>

struct TDataStreamAttribute {
  // значение атрибута; тип данных может быть любым, зависит от его ID
  wchar_t* Attribute;
  // размер данных в Attribute
  int Size;
  // идентификатор атрибута; все типы будут описаны в справочнике атрибутов
  int ID;
  // следующий атрибут
  TDataStreamAttribute* Next;
  // принадлежность атрибута
  TDataStream DataStreamOwner;
};
  
typedef TDataStreamAttribute* PDataStreamAttribute;

/// <summary> Таблица вывода статистики работы плагина </summary>
MIDL_INTERFACE("833B28D0-7D2F-4041-BA33-CCCABFF7901A")
IPluginTable: IUnknown {
public:
  virtual wchar_t* STDMETHODCALLTYPE GetName() = 0;
  virtual wchar_t* STDMETHODCALLTYPE GetCaption() = 0;
  virtual int STDMETHODCALLTYPE AddColumn(wchar_t * AName, wchar_t * ACaption, TCinemaFieldType AType) = 0;
  virtual int STDMETHODCALLTYPE AddRow() = 0;
  virtual int STDMETHODCALLTYPE GetCountColumn() = 0;
  virtual int STDMETHODCALLTYPE GetCountRows() = 0;
  virtual void STDMETHODCALLTYPE SetCell(int AColumn, int ARow, wchar_t * AValue) = 0;
  virtual wchar_t* STDMETHODCALLTYPE GetCell(int AColumn, int ARow) = 0;
  virtual wchar_t* STDMETHODCALLTYPE GetText() = 0;
  virtual void STDMETHODCALLTYPE SetText(wchar_t * AText) = 0;
  virtual wchar_t* STDMETHODCALLTYPE GetColumnCaption(int AIndex) = 0;
  /// <summary> если плагин выставит в true тогда ядро вызовет RedrawView у плагина при первой возможности где плагин
  /// может обновить свои графические данные, в частности и таблицы;
  /// использовать по мере острой необходимости, т.к. при запросе таблицы от GUI эта опция выставляется в автомате <summary>
  virtual void STDMETHODCALLTYPE SetNeedReload(bool AValue) = 0;
};

MIDL_INTERFACE("C081EBF9-0F68-43AF-8925-ACEAC0B1D9B3")
IValueHolder: IUnknown{
public:
  virtual byte * STDMETHODCALLTYPE GetValue() = 0;
  virtual int STDMETHODCALLTYPE GetSize() = 0;
};

/// <summary> интерфейс ядра; используется плагином для отправки результата выполнения работы </summary>

MIDL_INTERFACE("4F6E8DD5-FB09-4831-9188-FEB505A3A98D")
ICinema: IUnknown {
  ///  <summary> метод для журналирования событий от плагина </summary>
  virtual void STDMETHODCALLTYPE SendMessage(wchar_t* AMsg, TMessageType AMessageType) = 0;
  ///  <summary> Установка текущего состояния качества выполняемой работы плагином </summary>
  ///  TODO: логирование изменений статуса
  virtual void STDMETHODCALLTYPE SetHealthState(wchar_t* AMsg, byte AHealthDegree) = 0;
  ///  <summary> создать таблицу для вывода статистики работы плагина;
  ///    <return> возвращает эксземпляр с интерфейсом IPluginTable; тип определён как Pointer
  ///    для кросскомпиляторности (Delphi и C++), поэтому необходимо самостоятельное приведение к IPluginTable </return>
  ///  </summary>
  virtual void* STDMETHODCALLTYPE CreateTable(wchar_t* AName, wchar_t* ACaption) = 0;
  ///  <summary> метод создания потока для отправки данных плагином </summary>
  ///  <param Name="AConnectionNumber"> Номер соединения на котором следующий плагин в сценарии
  ///  будет получать данные потока </param>
  virtual TDataStream STDMETHODCALLTYPE CreateStream(int AConnectionNumber, TDataStream AParentDataStream) = 0;
  ///  <summary> отправить данные по ранее созданному потоку </summary>
  virtual void STDMETHODCALLTYPE SendData(TDataStream ADataStream, byte* AData, __int64 ASize) = 0;
  ///  <summary> метод уничтожения потока </summary>
  virtual void STDMETHODCALLTYPE DestroyStream(TDataStream ADataStream) = 0;
  ///  <summary> добавить атрибут потока </summary>
  virtual PDataStreamAttribute STDMETHODCALLTYPE AddDataStreamAttribute(TDataStream ADataStream, int ID, void* AData, int ADataSize) = 0;
    ///  <summary> найти атрибут по идентификатору </summary>
  virtual PDataStreamAttribute STDMETHODCALLTYPE FindAttribute(TDataStream ADataStream, int ID) = 0;
    ///  <summary> найти атрибут по идентификатору </summary>
  virtual void* STDMETHODCALLTYPE FindAttributeValuePtr(TDataStream ADataStream, int ID, void* ADefaultValuePtr) = 0;
    ///  <summary> найти атрибут типа NativeInt по идентификатору </summary>
  virtual NativeInt STDMETHODCALLTYPE FindAttributeNativeInt(TDataStream ADataStream, int ID, NativeInt ADefault) = 0;
    ///  <summary> упаковка всех атрибутов потока в буфер </summary>
  virtual byte* STDMETHODCALLTYPE PackAttributes(TDataStream ADataStreamSource, int* ALength) = 0;
  ///  <summary> распаковка всех атрибутов из буфера в поток </summary>
  virtual void STDMETHODCALLTYPE UnPackAttributes(TDataStream ADataStreamDest, byte* AAttributes, int ALength) = 0;
  ///  <summary> получить ссылку на все атрибуты </summary>
  virtual PDataStreamAttribute STDMETHODCALLTYPE GetAttributes(TDataStream ADataStream) = 0;
  ///  <summary> выставить атрибуты из списка </summary>
  virtual void STDMETHODCALLTYPE SetAttributes(TDataStream ADataStream, PDataStreamAttribute AAtributes) = 0;
  ///  <summary> если плагин может выполнять работу без длительного возврата захваченного потока слоя,
  ///  тогда разработчик по возможности должен предусмотреть прерывание текущей работы плагина для того
  ///  что бы при желании пользователя остановить сценарий, можно было корректно завершить его выполнение;
  ///  данный метод позволяет узнать плагину что поток был прерван и необходимо завершать работу </summary>
  virtual bool STDMETHODCALLTYPE IsThreadTerminated() = 0;
  ///  <summary> имя и путь выполняемого сценария </summary>
  virtual wchar_t* STDMETHODCALLTYPE GetScenarioName() = 0;
  ///  <summary> поучить путь до модуля плагина </summary>
  virtual wchar_t* STDMETHODCALLTYPE GetPluginPath() = 0;
  ///  <summary> поучить путь до приложения в котором выполняется сценарий </summary>
  virtual wchar_t* STDMETHODCALLTYPE GetAppPath() = 0;
  ///  <summary> TODO </summary>
  virtual void STDMETHODCALLTYPE SetForceAsStart(bool AValue)  = 0;
  /// <summary> Получить полный путь до каталога со сценариями </summary>
  virtual wchar_t* STDMETHODCALLTYPE GetRtlScenariosPath() = 0;
  /// <summary> Добавить глобальный атрибут, заменяет на новое значение если такой уже есть </summary>
  virtual void STDMETHODCALLTYPE AddGlobalAttribute(int AID, void* AData, int ADataSize) = 0;
  virtual IValueHolder* STDMETHODCALLTYPE FindGloabalAttribute(int AID) = 0;
  virtual void STDMETHODCALLTYPE RemoveGlobalAttribute(int AID) = 0;
  virtual void STDMETHODCALLTYPE AddGlobalNamedAttribute(wchar_t* AName, void* AData, int ADataSize) = 0;
  virtual IValueHolder* STDMETHODCALLTYPE FindGloabalNamedAttribute(wchar_t* AName) = 0;
  virtual void STDMETHODCALLTYPE RemoveGlobalNamedAttribute(wchar_t* AName) = 0;
  virtual void STDMETHODCALLTYPE ClearGlobalAttributes() = 0;
  virtual wchar_t* STDMETHODCALLTYPE GetVersionScenario() = 0;
};

  /// <summary> интерфейс плагина, реализуется модулем; ядро может создать несколько экземпляров,
  ///  в зависимости от необходимости при выполнении сценария </summary>

MIDL_INTERFACE("54475BE8-128D-4161-B58B-4AC32E9617FC")
ICinemaPlugin: IUnknown {
  /// <summary> создать обработчик данных передаваемых во входном потоке ADataStream;
  /// возвращаемый результат может быть любым, на усмотрение разработчика плагина;
  /// как правило, это класс который будет выполнять обработку
  /// входных данных получаемых в методе ProcessDataStream </summary>
  virtual void* STDMETHODCALLTYPE CreateDataStreamProcessor(TDataStream AInputDataStream, int ANumberInput) = 0;
  virtual void STDMETHODCALLTYPE ProcessDataStream(void* AProcessor, byte* AData, __int64 ASize) = 0;
  virtual void STDMETHODCALLTYPE DestroyDataStreamProcessor(void* AProcessor) = 0;
  /// <summary> обновить графику плагина (таблицы и прочее что в будущем ещё может понадобится вывести) </summary>
  virtual void STDMETHODCALLTYPE RedrawView() = 0;
};

#endif
