﻿unit Cinema.Hypervisor.Interfaces;

interface

uses
  System.JSON,
  Cinema.Types,
  Cinema.Log.Interfaces,
  Cinema.Scenario.Interfaces
  ;

type
  IHypervisorCinemaNode = interface;
  IHypervisorScenario = interface;
  IHypervisorNode = interface;

  TNodeStateMessages = array[TNodeState] of WideString;
  THypervisorNodeEvent = procedure(ASource: IHypervisorNode) of object;

  TSumResources = record
    TotalMemory: int64;
    UsageMemory: int64;
    AvailableMemory: int64;
    CPUUsage: byte;
  end;


  IHypervisorNode = interface
  ['{14D11C8B-3C87-4967-950F-C4D2C303F82A}']
    function GetObject: TInterfacedObject;
    function GetKey: WideString;
    function GetName: WideString;
    procedure SetName(const AValue: WideString);
    function GetParent: IHypervisorNode;
    procedure SetParent(const AValue: IHypervisorNode);
    function GetChild(Index: int32): IHypervisorNode;
    function GetFolded: boolean;
    procedure SetFolded(const Value: boolean);
    function GetCountChildren: int32;

    procedure SetWorkState(AState: TNodeState);
    procedure SetHealthState(const AMessage: WideString; AState: TNodeHalthDegreeState);
    procedure SetLogState(const AMessage: WideString; ALogState: TNodeLogState);

    function GetCinemaNode: IHypervisorCinemaNode;
    function GetScenarioNode: IHypervisorScenario;

    function GetSumWorkState: TNodeStates;
    function GetSumHealthState: TNodeHalthDegreeState;
    function GetSumHealthStateMsg: WideString;

    function ToJson: WideString;
    procedure FromJson(AJson: TJsonCinema);
    function AddNode(const AName: WideString): IHypervisorNode;
    function AddCinemaNode(const AName: WideString; const AAddress: AnsiString): IHypervisorCinemaNode;
    function AddScenario(const AFileName: UnicodeString): IHypervisorScenario;
    procedure DeleteChild(const AChild: IHypervisorNode);
    function ChildExists(const AName: WideString; const ANodeType: TGUID): boolean; overload;
    function ChildExists(const AName: WideString; const ANodeType: TGUID; out ANode: IHypervisorNode): boolean; overload;

    function CanAddNode: boolean;
    function CanAddCinemaNode: boolean;
    function CanAddScenario: boolean;
    function CanStart: boolean;
    function CanStop: boolean;

    procedure SendRequestRun(const AKey, AName: WideString; ACountInstances: int32);
    procedure SendRequestStop(const AKey: WideString);
    procedure SendRequestTerminate(const AKey: WideString);
    procedure SendRequestKill(const AKey: WideString);
    procedure SendRequestDeleteScenario(const AKey: WideString);

    function GetTotalMemory: int64;
    procedure SetTotalMemory(const Value: int64);
    function GetUsageMemory: int64;
    procedure SetUsageMemory(const Value: int64);
    function GetAvailableMemory: int64;
    procedure SetAvailableMemory(const Value: int64);
    function GetCPUUsage: byte;
    procedure SetCPUUsage(const Value: byte);
    function GetSumResources: TSumResources;

    function GetView: TObject;
    procedure SetView(AValue: TObject);

    property CinemaNode: IHypervisorCinemaNode read GetCinemaNode;
    property ScenarioNode: IHypervisorScenario read GetScenarioNode;

    property Name: WideString read GetName write SetName;
    property Key: WideString read GetKey;
    property Parent: IHypervisorNode read GetParent write SetParent;
    property CountChildren: int32 read GetCountChildren;
    property Child[Index: int32]: IHypervisorNode read GetChild;
    property Folded: boolean read GetFolded write SetFolded;

    property SumWorkState: TNodeStates read GetSumWorkState;
    property SumHealthState: TNodeHalthDegreeState read GetSumHealthState;
    property SumHealthStateMsg: WideString read GetSumHealthStateMsg;
    property TotalMemory: int64 read GetTotalMemory write SetTotalMemory;
    property UsageMemory: int64 read GetUsageMemory write SetUsageMemory;
    property AvailableMemory: int64 read GetAvailableMemory write SetAvailableMemory;
    property CPUUsage: byte read GetCPUUsage write SetCPUUsage;
    property SumResources: TSumResources read GetSumResources;

    property View: TObject read GetView write SetView;
  end;

  IHypervisorGroup = interface(IHypervisorNode)
  ['{8B925769-653A-4DF2-9B86-377411B11C89}']

  end;

  IHypervisorCinemaNode = interface(IHypervisorNode)
  ['{C8F8EA66-F2B7-4A89-9E2C-E21D6BE8DD37}']
    procedure SendUpdatedScenario(const AKey, AName: WideString; ACountInstances: int32);
    procedure SendRequestStatTables(const AKey: WideString; const AGuid: TGuid);
    procedure UpdateItself;
    procedure SendFile(const AFileName: string);
    function GetIsConnected: boolean;
    function GetAddress: string;
    procedure SetAddress(const Value: string);
    procedure SendRequestStat;
    function GetVersion: WideString;
    procedure SetVersion(const Value: WideString);
    function GetOS: WideString;

    property IsConnected: boolean read GetIsConnected;
    property Address: string read GetAddress write SetAddress;
    property Version: WideString read GetVersion write SetVersion;
    property OS: WideString read GetOS;
  end;

  IHypervisorScenario = interface(IHypervisorNode)
  ['{41C1134C-B16D-4613-B084-7C06F4B0C7AE}']
    function GetFileScenario: UnicodeString;
    procedure SetFileScenario(const AValue: UnicodeString);
    function GetScenario: ICinemaScenario;
    //function GetScenarioPrototype: ICinemaScenario;
    function GetCountInstances: int32;
    procedure SetCountInstances(const AValue: int32);
    procedure SendRequestStatTables(const AGuid: TGuid);
    procedure ResetToState(AState: TNodeState);
    procedure SendUpdatedScenario(ACountInstances: int32);

    property FileScenario: UnicodeString read GetFileScenario write SetFileScenario;
    property Scenario: ICinemaScenario read GetScenario;
    //property ScenarioPrototype: ICinemaScenario read GetScenarioPrototype;
    property CountInstances: int32 read GetCountInstances write SetCountInstances;
  end;

  THypervisorEventsListener = class
  public
    procedure OnChangeNodeState(const ASourceNode: IHypervisorNode); virtual; stdcall; abstract;
    procedure OnChangeNodeHealthState(const ASourceNode: IHypervisorNode); virtual; stdcall; abstract;
    procedure OnChangeNodeLog(const ASourceNode: IHypervisorNode); virtual; stdcall; abstract;
    procedure OnChangeScenarioProcessorStat(const ASourceNode: IDataStreamProcessor); virtual; stdcall; abstract;
    procedure OnChangeScenarioStat(const ASourceNode: IHypervisorScenario); virtual; stdcall; abstract;
    procedure OnChangeCinemaNodeStat(const ASourceCineamNode: IHypervisorCinemaNode); virtual; stdcall; abstract;
    procedure OnReloadScenario(const ASourceNode: IHypervisorNode); virtual; stdcall; abstract;
  end;

  IHypervisor = interface(IHypervisorNode)
  ['{E3AB4693-6882-479E-A203-29BF5759A06A}']
    function FindLikeNode(const ALikeName: WideString; const AFromNode: IHypervisorNode): IHypervisorNode;
    procedure Start;
    procedure Stop;
    procedure SendEvent(const AKey, ASource, AWhoIs, AMsg: WideString; ADateTime: TDateTime; AEventType: int32; ASourceId: TGuid);
    procedure SendHealthStateNodeEvent(const AKey: WideString; AState: byte; AHealthState: byte; const AHealthStateMsg: WideString; const AItemID: TGuid);
    procedure ChangeStateEvent(const AKey: WideString; const ASource: TGuid; AState: TNodeState);

    function GetLog: ICinemaLog;
    function GetIsRun: boolean;
    function GetEventsListener: THypervisorEventsListener;
    procedure SetEventsListener(const Value: THypervisorEventsListener);

    function GetOnCreateNode: THypervisorNodeEvent;
    function GetOnDestroyNode: THypervisorNodeEvent;
    procedure SetOnCreateNode(const Value: THypervisorNodeEvent);
    procedure SetOnDestroyNode(const Value: THypervisorNodeEvent);
    function GetOnRemouteDeleteNode: THypervisorNodeEvent;
    procedure SetOnRemouteDeleteNode(const Value: THypervisorNodeEvent);
    function GetOnConnectCinemaNodeEvent: THypervisorNodeEvent;
    procedure SetOnConnectCinemaNodeEvent(AValue: THypervisorNodeEvent);
    function GetOnDisconnectCinemaNodeEvent: THypervisorNodeEvent;
    procedure SetOnDisconnectCinemaNodeEvent(AValue: THypervisorNodeEvent);
    function GetCountCinemaNodes: int32;
    function GetCountConnectedCinemaNodes: int32;
    function GetSelectedCinemaNode: IHypervisorCinemaNode;
    procedure SetSelectedCinemaNode(const Value: IHypervisorCinemaNode);
    function GetIsDeserializingNow: Boolean;

    property IsRun: boolean read GetIsRun;
    property Log: ICinemaLog read GetLog;
    property EventListener: THypervisorEventsListener read GetEventsListener write SetEventsListener;
    property CountCinemaNodes: int32 read GetCountCinemaNodes;
    property CountConnectedCinemaNodes: int32 read GetCountConnectedCinemaNodes;
    property SelectedCinemaNode: IHypervisorCinemaNode read GetSelectedCinemaNode write SetSelectedCinemaNode;
    property IsDeserializingNow: Boolean read GetIsDeserializingNow;

    property OnCreateNode: THypervisorNodeEvent read GetOnCreateNode write SetOnCreateNode;
    property OnDestroyNode: THypervisorNodeEvent read GetOnDestroyNode write SetOnDestroyNode;
    property OnRemouteDeleteNode: THypervisorNodeEvent read GetOnRemouteDeleteNode write SetOnRemouteDeleteNode;
    property OnConnectCinemaNodeEvent: THypervisorNodeEvent read GetOnConnectCinemaNodeEvent write SetOnConnectCinemaNodeEvent;
    property OnDisconnectCinemaNodeEvent: THypervisorNodeEvent read GetOnDisconnectCinemaNodeEvent write SetOnDisconnectCinemaNodeEvent;
  end;

const
  csHypervisorNoError = 0;
  csHypervisorErrorNodeAlreadyExist = 1;

implementation

end.
