﻿unit Cinema.Hypervisor;

{$ifdef FPC}
  {$mode Delphi}
{$endif}

interface

uses
  {$ifdef FPC}
    Classes
  , fpjson
  , jsonparser
  , SyncObjs
  , Generics.Collections
  {$else}
  , System.Classes
  , System.JSON
  , System.SyncObjs
  , System.Generics.Collections
  {$endif}
  , Cinema.Types
  , Cinema.Log.Interfaces
  , Cinema.Net.SocketCommon
  , Cinema.Net.SocketClient
  , Cinema.Net.SocketServer
  , Cinema.Net.Protocol
  , Cinema.Scenario.Interfaces
  , Cinema.Hypervisor.Interfaces
  ;

type

  THypervisor = class;
  THypervisorNodeClass = class of THypervisorNode;

  { THypervisorNode }

  THypervisorNode = class(TInterfacedObject, IHypervisorNode)
  private
    class var
      FClassNodes: TDictionary<string, THypervisorNodeClass>;

    class constructor Create;
    class destructor Destroy;
  private
    FOwner: THypervisor;
    FParent: IHypervisorNode;
    FCinemaNode: IHypervisorCinemaNode;
    FScenarioNode: IHypervisorScenario;

    FChildren: TList<IHypervisorNode>;
    FName: WideString;
    FKey: WideString;

    FNodeState: TNodeState;
    FNodeHealthState: TNodeHalthDegreeState;
    FNodeHealthStateMsg: WideString;
    FNodeLogState: TNodeLogState;
    FNodeLogStateMsg: WideString;

    FFolded: boolean;

    FSumWorkState: TNodeStates;
    FSumHealthState: TNodeHalthDegreeState;
    FSumHealthStateMsg: WideString;

    FTotalMemory: int64;
    FUsageMemory: int64;
    FAvailableMemory: int64;
    FCPUUsage: byte;

    FView: TObject;

    procedure FindParentsTypes;

    procedure CalcSumState;
    procedure ResetState;
  protected
    procedure AddChild(const AChild: IHypervisorNode); virtual;
    procedure RemoveChild(const AChild: IHypervisorNode); virtual;
    procedure DoOnChangeChildState(ASource: THypervisorNode); virtual;
    procedure DoOnChangeChildLogState(ASource: THypervisorNode); virtual;
    procedure DoOnChangeChildHealthState(ASource: THypervisorNode); virtual;
    procedure ProcessError(const ActionName: WideString; AErrorCode: int32);

    property Owner: THypervisor read FOwner;
  public
    constructor Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode); virtual;
    destructor Destroy; override;

    class procedure RegisterClassNode(AClassNode: THypervisorNodeClass);
    class function CreateNode(const AClassName: string; AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode): THypervisorNode;

    function GetObject: TInterfacedObject;
    function GetKey: WideString; virtual;
    function GetName: WideString;
    procedure SetName(const AValue: WideString);
    function GetParent: IHypervisorNode;
    procedure SetParent(const AValue: IHypervisorNode);
    function GetChild(Index: int32): IHypervisorNode;
    function GetFolded: boolean;
    procedure SetFolded(const Value: boolean);
    function GetCountChildren: int32;
    function GetNodeLogState: TNodeLogState;
    function GetNodeState: TNodeState;

    procedure SetWorkState(AState: TNodeState);
    procedure SetHealthState(const AMessage: WideString; AState: TNodeHalthDegreeState);
    procedure SetLogState(const AMessage: WideString; ALogState: TNodeLogState);

    function GetCinemaNode: IHypervisorCinemaNode;
    function GetScenarioNode: IHypervisorScenario;

    function GetSumWorkState: TNodeStates;
    function GetSumHealthState: TNodeHalthDegreeState;
    function GetSumHealthStateMsg: WideString;

    function ToJson: WideString; virtual;
    procedure FromJson(AJson: TJSONCinema); virtual;

    function AddNode(const AName: WideString): IHypervisorNode;
    function AddCinemaNode(const AName: WideString; const AAddress: AnsiString): IHypervisorCinemaNode;
    function AddScenario(const AFileName: UnicodeString): IHypervisorScenario;
    procedure DeleteChild(const AChild: IHypervisorNode);
    function ChildExists(const AName: WideString; const ANodeType: TGUID): boolean; overload;
    function ChildExists(const AName: WideString; const ANodeType: TGUID; out ANode: IHypervisorNode): boolean; overload;

    function CanAddNode: boolean;
    function CanAddCinemaNode: boolean;
    function CanAddScenario: boolean;
    function CanStart: boolean;
    function CanStop: boolean;

    procedure SendRequestRun(const AKey, AName: WideString; ACountInstances: int32); virtual;
    procedure SendRequestStop(const AKey: WideString); virtual;
    procedure SendRequestTerminate(const AKey: WideString); virtual;
    procedure SendRequestKill(const AKey: WideString); virtual;
    procedure SendRequestDeleteScenario(const AKey: WideString); virtual;

    function GetTotalMemory: int64;
    procedure SetTotalMemory(const Value: int64);
    function GetUsageMemory: int64;
    procedure SetUsageMemory(const Value: int64);
    function GetAvailableMemory: int64;
    procedure SetAvailableMemory(const Value: int64);
    function GetCPUUsage: byte;
    procedure SetCPUUsage(const Value: byte);
    function GetSumResources: TSumResources;

    property Name: WideString read GetName write SetName;
    property Key: WideString read GetKey;
    property Parent: IHypervisorNode read GetParent write SetParent;
    property CountChildren: int32 read GetCountChildren;
    property Child[Index: int32]: IHypervisorNode read GetChild;
    property Folded: boolean read GetFolded write SetFolded;

    function GetView: TObject;
    procedure SetView(AValue: TObject);

    property CinemaNode: IHypervisorCinemaNode read GetCinemaNode;
    property ScenarioNode: IHypervisorScenario read GetScenarioNode;

    property SumWorkState: TNodeStates read GetSumWorkState;
    property SumHealthState: TNodeHalthDegreeState read GetSumHealthState;
    property SumHealthStateMsg: WideString read GetSumHealthStateMsg;

    property TotalMemory: int64 read GetTotalMemory write SetTotalMemory;
    property UsageMemory: int64 read GetUsageMemory write SetUsageMemory;
    property AvailableMemory: int64 read GetAvailableMemory write SetAvailableMemory;
    property CPUUsage: byte read GetCPUUsage write SetCPUUsage;
    property SumResources: TSumResources read GetSumResources;

    property View: TObject read GetView write SetView;
  end;

  THypervisorCinemaNode = class;

  THypervisorGroup = class(THypervisorNode, IHypervisorGroup)
  end;

  { TCinemaNodeClient }

  TCinemaNodeClient = class(TCinemaNetProtocol)
  private
    FHypervisor: THypervisor;
    FOwners: TList<THypervisorCinemaNode>;
    FClient: TCinemaSocketClient;
    FName: UnicodeString;
    FHostName: UnicodeString;
    FIsLocal: boolean;
    FOS: UnicodeString;
    FLastSendData: uint64;
    FLastTryConnected: uint64;

    procedure OnClientConnect(AClient: TCinemaSocketClient);
    procedure OnClientDisconnect(AClient: TCinemaSocketClient);

    function GetCountOwners: int32;
    function GetOwners(AIndex: int32): THypervisorCinemaNode;

    procedure OnProtocolError(ATlv: PTlvHeaderBase; const AMsg: UnicodeString);
  protected
    procedure DoEvent(const ASource, AWhoIs, AMsg: UnicodeString; ADateTime: TDateTime; AEventType: int32; const ASourceId: TGuid); override;
    procedure DoKey(const AKey: UnicodeString; APid: uint32; ANumber: int32); override;
    procedure DoHealthStateNode(const AKey: UnicodeString; AState: byte; AHealthState: byte; const AHealthStateMsg: UnicodeString; const AItemID: TGuid); override;
    procedure DoState(const ASource: TGuid; AState: int32); override;
    procedure DoAddOrUpdateScenario(const AKey, AName: UnicodeString; ACountInstances: int32); override;
    procedure DoDeleteScenario(const AKey: UnicodeString; ADeleteFile: boolean); override;
    // в данном случае это событие остановки
    procedure DoStopScenario(const AName: UnicodeString); override;
    procedure DoHostName(const AName: UnicodeString); override;
    procedure DoProcessorStat(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat; const ATables: UnicodeString); override;
    procedure DoProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString); override;
    procedure DoScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte); override;
    procedure DoNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte); override;
    procedure DoVersion(const AKey: UnicodeString; const AVersion: UnicodeString); override;
    procedure DoOs(const AOS: UnicodeString); override;
  public
    constructor Create(AHypervisor: THypervisor; const AName: UnicodeString; const AAddress: string); reintroduce;
    destructor Destroy; override;
    procedure AddOwner(AOwner: THypervisorCinemaNode);
    procedure RemoveOwner(AOwner: THypervisorCinemaNode);
    function Execute: boolean;

    property Name: UnicodeString read FName;
    property HostName: UnicodeString read FHostName;
    property IsLocal: boolean read FIsLocal;
    property CountOwners: int32 read GetCountOwners;
    property Owners[AIndex: int32]: THypervisorCinemaNode read GetOwners;
    property Client: TCinemaSocketClient read FClient;
    property OS: UnicodeString read FOS;
  end;

  { THypervisorCinemaNode }

  THypervisorCinemaNode = class(THypervisorNode, IHypervisorCinemaNode)
  private
    FClient: TCinemaNodeClient;
    FAddress: string;
    FVersion: WideString;
    // TODO: запоминать о необходимости обновить сцеанрий на удалённом узле
    //FNeedUpdateScenarios: boolean;
  public
    constructor Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode); override;
    destructor Destroy; override;

    function ToJson: WideString; override;
    procedure FromJson(AJson: TJSONCinema); override;

    procedure SendRequestRun(const AKey, AName: WideString; ACountInstances: int32); override;
    procedure SendRequestStop(const AKey: WideString); override;
    procedure SendRequestTerminate(const AKey: WideString); override;
    procedure SendRequestKill(const AKey: WideString); override;
    procedure SendRequestStatTables(const AKey: WideString; const AGuid: TGuid);
    procedure SendRequestStat;
    procedure SendUpdatedScenario(const AKey, AName: WideString; ACountInstances: int32);
    procedure SendRequestDeleteScenario(const AKey: WideString); override;
    procedure UpdateItself;
    procedure SendFile(const AFileName: string);

    function GetIsConnected: boolean;
    function GetAddress: string;
    procedure SetAddress(const Value: string);
    function GetUsageMemory: int64;
    function GetCPUUsage: byte;
    function GetTotalMemory: int64;
    procedure SetUsageMemory(const Value: int64);
    procedure SetCPUUsage(const Value: byte);
    procedure SetTotalMemory(const Value: int64);
    function GetVersion: WideString;
    procedure SetVersion(const Value: WideString);
    function GetOS: WideString;

    property Client: TCinemaNodeClient read FClient write FClient;

    property IsConnected: boolean read GetIsConnected;
    property Address: string read GetAddress write SetAddress;
    property Version: WideString read GetVersion write SetVersion;
    property OS: WideString read GetOS;
  end;

  { THypervisorScenario }

  THypervisorScenario = class(THypervisorNode, IHypervisorScenario)
  private
    FFileScenario: UnicodeString;
    FRelativeScenarioPath: UnicodeString;
    FScenario: ICinemaScenario;
    //FScenarioPrototype: ICinemaScenario;
    FCountInstances: int32;
    procedure SetFileScenario(const Value: UnicodeString);
  public
    constructor Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode); override;

    function ToJson: WideString; override;
    procedure FromJson(AJson: TJSONCinema); override;

    procedure SendRequestRun(const AKey, AName: WideString; ACountInstances: int32); override;
    procedure SendRequestStop(const AKey: WideString); override;
    procedure SendRequestTerminate(const AKey: WideString); override;
    procedure SendRequestKill(const AKey: WideString); override;
    procedure SendRequestStatTables(const AGuid: TGuid);
    procedure SendUpdatedScenario(ACountInstances: int32);
    procedure SendRequestDeleteScenario(const AKey: WideString); override;

    { IHypervisorScenario }

    function GetFileScenario: UnicodeString;
    function GetScenario: ICinemaScenario;
    //function GetScenarioPrototype: ICinemaScenario;
    function GetCountInstances: int32;
    procedure SetCountInstances(const AValue: int32);
    procedure ResetToState(AState: TNodeState);

    property FileScenario: UnicodeString read GetFileScenario write SetFileScenario;
    property Scenario: ICinemaScenario read GetScenario;
    //property ScenarioPrototype: ICinemaScenario read GetScenarioPrototype;
    property CountInstances: int32 read GetCountInstances write SetCountInstances;
  end;

  { TDefaultCinemaNodeClient }

  TDefaultCinemaNodeClient = class(TCinemaNetProtocol)
  private
    FOwner: THypervisor;
    FRemouteEndPoint: TNetEndpoint;
    FHostName: UnicodeString;
  protected
    procedure DoHostName(const AName: UnicodeString); override;
  public
    constructor Create(AOwner: THypervisor; ARemouteEndPoint: TNetEndpoint); reintroduce;
    property HostName: UnicodeString read FHostName;
    property RemouteEndPoint: TNetEndpoint read FRemouteEndPoint;
  end;

  THypervisor = class(THypervisorNode, IHypervisor)
  private
    type
      TWorkThread = class(TThread)
      private
        FOwner: THypervisor;
      protected
        procedure Execute; override;
      public
        constructor Create(AOwner: THypervisor);
      end;
  private
    class var FInstallDir: string;
    class var FLocalHostName: UnicodeString;
    class constructor Create;
  private
    FIsDeserializingNow: Boolean;
    FThread: TWorkThread;
    FCountConnectedCinemaNodes: int32;
    FCinemaNodes: TList<THypervisorCinemaNode>;
    FCinemaNodesDic: TDictionary<WideString, THypervisorCinemaNode>;
    FCinemaNodeClientsDic: TDictionary<string, TCinemaNodeClient>;
    FNodesDic: TDictionary<UnicodeString, IHypervisorNode>;
    FScenarios: TDictionary<TGuid, ICinemaScenario>;
    FLog: ICinemaLog;
    FOnCreateNode: THypervisorNodeEvent;
    FOnDestroyNode: THypervisorNodeEvent;
    FOnRemouteDeleteNode: THypervisorNodeEvent;
    FOnConnectCinemaNodeEvent: THypervisorNodeEvent;
    FOnDisconnectCinemaNodeEvent: THypervisorNodeEvent;
    FListener: THypervisorEventsListener;
    // узлы которые подключились сами к гипервизору, а не гипервизор к ним
    FDefaultCinemaNodes: TList<TDefaultCinemaNodeClient>;
    // сервер для подключения со стороны узлов; если подключится узел которого нет в конфигурации, он автоматически добавиться в корень
    FDefaultCinemaNodeServer: TCinemaSocketServer;
    FSelectedCinemaNode: IHypervisorCinemaNode;
    FServer: TCinemaSocketServer;

    function GetIsDeserializingNow: Boolean;

    //procedure OnAcceptNodeClient(ANodeClient: TCinemaNodeClient);
    procedure OnReceiveNodeClient(AOwner: Pointer; AData: PByte; ALen: Integer);
    procedure OnCloseNodeClient(AOwner: Pointer);

    function OnAcceptDefault(ASocket: TSocketHandle; ARemouteEndPoint: TNetEndpoint): Pointer;
    procedure OnReceiveDefault(AOwner: Pointer; AData: PByte; ALen: Integer);
    procedure OnCloseDefault(AOwner: Pointer);

    // если какие либо данные были получены или отправлены вернёт true
    function Ask: boolean;
    function GetSelectedCinemaNode: IHypervisorCinemaNode;
    procedure SetSelectedCinemaNode(const Value: IHypervisorCinemaNode);
    function FindNode(const AKey: WideString): IHypervisorNode;
  protected
    procedure OnCreateCinemaNode(ACinemaNode: THypervisorCinemaNode);
    procedure OnDestroyCinemaNode(ACinemaNode: THypervisorCinemaNode);
    procedure OnAddChild(const AChild: THypervisorNode); virtual;
    procedure OnRemoveChild(const AChild: THypervisorNode); virtual;
    procedure OnConnectCinemaNodeClient(AClient: TCinemaNodeClient);
    procedure OnDisconnectCinemaNodeClient(AClient: TCinemaNodeClient);
    procedure OnAcceptDefaultCinemaNode(ACinemNode: TDefaultCinemaNodeClient);

    procedure DoOnChangeChildState(ASource: THypervisorNode); override;
    procedure DoOnChangeChildLogState(ASource: THypervisorNode); override;
    procedure DoOnChangeChildHealthState(ASource: THypervisorNode); override;
    procedure DoProcessorStat(const AKey, AHostName: UnicodeString; const AStats: TTLVScenarioProcessorStat; const ATables: UnicodeString);
    procedure DoProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString);
    procedure DoScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte);
    procedure DoNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte);
    procedure DoProcessVersion(const AKey, AVersion: UnicodeString);

    procedure SendEvent(const AMsg: WideString; AEventType: int32); overload;
    procedure SendEvent(const AKey, ASource, AWhoIs, AMsg: WideString; ADateTime: TDateTime; AEventType: int32; ASourceId: TGuid); overload;
    procedure SendHealthStateNodeEvent(const AKey: WideString; AState: byte; AHealthState: byte; const AHealthStateMsg: WideString; const AItemID: TGuid);
    procedure ChangeStateEvent(const AKey: WideString; const ASource: TGuid; AState: TNodeState);
    procedure StopScenarioEvent(const AKey: WideString);
    procedure AddOrUpdateScenario(const AKey, AName: WideString; ACountInstances: int32);
    procedure DeleteScenario(const AKey: WideString; ADeleteFile: boolean);

  public
    constructor Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode); override;
    destructor Destroy; override;
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;

    function GetScenario(const AFileName: string): ICinemaScenario;
    function FindLikeNode(const ALikeName: WideString; const AFromNode: IHypervisorNode): IHypervisorNode;

    procedure Start;
    procedure Stop;

    function ToJson: WideString; override;
    procedure FromJson(AJson: TJSONCinema); override;

    function GetIsRun: boolean;
    function GetLog: ICinemaLog;

    function GetEventsListener: THypervisorEventsListener;
    procedure SetEventsListener(const Value: THypervisorEventsListener);
    function GetOnCreateNode: THypervisorNodeEvent;
    function GetOnDestroyNode: THypervisorNodeEvent;
    procedure SetOnCreateNode(const Value: THypervisorNodeEvent);
    procedure SetOnDestroyNode(const Value: THypervisorNodeEvent);
    function GetOnRemouteDeleteNode: THypervisorNodeEvent;
    procedure SetOnRemouteDeleteNode(const Value: THypervisorNodeEvent);
    function GetOnConnectCinemaNodeEvent: THypervisorNodeEvent;
    procedure SetOnConnectCinemaNodeEvent(AValue: THypervisorNodeEvent);
    function GetOnDisconnectCinemaNodeEvent: THypervisorNodeEvent;
    procedure SetOnDisconnectCinemaNodeEvent(AValue: THypervisorNodeEvent);
    function GetCountCinemaNodes: int32;
    function GetCountConnectedCinemaNodes: int32;

    property IsRun: boolean read GetIsRun;
    property Log: ICinemaLog read FLog;
    property EventsListener: THypervisorEventsListener read GetEventsListener write SetEventsListener;
    property CountCinemaNodes: int32 read GetCountCinemaNodes;
    property CountConnectedCinemaNodes: int32 read GetCountConnectedCinemaNodes;
    property SelectedCinemaNode: IHypervisorCinemaNode read GetSelectedCinemaNode write SetSelectedCinemaNode;
    property IsDeserializingNow: Boolean read GetIsDeserializingNow;

    property OnCreateNode: THypervisorNodeEvent read GetOnCreateNode write SetOnCreateNode;
    property OnDestroyNode: THypervisorNodeEvent read GetOnDestroyNode write SetOnDestroyNode;
    property OnRemouteDeleteNode: THypervisorNodeEvent read GetOnRemouteDeleteNode write SetOnRemouteDeleteNode;
    property OnConnectCinemaNodeEvent: THypervisorNodeEvent read GetOnConnectCinemaNodeEvent write SetOnConnectCinemaNodeEvent;
    property OnDisconnectCinemaNodeEvent: THypervisorNodeEvent read GetOnDisconnectCinemaNodeEvent write SetOnDisconnectCinemaNodeEvent;

    class property LocalHostName: UnicodeString read FLocalHostName;
  end;


implementation

uses
    Math
  , System.IOUtils
  , SysUtils
  , StrUtils
  {$ifndef fpc}
    {$ifdef posix}
  , Posix.Unistd
    {$endif}
  {$endif}
  , Cinema.Log
  , Cinema.Utils
  , Cinema.Exceptions
  , Cinema.Interfaces
  , Cinema.Interfaces.FuncExport
  , Cinema.Configuration
  , Cinema.FPC.Threading
  //, Cinema.Scenario
  ;

{ THypervisorNode }

procedure THypervisorNode.AddChild(const AChild: IHypervisorNode);
begin
  TMonitor.Enter(Self);
  try
    FChildren.Add(AChild);
  finally
    TMonitor.Exit(Self);
  end;
end;

function THypervisorNode.AddCinemaNode(const AName: WideString;
  const AAddress: AnsiString): IHypervisorCinemaNode;
var
  cinemaNode: THypervisorCinemaNode;
begin
  TMonitor.Enter(Self);
  try
    cinemaNode := THypervisorCinemaNode.Create(Owner, AName, Self);
    cinemaNode.Address := AAddress;
    Result := cinemaNode;
  finally
    TMonitor.Exit(Self);
  end;
  FOwner.OnAddChild(cinemaNode);
end;

function THypervisorNode.AddNode(const AName: WideString): IHypervisorNode;
var
  node: THypervisorNode;
begin
  TMonitor.Enter(Self);
  try
    node := THypervisorGroup.Create(Owner, AName, Self);
    Result := node;
  finally
    TMonitor.Exit(Self);
  end;
  FOwner.OnAddChild(node);
end;

function THypervisorNode.AddScenario(const AFileName: UnicodeString): IHypervisorScenario;
var
  scen: THypervisorScenario;
  dir: UnicodeString;
  fileScen: UnicodeString;
  fileScenSrc: UnicodeString;
  splitted: TArray<UnicodeString>;
begin
  TMonitor.Enter(Self);
  try
    if Assigned(CinemaNode) then
    begin
      dir := TCinemaConfiguration.DtlAppPath + CinemaNode.Name + PathDelim + TCinemaNetProtocol.csWorkSubDirScens;
      if not DirectoryExists(dir) then
        ForceDirectories(dir);
      fileScen := dir + ExtractFileName(AFileName);
      if fileScen.Contains(AFileName) then
      begin
        splitted := AFileName.Split([TCinemaNetProtocol.csWorkDirectory + CinemaNode.Name + PathDelim + TCinemaNetProtocol.csWorkSubDirScens]);
        if (length(splitted) = 2) and (splitted[0] = '') then
          fileScenSrc := dir + ExtractFileName(AFileName)
        else
          fileScenSrc := AFileName;
      end else
        fileScenSrc := AFileName;

      if fileScen <> fileScenSrc then
        TFile.Copy(WideToString(fileScenSrc), WideToString(fileScen), true);
    end else
    begin
      fileScen := AFileName;
    end;

    scen := THypervisorScenario.Create(Owner, ExtractFileName(ChangeFileExt(fileScen, UnicodeString(''))), Self);
    scen.FileScenario := fileScen;
    Result := scen;
  finally
    TMonitor.Exit(Self);
  end;
  FOwner.OnAddChild(scen);
end;

procedure THypervisorNode.CalcSumState;
var
  child: IHypervisorNode;
begin

  FSumHealthState := min(FNodeHealthState, MSG_TYPE_TO_HEALTH_DEGREE[FNodeLogState]);
  if FNodeHealthState > MSG_TYPE_TO_HEALTH_DEGREE[FNodeLogState] then
    FSumHealthStateMsg := FNodeLogStateMsg;

  if FNodeState <> TNodeState.nsNone then
    FSumWorkState := [FNodeState]
  else
    FSumWorkState := [];


  for child in FChildren do
  begin
    FSumWorkState := FSumWorkState + child.SumWorkState;
    FSumHealthState := FSumHealthState * child.SumHealthState div 100;
  end;

  if FSumWorkState = [] then
    FSumWorkState := [TNodeState.nsNone];
end;

function THypervisorNode.CanAddCinemaNode: boolean;
begin
  Result := not Assigned(FCinemaNode);
end;

function THypervisorNode.CanAddNode: boolean;
begin
  Result := not Assigned(FCinemaNode) or not Assigned(FScenarioNode);
end;

function THypervisorNode.CanAddScenario: boolean;
begin
  Result := not Assigned(FScenarioNode);
end;

function THypervisorNode.CanStart: boolean;
var
  child: IHypervisorNode;
begin
  Result := Assigned(FCinemaNode) and Assigned(FScenarioNode) and (TNodeState.nsNone = FNodeState) and (FCinemaNode.IsConnected);
  if not Result then
    for child in FChildren do
    begin
      Result := child.CanStart;
      if Result then
        break;
    end;
end;

function THypervisorNode.CanStop: boolean;
var
  child: IHypervisorNode;
begin
  Result := Assigned(FCinemaNode) and Assigned(FScenarioNode) and (FNodeState <> TNodeState.nsNone);
  if not Result then
    for child in FChildren do
    begin
      Result := child.CanStop;
      if Result then
        break;
    end;
end;

function THypervisorNode.ChildExists(const AName: WideString; const ANodeType: TGUID; out ANode: IHypervisorNode): boolean;
var
  child: IHypervisorNode;
begin
  Result := false;
  ANode := nil;
  for child in FChildren do
  begin
    Result := (child.Name = AName) and Supports(child, ANodeType);
    if Result then
    begin
      ANode := child;
      break;
    end;
  end;
end;

function THypervisorNode.ChildExists(const AName: WideString; const ANodeType: TGUID): boolean;
var
  child: IHypervisorNode;
begin
  Result := false;
  for child in FChildren do
  begin
    Result := (child.Name = AName) and Supports(child, ANodeType);
    if Result then
      break;
  end;
end;

class constructor THypervisorNode.Create;
begin
  FClassNodes := TDictionary<string, THypervisorNodeClass>.Create;
  THypervisorNode.RegisterClassNode(THypervisorNode);
  THypervisorNode.RegisterClassNode(THypervisorGroup);
  THypervisorNode.RegisterClassNode(THypervisorCinemaNode);
  THypervisorNode.RegisterClassNode(THypervisorScenario);
  THypervisorNode.RegisterClassNode(THypervisor);
end;

constructor THypervisorNode.Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode);
begin
  FOwner := AOwner;
  FChildren := TList<IHypervisorNode>.Create;
  FParent := AParent;
  FName := FName.Copy(AName);
  ResetState;
  if Assigned(FParent) then
    (FParent.GetObject as THypervisorNode).AddChild(Self);
  FindParentsTypes;
end;

class function THypervisorNode.CreateNode(const AClassName: string; AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode): THypervisorNode;
var
  classNode: THypervisorNodeClass;
begin
  if FClassNodes.TryGetValue(AClassName, classNode) then
    Result := classNode.Create(AOwner, AName, AParent)
  else
    raise ECinema.Create('Uncknown node class:' + AClassName);
end;

procedure THypervisorNode.DeleteChild(const AChild: IHypervisorNode);
begin
  RemoveChild(AChild);
  if Assigned(FOwner) then
    FOwner.OnRemoveChild(THypervisorNode(AChild.GetObject));

  while AChild.CountChildren > 0 do
    AChild.DeleteChild(AChild.Child[AChild.CountChildren-1]);

  CalcSumState;
end;

class destructor THypervisorNode.Destroy;
begin
  FClassNodes.Free;
end;

destructor THypervisorNode.Destroy;
begin

  FChildren.Clear; // before need to clear children

  Pointer(FParent) := nil;
  Pointer(FCinemaNode) := nil;
  Pointer(FScenarioNode) := nil;

  FChildren.Free;
  inherited;
end;

procedure THypervisorNode.DoOnChangeChildState(ASource: THypervisorNode);
begin
  CalcSumState;
  if Assigned(FParent) then
    THypervisorNode(FParent.GetObject).DoOnChangeChildState(ASource);
end;

procedure THypervisorNode.DoOnChangeChildHealthState(ASource: THypervisorNode);
begin
  CalcSumState;
  if Assigned(FParent) then
    THypervisorNode(FParent.GetObject).DoOnChangeChildHealthState(ASource);
end;

procedure THypervisorNode.DoOnChangeChildLogState(ASource: THypervisorNode);
begin
  CalcSumState;
  if Assigned(FParent) then
    THypervisorNode(FParent.GetObject).DoOnChangeChildLogState(ASource);
end;

procedure THypervisorNode.FindParentsTypes;
var
  p: THypervisorNode; // если использовать интерфейс то объект релизнется не успев отдаться создателю, т.к. этот метод вызывается из AfterConstruction
  cinemaNode: IHypervisorCinemaNode;
  scenario: IHypervisorScenario;
  simpleKey: WideString;
begin
  p := Self;
  simpleKey := '';
  FKey := '';
  while Assigned(p) do
  begin
    if not Assigned(FCinemaNode) and Supports(p, IHypervisorCinemaNode, cinemaNode) then
    begin
      Pointer(FCinemaNode) := Pointer(cinemaNode);
      cinemaNode := nil;
//      if Assigned(FScenarioNode) then
//      begin
//        //p := THypervisorNode(THypervisorNode(FScenarioNode.GetObject).Parent.GetObject);
//        //FKey := p.Key + '/' + FScenarioNode.Name;
//        break;
//      end;
    end;

    if not Assigned(FScenarioNode) and Supports(p, IHypervisorScenario, scenario) then
    begin
      Pointer(FScenarioNode) := Pointer(scenario);
      scenario := nil;
//      if Assigned(FCinemaNode) then
//      begin
//        //FKey := cinemaNode.Key + '/' + FScenarioNode.Name;
//        break;
//      end;
    end;

//    if Assigned(FScenarioNode) or Assigned(FCinemaNode) then
//    begin
//      if FKey.IsEmpty then
//        FKey := p.Name
//      else
//        FKey := FKey + '/' + p.Name;
//    end;
//
    if Assigned(FScenarioNode) and Assigned(FCinemaNode) then
    begin
      FKey := FCinemaNode.Name + '/' + FScenarioNode.Name;
      break;
    end;

    if simpleKey.IsEmpty then
      simpleKey := p.Name
    else
      simpleKey := simpleKey + '/' + p.Name;

    if Assigned(p.Parent) and (p.Parent.GetObject <> FOwner) then
      p := THypervisorNode(p.Parent.GetObject)
    else
      break;
  end;

  if FKey.IsEmpty then
  begin
    if Assigned(FCinemaNode) and (FCinemaNode.GetObject = Self) then
      FKey := FName
    else
      FKey := simpleKey;
  end;
end;

procedure THypervisorNode.FromJson(AJson: TJSONCinema);
begin
  //FName := AJson.GetValue<string>('name');
  FFolded := AJson.GetValue<boolean>('folded');
end;

function THypervisorNode.GetChild(Index: int32): IHypervisorNode;
begin
  Result := FChildren.Items[Index];
end;

function THypervisorNode.GetCinemaNode: IHypervisorCinemaNode;
begin
  Result := FCinemaNode;
end;

function THypervisorNode.GetCountChildren: int32;
begin
  Result := FChildren.Count;
end;

function THypervisorNode.GetFolded: boolean;
begin
  Result := FFolded;
end;

function THypervisorNode.GetKey: WideString;
begin
  if FKey.IsEmpty then
  begin
    if Assigned(FParent) then
      Result := FParent.Key + '/' + FName
    else
      Result := FName
  end else
    Result := FKey;
end;

function THypervisorNode.GetName: WideString;
begin
  Result := FName;
end;

function THypervisorNode.GetNodeLogState: TNodeLogState;
begin
  Result := FNodeLogState;
end;

function THypervisorNode.GetNodeState: TNodeState;
begin
  Result := FNodeState;
end;

function THypervisorNode.GetObject: TInterfacedObject;
begin
  Result := Self;
end;

function THypervisorNode.GetParent: IHypervisorNode;
begin
  Result := FParent;
end;

function THypervisorNode.GetScenarioNode: IHypervisorScenario;
begin
  Result := FScenarioNode;
end;

function THypervisorNode.GetSumHealthState: TNodeHalthDegreeState;
begin
  Result := FSumHealthState;
end;

function THypervisorNode.GetSumHealthStateMsg: WideString;
begin
  Result := FSumHealthStateMsg;
end;

function THypervisorNode.GetSumResources: TSumResources;
var
  child: IHypervisorNode;
begin
  Result.TotalMemory := FTotalMemory;
  Result.AvailableMemory := FAvailableMemory;
  Result.UsageMemory := FUsageMemory;
  Result.CPUUsage := FCPUUsage;
  for child in FChildren do
  begin
    Result.TotalMemory := Result.TotalMemory + child.TotalMemory;
    Result.UsageMemory := Result.UsageMemory + child.UsageMemory;
    Result.AvailableMemory := Result.AvailableMemory + child.AvailableMemory;
    Result.CPUUsage := (Result.CPUUsage + child.CPUUsage) shr 1;
  end;
end;

function THypervisorNode.GetSumWorkState: TNodeStates;
begin
  Result := FSumWorkState;
end;

function THypervisorNode.GetCPUUsage: byte;
begin
  Result := FCPUUsage;
end;

procedure THypervisorNode.SetCPUUsage(const Value: byte);
begin
  FCPUUsage := Value;
end;

function THypervisorNode.GetUsageMemory: int64;
begin
  Result := FUsageMemory;
end;

function THypervisorNode.GetTotalMemory: int64;
begin
  Result := FTotalMemory;
end;

function THypervisorNode.GetAvailableMemory: int64;
begin
  Result := FAvailableMemory;
end;

procedure THypervisorNode.SetAvailableMemory(const Value: int64);
begin
  FAvailableMemory := Value;
end;

procedure THypervisorNode.SetUsageMemory(const Value: int64);
begin
  FUsageMemory := Value;
end;

procedure THypervisorNode.SetTotalMemory(const Value: int64);
begin
  FTotalMemory := Value;
end;

procedure THypervisorNode.SendRequestDeleteScenario(const AKey: WideString);
var
  node: IHypervisorNode;
begin
  TMonitor.Enter(Self);
  try
    if Assigned(FScenarioNode) and Assigned(FCinemaNode) then
      FCinemaNode.SendRequestDeleteScenario(FScenarioNode.Key)
    else
    for node in FChildren do
      node.SendRequestDeleteScenario('');
  finally
    TMonitor.Exit(Self);
  end;
end;

procedure THypervisorNode.SendRequestKill(const AKey: WideString);
var
  node: IHypervisorNode;
begin
  TMonitor.Enter(Self);
  try
    if Assigned(FScenarioNode) and Assigned(FCinemaNode) then
      FCinemaNode.SendRequestKill(FScenarioNode.Key)
    else
    for node in FChildren do
      node.SendRequestKill('');
  finally
    TMonitor.Exit(Self);
  end;
end;

class procedure THypervisorNode.RegisterClassNode(AClassNode: THypervisorNodeClass);
begin
  FClassNodes.Add(AClassNode.ClassName, AClassNode);
end;

procedure THypervisorNode.RemoveChild(const AChild: IHypervisorNode);
begin
  TMonitor.Enter(Self);
  try
    FChildren.Remove(AChild);
  finally
    TMonitor.Exit(Self);
  end;
end;

procedure THypervisorNode.ResetState;
begin
  FNodeHealthState := HEALTH_STATE_OK;
  FSumHealthState := HEALTH_STATE_OK;
  FSumWorkState := [TNodeState.nsNone];
  FNodeLogState := TNodeLogState.lsInfo;
  FNodeLogStateMsg := '';
end;

procedure THypervisorNode.SendRequestRun(const AKey, AName: WideString; ACountInstances: int32);
var
  node: IHypervisorNode;
begin
  TMonitor.Enter(Self);
  try
    if Assigned(FScenarioNode) and Assigned(FCinemaNode) then
      FCinemaNode.SendRequestRun(FScenarioNode.Key, FScenarioNode.FileScenario, ACountInstances)
    else
    for node in FChildren do
      node.SendRequestRun('', '', ACountInstances);
  finally
    TMonitor.Exit(Self);
  end;
end;

procedure THypervisorNode.SetFolded(const Value: boolean);
begin
  FFolded := Value;
end;

procedure THypervisorNode.SetHealthState(const AMessage: WideString; AState: TNodeHalthDegreeState);
begin
  if (AState = FNodeHealthState) and (AMessage = FNodeHealthStateMsg) then
    exit;
  TMonitor.Enter(Self);
  try
    FNodeHealthStateMsg := AMessage;
    FNodeHealthState := AState;
    CalcSumState;
  finally
    TMonitor.Exit(Self);
  end;
  if Assigned(FParent) then
    THypervisorNode(FParent.GetObject).DoOnChangeChildHealthState(Self);
end;

procedure THypervisorNode.SetLogState(const AMessage: WideString; ALogState: TNodeLogState);
begin
  // состоянием лога сигнализируем только об ухудшении ситуации: Info->Warning->Error
  if ALogState > FNodeLogState then
  begin
    TMonitor.Enter(Self);
    try
      FNodeLogState := ALogState;
      FNodeLogStateMsg := AMessage;

      CalcSumState;
    finally
      TMonitor.Exit(Self);
    end;

    if Assigned(FParent) then
      THypervisorNode(FParent.GetObject).DoOnChangeChildLogState(Self);
  end;
end;

procedure THypervisorNode.SetName(const AValue: WideString);
begin
  FName := AValue;
end;

procedure THypervisorNode.SetParent(const AValue: IHypervisorNode);
begin
  if Assigned(FParent) then
    (FParent as THypervisorNode).RemoveChild(Self);
  Pointer(FParent) := Pointer(AValue);
  if Assigned(FParent) then
    (FParent as THypervisorNode).AddChild(Self);
  FindParentsTypes;
end;

procedure THypervisorNode.SetWorkState(AState: TNodeState);
begin
  if FNodeState = AState then
    exit;
  TMonitor.Enter(Self);
  try
    FNodeState := AState;
    CalcSumState;
  finally
    TMonitor.Exit(Self);
  end;
  if Assigned(FParent) then
    THypervisorNode(FParent.GetObject).DoOnChangeChildState(Self);
end;

function THypervisorNode.GetView: TObject;
begin
  Result := FView;
end;

procedure THypervisorNode.ProcessError(const ActionName: WideString; AErrorCode: int32);
begin
  case AErrorCode of
    csHypervisorErrorNodeAlreadyExist: FOwner.SendEvent(ActionName + ': ' + 'подобный эелемент уже существует', MESSAGE_INFO);
  end;
end;

procedure THypervisorNode.SetView(AValue: TObject);
begin
  FView := AValue;
end;

procedure THypervisorNode.SendRequestStop(const AKey: WideString);
var
  node: IHypervisorNode;
begin
  TMonitor.Enter(Self);
  try
    if Assigned(FScenarioNode) and Assigned(FCinemaNode) then
      FCinemaNode.SendRequestStop(FScenarioNode.Key)
    else
    for node in FChildren do
      node.SendRequestStop('');
  finally
    TMonitor.Exit(Self);
  end;
end;

procedure THypervisorNode.SendRequestTerminate(const AKey: WideString);
var
  node: IHypervisorNode;
begin
  TMonitor.Enter(Self);
  try
    if Assigned(FScenarioNode) and Assigned(FCinemaNode) then
      FCinemaNode.SendRequestTerminate(FScenarioNode.Key)
    else
    for node in FChildren do
      node.SendRequestTerminate('');
  finally
    TMonitor.Exit(Self);
  end;
end;

function THypervisorNode.ToJson: WideString;
var
  i: int32;
  child: IHypervisorNode;
begin
  Result :=
    '"name":"' + FName + '",' + sLineBreak +
    '"folded":"' + WideString(BoolToStr(FFolded, true)) + '",' + sLineBreak +
    '"className":"' + WideString(ClassName) + '",' + sLineBreak +
    '"children": [' + sLineBreak;

  for i := 0 to FChildren.Count - 1 do
  begin
    child := FChildren.Items[i];
    if i > 0 then
      Result := Result + ',{' + child.ToJson + '}' +  sLineBreak
    else
      Result := Result + '{' + child.ToJson + '}' +  sLineBreak;
  end;

  Result := Result + ']';
end;

{ THypervisorScenario }

function THypervisorScenario.GetCountInstances: int32;
begin
  Result := FCountInstances;
end;

constructor THypervisorScenario.Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode);
begin
  inherited Create(AOwner, AName, AParent);
  FCountInstances := 1;
  FScenario := Cinema.Interfaces.FuncExport.CreateScenario(PWideChar(@FName[1]));
end;

procedure THypervisorScenario.FromJson(AJson: TJSONCinema);
var
  path: UnicodeString;
begin
  inherited;
  path := AJson.GetValue<UnicodeString>('FileScenario');
  if FileExists(path) then
    FileScenario := path
  else
  if FileExists(TCinemaConfiguration.DtlWorkDir + path) then
    FileScenario := TCinemaConfiguration.DtlWorkDir + path
  else
  if FileExists(TCinemaConfiguration.CinemaPath + path) then
    FileScenario := TCinemaConfiguration.CinemaPath + path
  else
    FileScenario := path;
end;

function THypervisorScenario.GetFileScenario: UnicodeString;
begin
  Result := FFileScenario;
end;

function THypervisorScenario.GetScenario: ICinemaScenario;
begin
  Result := FScenario;
end;

procedure THypervisorScenario.ResetToState(AState: TNodeState);

  procedure ResetLayerState(ALayer: IScenarioLayer);
  var
    i: int32;
    node: IScenarioNode;
    layerChild: IScenarioNode;
    layer: IScenarioLayer;
  begin
    for i := 0 to ALayer.CountChildren - 1 do
    begin
      layerChild := IScenarioNode(ALayer.Children[i]);
      if layerChild.QueryInterface(IScenarioLayer, layer) = 0 then
        ResetLayerState(layer);
      layerChild.State := AState;
      if AState = TNodeState.nsNone then
      begin
        layerChild.ChangeHealthState('Ок', 100);
        layerChild.ResetLogState;
      end;
    end;
    ALayer.QueryInterface(IScenarioNode, node);
    node.State := AState;
    if AState = TNodeState.nsNone then
      node.ResetLogState;
  end;

var
  mainLayer: IScenarioLayer;
begin
  TMonitor.Enter(Self);
  try
    if Scenario.QueryInterface(IScenarioLayer, mainLayer) = S_OK then
      ResetLayerState(mainLayer);
    if AState = TNodeState.nsNone then
      ResetState;
    SetWorkState(AState);
  finally
    TMonitor.Exit(Self);
  end;
  if AState = TNodeState.nsNone then
    SetHealthState('', 100);
end;

procedure THypervisorScenario.SendRequestDeleteScenario(const AKey: WideString);
begin
  if Assigned(CinemaNode) then
    CinemaNode.SendRequestDeleteScenario(Key)
  else
    inherited;
end;

procedure THypervisorScenario.SendRequestKill(const AKey: WideString);
begin
  if Assigned(CinemaNode) then
    CinemaNode.SendRequestKill(Key)
  else
    inherited;
end;

procedure THypervisorScenario.SendRequestRun(const AKey, AName: WideString; ACountInstances: int32);
begin
  if Assigned(CinemaNode) then
  begin
    if (Scenario as IScenarioNode).State = TNodeState.nsNone then
      ResetToState(TNodeState.nsNone);
    CinemaNode.SendRequestRun(Key, FFileScenario, ACountInstances)
  end else
    inherited;
end;

procedure THypervisorScenario.SetCountInstances(const AValue: int32);
begin
  if AValue < 1 then
    exit;
  FCountInstances := AValue;
end;

procedure THypervisorScenario.SetFileScenario(const Value: UnicodeString);
var
  splitted: TArray<UnicodeString>;
begin
  FFileScenario := Value;
  splitted := FFileScenario.Split([TCinemaConfiguration.DtlWorkDir]);
  if length(splitted) = 2 then
    FRelativeScenarioPath := splitted[1]
  else
  begin
    splitted := FFileScenario.Split([TCinemaConfiguration.CinemaPath]);
    if length(splitted) = 2 then
      FRelativeScenarioPath := splitted[1]
    else
      FRelativeScenarioPath := FFileScenario;
  end;

  if FileExists(FFileScenario) then
  begin
    FScenario.LoadScenario(PWideChar(FFileScenario));
  end;
end;

procedure THypervisorScenario.SendRequestStatTables(const AGuid: TGuid);
begin
  if Assigned(FScenarioNode) and Assigned(FCinemaNode) then
    FCinemaNode.SendRequestStatTables(FScenarioNode.Key, AGuid)
  else
    inherited;
end;

procedure THypervisorScenario.SendRequestStop(const AKey: WideString);
begin
  if Assigned(CinemaNode) then
    CinemaNode.SendRequestStop(Key)
  else
    inherited;
end;

procedure THypervisorScenario.SendRequestTerminate(const AKey: WideString);
begin
  if Assigned(CinemaNode) then
    CinemaNode.SendRequestTerminate(Key)
  else
    inherited;
end;

procedure THypervisorScenario.SendUpdatedScenario(ACountInstances: int32);
begin
  FCountInstances := ACountInstances;
  if Assigned(CinemaNode) then
    CinemaNode.SendUpdatedScenario(Key, FFileScenario, ACountInstances);
end;

function THypervisorScenario.ToJson: WideString;
var
  jsnString: TJSONString;
{$ifndef fpc}
  builder: TStringBuilder;
begin
  jsnString := TJSONString.Create(FRelativeScenarioPath);
  builder := TStringBuilder.Create;
  try
    jsnString.ToJSON(builder, []);
    Result := '"FileScenario":' + builder.ToString + ',' + sLineBreak + inherited;
  finally
    jsnString.Free;
    builder.Free;
  end;
end;
{$else}
 begin
  jsnString := TJSONString.Create(FRelativeScenarioPath);
  try
    // TODO check
    Result := WideString('"FileScenario":' + jsnString.AsJSON + ',' + sLineBreak) + inherited;
  finally
    jsnString.Free;
  end;
 end;
{$endif}

{ THypervisor }

procedure THypervisor.AddOrUpdateScenario(const AKey, AName: WideString; ACountInstances: int32);
var
  parentNode: IHypervisorNode;
  cinemaNode: IHypervisorNode;
  i: int32;
  snipents: TArray<UnicodeString>;
  lowCase: WideString;
begin

  lowCase := LowerCase(AKey);
  snipents := UnicodeString(lowCase).Split(['/', '\']);
  if length(snipents) < 2 then
    exit;

  if not FNodesDic.TryGetValue(snipents[0], cinemaNode) then
  begin
    SendEvent('В конфигурации не найден узел "' + snipents[0] + '"',  MESSAGE_WARN);
    exit;
  end;

  parentNode := cinemaNode;

  // проверим что сценарий ещё не добавлен к CinemaNode
  if FNodesDic.TryGetValue(lowCase, cinemaNode) then
  begin
    // перегрузим сценарий
    if Assigned(cinemaNode.ScenarioNode) and FileExists(cinemaNode.ScenarioNode.FileScenario) then
    begin
      cinemaNode.ScenarioNode.Scenario.LoadScenario(PWideChar(cinemaNode.ScenarioNode.FileScenario));
      // нужно отправить событие о перегрузке сценария
      if Assigned(FListener) then
        FListener.OnReloadScenario(cinemaNode);
    end;
    exit;
  end;

  for i := 1 to length(snipents) - 2 do
  begin
    parentNode := parentNode.AddNode(snipents[i]);
    {$ifdef DEBUG}
    SendEvent('Добавлен узел "' + snipents[i] + '"',  MESSAGE_INFO);
    {$endif}
  end;

  parentNode.AddScenario(AName);
  {$ifdef DEBUG}
  SendEvent('Добавлен сценарий "' + snipents[0] + '"',  MESSAGE_INFO);
  {$endif}
end;

procedure THypervisor.AfterConstruction;
begin
  inherited;
  FDefaultCinemaNodeServer.Start(true);
end;

function THypervisor.Ask: boolean;
var
  i: Integer;
  node: THypervisorCinemaNode;
begin
  Result := false;
  TMonitor.Enter(FCinemaNodeClientsDic);
  try
    //if Assigned(FSelectedCinemaNode) then
    //  if (FSelectedCinemaNode as THypervisorCinemaNode).Client.Execute then
    //    Result := true;

    if FServer.ReceiveFromClients(0) then
    begin
      Result := true;
    end;
    for i := 0 to FCinemaNodes.Count - 1 do
    begin
      node := FCinemaNodes.Items[i];
      if node.Client.Execute then
        Result := true;
    end;
  finally
    TMonitor.Exit(FCinemaNodeClientsDic);
  end;

  // сервер приёма самоподключающихся узлов по дефолтному адресу прописанному в hypervisors.json
  FDefaultCinemaNodeServer.Receive;
end;

procedure THypervisor.BeforeDestruction;
begin
  inherited;
  FDefaultCinemaNodeServer.Stop;

end;

procedure THypervisor.ChangeStateEvent(const AKey: WideString; const ASource: TGuid; AState: TNodeState);
var
  child: IHypervisorNode;
  source: IScenarioNode;
  //splitted:
begin
  child := FindNode(AKey);
  if Assigned(child) then
  begin
    if Assigned(child.ScenarioNode) and not ASource.IsEmpty then
    begin
      source := IScenarioNode(child.ScenarioNode.Scenario.GetChildByGuid(ASource));
      if Assigned(source) then
      begin
        source.State := AState;
        child.SetWorkState((child.ScenarioNode.Scenario as IScenarioNode).State);
      end;
    end else
      child.SetWorkState(AState);
  end;
end;

class constructor THypervisor.Create;
begin
  FInstallDir := IncludeTrailingPathDelimiter(ExtractFilePath(ParamStr(0)));
  FLocalHostName := WideString(GetLocalHostName);
end;

constructor THypervisor.Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode);
begin
  inherited Create(Self, AName, AParent);
  FCinemaNodes := TList<THypervisorCinemaNode>.Create;
  FCinemaNodesDic := TDictionary<WideString, THypervisorCinemaNode>.Create;
  FCinemaNodeClientsDic := TDictionary<string, TCinemaNodeClient>.Create;
  FNodesDic := TDictionary<UnicodeString, IHypervisorNode>.Create;
  FScenarios := TDictionary<TGuid, ICinemaScenario>.Create;
  FLog := TCinemaLog.Create;
  FDefaultCinemaNodes := TList<TDefaultCinemaNodeClient>.Create;
  FDefaultCinemaNodeServer := TCinemaSocketServer.Create('0.0.0.0', TCinemaNetProtocol.csServerForCinemaNodePort, TSocketType.TCP);
  FDefaultCinemaNodeServer.OnAccept := OnAcceptDefault;
  FDefaultCinemaNodeServer.OnReceive := OnReceiveDefault;
  FDefaultCinemaNodeServer.OnClose := OnCloseDefault;
  FServer := TCinemaSocketServer.Create('', 0, TSocketType.TCP);
  FServer.OnReceive := OnReceiveNodeClient;
  FServer.OnClose := OnCloseNodeClient;
end;

procedure THypervisor.DeleteScenario(const AKey: WideString; ADeleteFile: boolean);
var
  parentNode: IHypervisorNode;
  cinemaNode: IHypervisorNode;
  scenario: IHypervisorScenario;
  snipents: TArray<UnicodeString>;
  lowCase: WideString;
begin

  lowCase := LowerCase(AKey);
  snipents := UnicodeString(lowCase).Split(['/', '\']);
  if length(snipents) < 2 then
    exit;

  if not FNodesDic.TryGetValue(snipents[0], cinemaNode) then
  begin
    SendEvent('Пришла команда на удаление сценария но в конфигурации не найден родительский узел "' + snipents[0] + '"',  MESSAGE_WARN);
    exit;
  end;

  parentNode := cinemaNode;

  if FNodesDic.TryGetValue(lowCase, cinemaNode) then
  begin
    if Supports(cinemaNode.GetObject, IHypervisorScenario, scenario) then
    begin
      snipents := scenario.FileScenario.Split([TCinemaConfiguration.DtlWorkDir + cinemaNode.Name + PathDelim + TCinemaConfiguration.DtlWorkDirScens]);
      if (length(snipents) = 2) and (snipents[0] = '') then
        DeleteFile(scenario.FileScenario);
    end;
    FNodesDic.Remove(lowCase);
    parentNode.DeleteChild(cinemaNode);
    if Assigned(FOnRemouteDeleteNode) then
      FOnRemouteDeleteNode(cinemaNode);
    SendEvent('Сценарий "' + AKey + '" удалён другим гипервизором "',  MESSAGE_INFO);
  end;
end;

destructor THypervisor.Destroy;
begin
  inherited;
  FSelectedCinemaNode := nil;
  FScenarios.Free;
  FNodesDic.Free;
  FCinemaNodesDic.Free;
  FCinemaNodes.Free;
  FCinemaNodeClientsDic.Free;
  FDefaultCinemaNodeServer.Free;
  FDefaultCinemaNodes.Free;
  FServer.Free;
end;

procedure THypervisor.DoProcessorStat(const AKey, AHostName: UnicodeString; const AStats: TTLVScenarioProcessorStat; const ATables: UnicodeString);
var
  i: int32;
  node: IHypervisorNode;
  processor: IDataStreamProcessor;
  scenarioNode: IScenarioNode;
  plugin: IScenarioPlugin;
  cinema: ICinema;
  tables: TArray<UnicodeString>;
  table: IPluginTable;
begin
  node := FindNode(AHostName + '/' + AKey);
  if Assigned(node) and Assigned(node.ScenarioNode) then
  begin
    scenarioNode := IScenarioNode(node.ScenarioNode.Scenario.GetChildByGuid(AStats.Guid));
    if Assigned(scenarioNode) then
    begin
      processor := nil;
      if scenarioNode.QueryInterface(IDataStreamProcessor, processor) = S_OK then
      begin
        processor.CountInputDataStreams := AStats.CountInputDataStreams;
        processor.CountOutputDataStreams := AStats.CountOutputDataStreams;
        processor.InputSpeed := AStats.InputSpeed;
        processor.OutputSpeed := AStats.OutputSpeed;
      end;

      if (scenarioNode.QueryInterface(IScenarioPlugin, plugin) = S_OK) and (scenarioNode.QueryInterface(ICinema, cinema) = S_OK) then
      begin
        tables := ATables.Split([csPluginTablesDelimeter]);
        if length(tables) <> AStats.CountTables then
        begin
          SendEvent('Ошибка передачи таблицы плагина', MESSAGE_ERROR);
          exit;
        end;

        for i := 0 to AStats.CountTables - 1 do
        begin
          if i < plugin.CountTables then
            table := IPluginTable(plugin.Table[i])
          else
            table := IPluginTable(cinema.CreateTable('', ''));

          table.Text := PWideChar(Trim(tables[i]));
        end;
      end;

      if Assigned(FListener) and Assigned(processor) then
        FListener.OnChangeScenarioProcessorStat(processor);
    end;
  end;
end;

procedure THypervisor.DoProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString);
var
  i: int32;
  node: IHypervisorNode;
  scenarioNode: IScenarioNode;
  processor: IDataStreamProcessor;
  plugin: IScenarioPlugin;
  cinema: ICinema;
  tables: TArray<UnicodeString>;
  table: IPluginTable;
begin
  node := FindNode(AHostName + '/' + AKey);
  if Assigned(node) and Assigned(node.ScenarioNode) then
  begin
    scenarioNode := IScenarioNode(node.ScenarioNode.Scenario.GetChildByGuid(AStats.Guid));
    processor := nil;
    if Assigned(scenarioNode) then
    begin
      if scenarioNode.QueryInterface(IDataStreamProcessor, processor) = S_OK then
      begin
        processor.CountInputDataStreams := AStats.CountInputDataStreams;
        processor.CountOutputDataStreams := AStats.CountOutputDataStreams;
        processor.InputSpeed := AStats.InputSpeed;
        processor.OutputSpeed := AStats.OutputSpeed;
        processor.QueuesSize := AStats.QueuesSize;
        processor.QueuesItemsCount := AStats.QueuesItemsCount;
      end;

      if (scenarioNode.QueryInterface(IScenarioPlugin, plugin) = S_OK) and (scenarioNode.QueryInterface(ICinema, cinema) = S_OK) then
      begin
        tables := ATables.Split([csPluginTablesDelimeter]);
        if length(tables) <> AStats.CountTables then
        begin
          SendEvent('Ошибка передачи таблицы плагина', MESSAGE_ERROR);
          exit;
        end;

        for i := 0 to AStats.CountTables - 1 do
        begin
          if i < plugin.CountTables then
            table := IPluginTable(plugin.Table[i])
          else
            table := IPluginTable(cinema.CreateTable('', ''));

          table.Text := PWideChar(tables[i]);
        end;
      end;

      if Assigned(FListener) and Assigned(processor) then
        FListener.OnChangeScenarioProcessorStat(processor);
    end;

  end;
end;

procedure THypervisor.DoProcessVersion(const AKey, AVersion: UnicodeString);
var
  node: IHypervisorNode;
begin
  node := FindNode(AKey);
  if Assigned(node) and Assigned(node.CinemaNode) then
  begin
    if Assigned(FListener) then
      FListener.OnChangeCinemaNodeStat(node.CinemaNode);
  end;
end;

procedure THypervisor.DoNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte);
var
  node: IHypervisorNode;
begin
  node := FindNode(AKey);
  if Assigned(node) and Assigned(node.CinemaNode) then
  begin
    node.CinemaNode.TotalMemory := ATotalMem;
    node.CinemaNode.UsageMemory := AMemUsage;
    node.CinemaNode.CPUUsage := ACPUUsage;
    if Assigned(FListener) then
      FListener.OnChangeCinemaNodeStat(node.CinemaNode);
  end;
end;

procedure THypervisor.DoScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte);
var
  node: IHypervisorNode;
begin
  node := FindNode(AHostName + '/' + AKey);
  if Assigned(node) and Assigned(node.ScenarioNode) then
  begin
    node.ScenarioNode.TotalMemory := ATotalMem;
    node.ScenarioNode.UsageMemory := AMemUsage;
    node.ScenarioNode.AvailableMemory := AAvailableMem;
    node.ScenarioNode.CPUUsage := ACPUUsage;
    if Assigned(FListener) then
      FListener.OnChangeScenarioStat(node.ScenarioNode);
  end;
end;

procedure THypervisor.DoOnChangeChildHealthState(ASource: THypervisorNode);
begin
  inherited;
  if Assigned(FListener) then
    FListener.OnChangeNodeHealthState(ASource);
end;

procedure THypervisor.DoOnChangeChildLogState(ASource: THypervisorNode);
begin
  inherited;
  if Assigned(FListener) then
    FListener.OnChangeNodeLog(ASource);
end;

procedure THypervisor.DoOnChangeChildState(ASource: THypervisorNode);
begin
  inherited;
  if Assigned(FListener) then
    FListener.OnChangeNodeState(ASource);
end;

procedure THypervisor.FromJson(AJson: TJSONCinema);

  procedure LoadNode(AParentNode: THypervisorNode; AParenJson: TJSONCinema);
  var
    i: Int32;
    children: TJsonArray;
    childJson: TJSONCinema;
    child: THypervisorNode;
  begin
    children := AParenJson.GetValue<TJsonArray>('children');
    for i := 0 to children.Count - 1 do
    begin
      childJson := children.Items[i];
      child := CreateNode(childJson.GetValue<string>('className'), Self, childJson.GetValue<WideString>('name'), AParentNode);
      child.FromJson(childJson);
      FOwner.OnAddChild(child);
      LoadNode(child, childJson);
    end;
  end;

begin
  FIsDeserializingNow := true;
  try
    inherited;
    LoadNode(Self, AJson);
  finally
    FIsDeserializingNow := false;
  end;
end;

function THypervisor.GetCountCinemaNodes: int32;
begin
  Result := FCinemaNodes.Count;
end;

function THypervisor.GetCountConnectedCinemaNodes: int32;
begin
  Result := FCountConnectedCinemaNodes;
end;

function THypervisor.GetEventsListener: THypervisorEventsListener;
begin
  Result := FListener;
end;

function THypervisor.GetIsRun: boolean;
begin
  Result := Assigned(FThread);
end;

function THypervisor.GetLog: ICinemaLog;
begin
  Result := FLog;
end;

function THypervisor.GetOnConnectCinemaNodeEvent: THypervisorNodeEvent;
begin
  Result := FOnConnectCinemaNodeEvent;
end;

function THypervisor.GetOnCreateNode: THypervisorNodeEvent;
begin
  Result := FOnCreateNode;
end;

function THypervisor.GetOnDestroyNode: THypervisorNodeEvent;
begin
  Result := FOnDestroyNode;
end;

function THypervisor.GetOnDisconnectCinemaNodeEvent: THypervisorNodeEvent;
begin
  Result := FOnDisconnectCinemaNodeEvent;
end;

function THypervisor.GetOnRemouteDeleteNode: THypervisorNodeEvent;
begin
  Result := FOnRemouteDeleteNode;
end;

function THypervisor.GetScenario(const AFileName: string): ICinemaScenario;
var
  json: TJSONCinema;
  uid: UnicodeString;
  guid: TGuid;
  resOk: boolean;
begin
  json := Cinema.Utils.GetJson(AFileName, true);
  try
    resOk := json.TryGetValue<UnicodeString>('uid', uid);
    if resOk then
    begin
      guid := StringToGuid(string(uid));
      if not FScenarios.TryGetValue(guid, Result) then
      begin
        Result := CreateScenario(PWideChar(ExtractFileName(AFileName)));
        Result.LoadScenario(PWideChar(AFileName));
        FScenarios.Add((Result as IScenarioNode).GUID^, Result);
      end;
    end;

  finally
    json.Free;
  end;
end;

function THypervisor.FindLikeNode(const ALikeName: WideString; const AFromNode: IHypervisorNode): IHypervisorNode;
var
  upLikeName: WideString;
  fromNode: IHypervisorNode;

  function CheckMatch(const ANode: IHypervisorNode; out ALikeNode: IHypervisorNode): boolean;
  var
    i: Int32;
  begin
    if Assigned(fromNode) then
    begin
      Result := false;
      if (ANode.GetObject = fromNode.GetObject) then
        fromNode := nil;
    end else
    begin
      Result := UpperCase(ANode.Name).Contains(upLikeName);
      if Result then
      begin
        ALikeNode := ANode;
        exit;
      end;
    end;

    for i := 0 to ANode.CountChildren - 1 do
    begin
      Result := CheckMatch(ANode.Child[i], ALikeNode);
      if Result then
        break;
    end;
  end;

  function CheckMatchBefore(const ANode: IHypervisorNode; out ALikeNode: IHypervisorNode): boolean;
  var
    i: Int32;
  begin
    if not Assigned(fromNode) or (ANode.GetObject = fromNode.GetObject) then
    begin
      fromNode := nil;
      exit(false);
    end;
    Result := UpperCase(ANode.Name).Contains(upLikeName);
    if Result then
    begin
      ALikeNode := ANode;
      exit;
    end;

    for i := 0 to ANode.CountChildren - 1 do
    begin
      Result := CheckMatchBefore(ANode.Child[i], ALikeNode);
      if Result then
        break;
    end;
  end;

begin
  if ALikeName = '' then
    exit;
  fromNode := AFromNode;
  if Assigned(fromNode) and (fromNode.GetObject = Self) then
    fromNode := nil;
  upLikeName := UpperCase(ALikeName);
  CheckMatch(Self, Result);
  if not Assigned(Result) and Assigned(AFromNode) then
  begin
    fromNode := AFromNode;
    CheckMatchBefore(Self, Result);
  end;
end;

function THypervisor.GetSelectedCinemaNode: IHypervisorCinemaNode;
begin
  Result := FSelectedCinemaNode;
end;

function THypervisor.OnAcceptDefault(ASocket: TSocketHandle; ARemouteEndPoint: TNetEndpoint): Pointer;
var
  nodeClient: TDefaultCinemaNodeClient;
begin
  nodeClient := TDefaultCinemaNodeClient.Create(Self, ARemouteEndPoint);
  FDefaultCinemaNodes.Add(nodeClient);
  Result := nodeClient;
end;

function THypervisor.GetIsDeserializingNow: Boolean;
begin
  Result := FIsDeserializingNow;
end;

procedure THypervisor.OnReceiveNodeClient(AOwner: Pointer; AData: PByte; ALen: Integer);
begin
  TCinemaNodeClient(AOwner).ProcessTlv(AData, ALen);
end;

procedure THypervisor.OnCloseNodeClient(AOwner: Pointer);
begin
  TCinemaNodeClient(AOwner).Client.IsConnected := false;
end;

procedure THypervisor.OnAcceptDefaultCinemaNode(ACinemNode: TDefaultCinemaNodeClient);
var
  pair: TPair<string, TCinemaNodeClient>;
begin
  if FCinemaNodesDic.ContainsKey(ACinemNode.HostName) or FCinemaNodeClientsDic.ContainsKey(string(ACinemNode.HostName)) then
    exit;

  for pair in FCinemaNodeClientsDic do
  begin
    if (pair.Value.HostName = ACinemNode.HostName) then
      exit;

    if (pair.Value.Client.EndPoint.Address.Address = ACinemNode.RemouteEndPoint.Address.Address) then
      exit;

    if (pair.Value.IsLocal) and (ACinemNode.HostName = THypervisor.LocalHostName) then
      exit;

  end;

  AddCinemaNode(ACinemNode.HostName, ACinemNode.RemouteEndPoint.Address.Address);
end;

procedure THypervisor.OnAddChild(const AChild: THypervisorNode);
begin
  TMonitor.Enter(FNodesDic);
  try
    FNodesDic.Add(LowerCase(AChild.Key), AChild);
  finally
    TMonitor.Exit(FNodesDic);
  end;

  if AChild.InheritsFrom(THypervisorCinemaNode) then
    OnCreateCinemaNode(THypervisorCinemaNode(AChild));

  if Assigned(FOnCreateNode) then
    FOnCreateNode(AChild);
end;

procedure THypervisor.OnCloseDefault(AOwner: Pointer);
var
  nodeClient: TDefaultCinemaNodeClient;
begin
  nodeClient := AOwner;
  FDefaultCinemaNodes.Remove(nodeClient);
end;

procedure THypervisor.OnConnectCinemaNodeClient(AClient: TCinemaNodeClient);

  procedure ResetChild(const ANode: IHypervisorNode);
  var
    i: int32;
  begin
    if Assigned(ANode.ScenarioNode) then
      ANode.ScenarioNode.ResetToState(TNodeState.nsNone);

    for i := 0 to ANode.CountChildren - 1 do
    begin
      ResetChild(ANode.Child[i]);
    end;
  end;

var
  i: int32;
  cinemaNode: THypervisorCinemaNode;
begin
  inc(FCountConnectedCinemaNodes);
  FServer.Add(AClient.Client.Handle, AClient, false);
  if Assigned(FOnConnectCinemaNodeEvent) then
  begin
    for i := 0 to AClient.CountOwners - 1 do
    begin
      cinemaNode := AClient.Owners[i];
      ResetChild(cinemaNode);
      FOnConnectCinemaNodeEvent(cinemaNode);
    end;
  end;
end;

procedure THypervisor.OnCreateCinemaNode(ACinemaNode: THypervisorCinemaNode);
var
  list: TList<TProcessInfo>;
  client: TCinemaNodeClient;
begin
  TMonitor.Enter(Self);
  try

    TMonitor.Enter(FCinemaNodeClientsDic);
    try
      FCinemaNodes.Add(ACinemaNode);
      FCinemaNodesDic.Add(ACinemaNode.Name, ACinemaNode);
      if not FCinemaNodeClientsDic.TryGetValue(ACinemaNode.Address, client) then
      begin
        client := TCinemaNodeClient.Create(Self, ACinemaNode.Name, ACinemaNode.Address);
        FCinemaNodeClientsDic.Add(ACinemaNode.Address, client);
      end;
    finally
      TMonitor.Exit(FCinemaNodeClientsDic);
    end;

    client.AddOwner(ACinemaNode);

    if ACinemaNode.Client.IsLocal then
    begin
      // проверим что локально процесс ноды запущен
      list := GetListProcesses(WideToString(csCineamNodeName));
      try
        if list.Count = 0 then
          FileExecute(WideToString(TCinemaConfiguration.RtlAppPath + csCineamNodeName));
      finally
        list.Free;
      end;
    end;

  finally
    TMonitor.Exit(Self);
  end;
end;

procedure THypervisor.OnDestroyCinemaNode(ACinemaNode: THypervisorCinemaNode);
var
  client: TCinemaNodeClient;
begin
  if Assigned(FSelectedCinemaNode) and (FSelectedCinemaNode.GetObject = ACinemaNode) then
    FSelectedCinemaNode := nil;

  TMonitor.Enter(Self);
  try
    FCinemaNodes.Remove(ACinemaNode);
    FCinemaNodesDic.Remove(ACinemaNode.Name);
  finally
    TMonitor.Exit(Self);
  end;

  TMonitor.Enter(FCinemaNodeClientsDic);
  try
    client := FCinemaNodeClientsDic.Items[ACinemaNode.Address];
    client.RemoveOwner(ACinemaNode);
    if client.CountOwners = 0 then
    begin
      FCinemaNodeClientsDic.Remove(ACinemaNode.Address);
      client.Free;
    end;
  finally
    TMonitor.Exit(FCinemaNodeClientsDic);
  end;
end;

procedure THypervisor.OnDisconnectCinemaNodeClient(AClient: TCinemaNodeClient);
var
  i: int32;
  cinemaNode: THypervisorCinemaNode;
begin
  dec(FCountConnectedCinemaNodes);
  FServer.Remove(AClient.Client.Handle);
  if Assigned(FOnDisconnectCinemaNodeEvent) then
  begin
    for i := 0 to AClient.CountOwners - 1 do
    begin
      cinemaNode := AClient.Owners[i];
      FOnDisconnectCinemaNodeEvent(cinemaNode);
    end;
  end;
end;

procedure THypervisor.SendEvent(const AKey, ASource, AWhoIs, AMsg: WideString; ADateTime: TDateTime; AEventType: int32; ASourceId: TGuid);
var
  child: IHypervisorNode;
  node: IScenarioNode;
  path: WideString;
begin
  child := nil;
  path := AKey + '/' + ASource;
  if not FNodesDic.TryGetValue(path, child) then
  begin
    path := ASource;
    if not FNodesDic.TryGetValue(path, child) then
    begin
      path := AKey;
      FNodesDic.TryGetValue(path, child);
    end;
    if not Assigned(child) then
      child := FindNode(AKey + '/' + ASource);
  end;
  FLog.AddMessage(PWideChar(AMsg), PWideChar(path), ASourceId, AEventType);
  if Assigned(child) then
  begin
    if Assigned(child.ScenarioNode) and not ASourceId.IsEmpty then
    begin
      node := IScenarioNode(child.ScenarioNode.Scenario.GetChildByGuid(ASourceId));
      if Assigned(node) then
        node.SendNodeMessage(PWideChar(AMsg), AEventType);
    end;
    child.SetLogState(AMsg, TNodeLogState(AEventType));
  end else
    SetLogState(AMsg, TNodeLogState(AEventType));
end;

procedure THypervisor.SendEvent(const AMsg: WideString; AEventType: int32);
begin
  SendEvent(Key, Key, Name, AMsg, now, AEventType, TGuid.Empty);
end;

procedure THypervisor.SendHealthStateNodeEvent(const AKey: WideString; AState: byte; AHealthState: byte; const AHealthStateMsg: WideString; const AItemID: TGuid
  );
var
  child: IHypervisorNode;
  node: IScenarioNode;
begin
  child := FindNode(AKey);
  if Assigned(child) then
  begin
    if Assigned(child.ScenarioNode) then
    begin
      if not AItemID.IsEmpty then
      begin
        node := IScenarioNode(child.ScenarioNode.Scenario.GetChildByGuid(AItemID));
        if Assigned(node) then
        begin
          if node.State <> TNodeState(AState) then
            node.State := TNodeState(AState);
          node.ChangeHealthState(PWideChar(AHealthStateMsg), AHealthState);
        end;
      end;
      child.SetHealthState((child.ScenarioNode.Scenario as IScenarioNode).HealthStateMessage, (child.ScenarioNode.Scenario as IScenarioNode).HealthState);
    end else
      child.SetHealthState(AHealthStateMsg, AHealthState);
  end else
    SetHealthState(AHealthStateMsg, AHealthState);
end;

procedure THypervisor.SetEventsListener(const Value: THypervisorEventsListener);
begin
  FListener := Value;
end;

procedure THypervisor.SetOnConnectCinemaNodeEvent(AValue: THypervisorNodeEvent);
begin
  FOnConnectCinemaNodeEvent := AValue;
end;

procedure THypervisor.SetOnCreateNode(const Value: THypervisorNodeEvent);
begin
  FOnCreateNode := Value;
end;

procedure THypervisor.SetOnDestroyNode(const Value: THypervisorNodeEvent);
begin
  FOnDestroyNode := Value;
end;

procedure THypervisor.SetOnDisconnectCinemaNodeEvent(AValue: THypervisorNodeEvent);
begin
  FOnDisconnectCinemaNodeEvent := AValue;
end;

procedure THypervisor.SetOnRemouteDeleteNode(const Value: THypervisorNodeEvent);
begin
  FOnRemouteDeleteNode := Value;
end;

procedure THypervisor.SetSelectedCinemaNode(const Value: IHypervisorCinemaNode);
begin
  FSelectedCinemaNode := Value;
end;

function THypervisor.FindNode(const AKey: WideString): IHypervisorNode;
var
  i: int32;
  splitted: TArray<WideString>;
  key: WideString;
  lowKey: Widestring;
begin
  lowKey := LowerCase(AKey);
  if not FNodesDic.TryGetValue(lowKey, Result) then
  begin
    splitted := lowKey.Split(['/']);
    if length(splitted) > 1 then
    begin
      if LowerCase(splitted[0]) = LowerCase(THypervisor.LocalHostName) then
      begin
        key := 'localhost';
        for i := 1 to length(splitted) - 1 do
          key := key + '/' + splitted[i];
        FNodesDic.TryGetValue(key, Result);
      end;
    end;
  end;
end;

procedure THypervisor.Start;
begin
  if Assigned(FThread) then
    exit;
  SendEvent('Гипервизор запущен', MESSAGE_INFO);
  FThread := TWorkThread.Create(self);
end;

procedure THypervisor.Stop;
begin
  if not Assigned(FThread) then
    exit;
  FThread.Terminate;
  FThread.WaitFor;
  FreeAndNil(FThread);
end;

procedure THypervisor.StopScenarioEvent(const AKey: WideString);
var
  child: IHypervisorNode;
begin
  child := nil;
  TMonitor.Enter(FNodesDic);
  try
    child := FindNode(AKey);
    if Assigned(child) then
    begin
      if Assigned(child.ScenarioNode) then
      begin
        child.ScenarioNode.ResetToState(TNodeState.nsNone);
      end;
    end;
  finally
    TMonitor.Exit(FNodesDic);
  end;
end;

procedure THypervisor.OnReceiveDefault(AOwner: Pointer; AData: PByte; ALen: Integer);
begin
  TDefaultCinemaNodeClient(AOwner).ProcessTlv(AData, ALen);
end;

procedure THypervisor.OnRemoveChild(const AChild: THypervisorNode);
begin
  TMonitor.Enter(FNodesDic);
  try
    FNodesDic.Remove(LowerCase(AChild.Key));
  finally
    TMonitor.Exit(FNodesDic);
  end;
  if Assigned(FOnDestroyNode) then
    FOnDestroyNode(AChild);
end;

function THypervisor.ToJson: WideString;
begin
  Result := '{' + sLineBreak +
    inherited
    + '}'+ sLineBreak;
end;

{ THypervisorCinemaNode }

constructor THypervisorCinemaNode.Create(AOwner: THypervisor; const AName: WideString; const AParent: IHypervisorNode);
begin
  inherited Create(AOwner, AName, AParent);
  FVersion := '-';
end;

destructor THypervisorCinemaNode.Destroy;
begin
  inherited;
  FOwner.OnDestroyCinemaNode(Self);
end;

procedure THypervisorCinemaNode.FromJson(AJson: TJSONCinema);
begin
  inherited;
  AJson.TryGetValue<string>('Address', FAddress);
end;

function THypervisorCinemaNode.GetAddress: string;
begin
  if FAddress.IsEmpty then
    Result := WideToString(Name)
  else
    Result := FAddress;
end;

function THypervisorCinemaNode.GetUsageMemory: int64;
begin
  Result := FUsageMemory;
end;

function THypervisorCinemaNode.GetVersion: WideString;
begin
  Result := FVersion;
end;

function THypervisorCinemaNode.GetCPUUsage: byte;
begin
  Result := FCPUUsage;
end;

function THypervisorCinemaNode.GetIsConnected: boolean;
begin
  Result := Assigned(FClient) and FClient.Client.IsConnected;
end;

function THypervisorCinemaNode.GetOS: WideString;
begin
  Result := FClient.OS;
end;

function THypervisorCinemaNode.GetTotalMemory: int64;
begin
  Result := FTotalMemory;
end;

procedure THypervisorCinemaNode.SendFile(const AFileName: string);
var
  src: TArray<WideString>;
  src2: TArray<WideString>;
begin
  if not IsConnected then
    exit;

  TMonitor.Enter(FClient);
  try
    src := WideString(AFileName).Split(WideString(TCinemaConfiguration.RtlPath));
    if length(src) = 2 then
    begin
      src2 := src[1].Split(WideString(FClient.OS));
      if (length(src2) = 2) and (src2[0] = '') then
        FClient.PutTlvFile(StringToWide(AFileName), src2[1], csFile)
      else
        FClient.PutTlvFile(StringToWide(AFileName), src[1], csFile);
    end else
      FClient.PutTlvFile(StringToWide(AFileName), StringToWide(ExtractFileName(AFileName)), csFile);
  finally
    TMonitor.Exit(FClient);
  end;
end;

procedure THypervisorCinemaNode.SendRequestDeleteScenario(const AKey: WideString);
begin
  if not AKey.IsEmpty then
  begin
    TMonitor.Enter(FClient);
    try
      FClient.PutTlvDeleteScenario(TrancParentKey(AKey), true);
    finally
      TMonitor.Exit(FClient);
    end;
  end else
  if Assigned(ScenarioNode) then
    SendRequestKill(Key)
  else
    inherited;
end;

procedure THypervisorCinemaNode.SendRequestKill(const AKey: WideString);
begin
  if not IsConnected then
    exit;

  if not AKey.IsEmpty then
  begin
    TMonitor.Enter(FClient);
    try
      FClient.PutTlvKillScenario(TrancParentKey(AKey))
    finally
      TMonitor.Exit(FClient);
    end;
  end else
  if Assigned(ScenarioNode) then
    SendRequestKill(Key)
  else
    inherited;
end;

procedure THypervisorCinemaNode.SendRequestRun(const AKey, AName: WideString; ACountInstances: int32);
begin
  if not IsConnected then
    exit;
  if not AKey.IsEmpty and not AName.IsEmpty then
  begin
    TMonitor.Enter(FClient);
    try
      //FClient.PutTlvAddOrUpdateScenario(TrancParentKey(AKey), TCinemaNetProtocol.csWorkDirectory + TCinemaNetProtocol.csWorkSubDirScens + ExtractFileName(AName), ACountInstances);
      FClient.PutTlvRunScenario(TrancParentKey(AKey), AName, ACountInstances);
    finally
      TMonitor.Exit(FClient);
    end;
  end else
  if Assigned(ScenarioNode) then
    SendRequestRun(Key, ScenarioNode.FileScenario, ScenarioNode.CountInstances)
  else
    inherited;
end;

procedure THypervisorCinemaNode.SendRequestStat;
begin
  if not IsConnected then
    exit;
  TMonitor.Enter(FClient);
  try
    FClient.PutTlvGetNodeStat(Key);
  finally
    TMonitor.Exit(FClient);
  end;
end;

procedure THypervisorCinemaNode.SendRequestStatTables(const AKey: WideString; const AGuid: TGuid);
begin
  if not IsConnected then
    exit;

  if not AKey.IsEmpty then
  begin
    TMonitor.Enter(FClient);
    try
      FClient.PutTlvGetStatTables(TrancParentKey(AKey), THypervisor.LocalHostName, AGuid);
    finally
      TMonitor.Exit(FClient);
    end;
  end else
  if Assigned(ScenarioNode) then
    SendRequestStatTables(Key, AGuid)
  else
    inherited;
end;

procedure THypervisorCinemaNode.SendRequestStop(const AKey: WideString);
begin
  if not IsConnected then
    exit;

  if not AKey.IsEmpty then
  begin
    TMonitor.Enter(FClient);
    try
      FClient.PutTlvStopScenario(TrancParentKey(AKey));
    finally
      TMonitor.Exit(FClient);
    end;
  end else
  if Assigned(ScenarioNode) then
    SendRequestStop(Key)
  else
    inherited;
end;

procedure THypervisorCinemaNode.SendRequestTerminate(const AKey: WideString);
begin
  if not IsConnected then
    exit;

  if not AKey.IsEmpty then
  begin
    TMonitor.Enter(FClient);
    try
      FClient.PutTlvTerminateScenario(TrancParentKey(AKey))
    finally
      TMonitor.Exit(FClient);
    end;
  end else
  if Assigned(ScenarioNode) then
    SendRequestTerminate(Key)
  else
    inherited;
end;

procedure THypervisorCinemaNode.SendUpdatedScenario(const AKey, AName: WideString; ACountInstances: int32);
begin
  if not IsConnected then
    exit;

  if not AKey.IsEmpty and not AName.IsEmpty then
  begin
    TMonitor.Enter(FClient);
    try
      FClient.PutTlvFile(AName, TCinemaNetProtocol.csWorkDirectory + TCinemaNetProtocol.csWorkSubDirScens + ExtractFileName(AName), csFileTypeScenario);
      FClient.PutTlvAddOrUpdateScenario(TrancParentKey(AKey), TCinemaNetProtocol.csWorkDirectory + TCinemaNetProtocol.csWorkSubDirScens + ExtractFileName(AName), ACountInstances)
    finally
      TMonitor.Exit(FClient);
    end;
  end;
end;

procedure THypervisorCinemaNode.SetAddress(const Value: string);
begin
  FAddress := Value;
end;

procedure THypervisorCinemaNode.SetUsageMemory(const Value: int64);
begin
  FUsageMemory := Value;
end;

procedure THypervisorCinemaNode.SetVersion(const Value: WideString);
begin
  FVersion := Value;
end;

procedure THypervisorCinemaNode.SetCPUUsage(const Value: byte);
begin
  FCPUUsage := Value;
end;

procedure THypervisorCinemaNode.SetTotalMemory(const Value: int64);
begin
  FTotalMemory := Value;
end;

function THypervisorCinemaNode.ToJson: WideString;
var
  jsnString: TJSONString;
{$ifndef fpc}
  builder: TStringBuilder;
begin
  jsnString := TJSONString.Create(FAddress);
  builder := TStringBuilder.Create;
  try
    jsnString.ToChars(builder, []);
    Result :=
      '"Address":' + builder.ToString + ',' + sLineBreak +
      inherited;
  finally
    jsnString.Free;
    builder.Free;
  end;
end;
{$else}
begin
  jsnString := TJSONString.Create(FAddress);
  try
    Result :=
      '"Address":' + WideString(jsnString.ToJSON) + ',' + sLineBreak +
      inherited;
  finally
    jsnString.Free;
  end;
end;
{$endif}

procedure THypervisorCinemaNode.UpdateItself;
var
  src: WideString;
begin
  if not IsConnected then
    exit;
  TMonitor.Enter(FClient);
  try
    src := TCinemaConfiguration.RtlPath + FClient.OS + PathDelim + 'CinemaNode';
    if LowerCase(FClient.OS).Contains('win') then
      src := src + '.exe';
    FClient.PutTlvCinemaNode(src, 'CinemaNode.copy');
  finally
    TMonitor.Exit(FClient);
  end;
end;

{ TCinemaNodeClient }

procedure TCinemaNodeClient.AddOwner(AOwner: THypervisorCinemaNode);
begin
  AOwner.Client := Self;
  FOwners.Add(AOwner);
end;

constructor TCinemaNodeClient.Create(AHypervisor: THypervisor; const AName: UnicodeString; const AAddress: string);
begin
  inherited Create;
  FName := AName;
  FIsLocal := (LowerCase(FName) = 'localhost') or (AAddress = '127.0.0.1') or (FName = THypervisor.LocalHostName);
  FHypervisor := AHypervisor;
  FOwners := TList<THypervisorCinemaNode>.Create;
  FClient := TCinemaSocketClient.Create(TNetEndpoint.Create(TIpAddress.Create(IfThen(FIsLocal, '127.0.0.1', AAddress)), TCinemaNetProtocol.csServerForHypervisorPort), TSocketType.TCP);
  FClient.OnConnect := OnClientConnect;
  FClient.OnDisconnect := OnClientDisconnect;
  FClient.TimeoutOnConnect := 300;
  OnProcessError := OnProtocolError;
end;

destructor TCinemaNodeClient.Destroy;
begin
  FClient.Free;
  FOwners.Free;
  inherited;
end;

procedure TCinemaNodeClient.DoAddOrUpdateScenario(const AKey, AName: UnicodeString; ACountInstances: int32);
begin
  inherited;
  FHypervisor.AddOrUpdateScenario(FOwners.Items[0].Key + '/' + AKey, AName, ACountInstances);
end;

procedure TCinemaNodeClient.DoDeleteScenario(const AKey: UnicodeString; ADeleteFile: boolean);
begin
  inherited;
  FHypervisor.DeleteScenario(AKey, ADeleteFile);
end;

procedure TCinemaNodeClient.DoEvent(const ASource, AWhoIs, AMsg: UnicodeString; ADateTime: TDateTime; AEventType: int32; const ASourceId: TGuid);
begin
  inherited;
  FHypervisor.SendEvent(FOwners.Items[0].Key, ASource, AWhoIs, AMsg, ADateTime, AEventType, ASourceId);
end;

procedure TCinemaNodeClient.DoHealthStateNode(const AKey: UnicodeString; AState: byte; AHealthState: byte; const AHealthStateMsg: UnicodeString;
  const AItemID: TGuid);
begin
  inherited;
  FHypervisor.SendHealthStateNodeEvent(FOwners.Items[0].Key + '/' + AKey, AState, AHealthState, AHealthStateMsg, AItemID);
end;

procedure TCinemaNodeClient.DoHostName(const AName: UnicodeString);
begin
  inherited;
  FHostName := AName;
end;

procedure TCinemaNodeClient.DoKey(const AKey: UnicodeString; APid: uint32; ANumber: int32);
begin
  inherited;
end;

procedure TCinemaNodeClient.DoNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte);
begin
  inherited;
  FHypervisor.DoNodeStat(AKey, ATotalMem, AMemUsage, ACPUUsage);
end;

procedure TCinemaNodeClient.DoOs(const AOS: UnicodeString);
begin
  inherited;
  FOS := AOS;
end;

procedure TCinemaNodeClient.DoScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte);
begin
  inherited;
  FHypervisor.DoScenarioStat(AKey, AHostName, ATotalMem, AAvailableMem, AMemUsage, ACPUUsage);
end;

procedure TCinemaNodeClient.DoState(const ASource: TGuid; AState: int32);
begin
  inherited;
  FHypervisor.ChangeStateEvent(Key, ASource, TNodeState(AState));
end;

procedure TCinemaNodeClient.DoProcessorStat(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat; const ATables: UnicodeString);
begin
  FHypervisor.DoProcessorStat(AKey, AHostName, AStats, ATables);
end;

procedure TCinemaNodeClient.DoProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString);
begin
  inherited;
  FHypervisor.DoProcessorStat_v2(AKey, AHostName, AStats, ATables);
end;

procedure TCinemaNodeClient.DoStopScenario(const AName: UnicodeString);
begin
  inherited;
  FHypervisor.StopScenarioEvent(Key + '/' + AName);
end;

procedure TCinemaNodeClient.DoVersion(const AKey: UnicodeString; const AVersion: UnicodeString);
var
  owner: THypervisorCinemaNode;
begin
  inherited;
  for owner in FOwners do
  begin
    owner.Version := AVersion;
  end;
  FHypervisor.DoProcessVersion(AKey, AVersion);
end;

function TCinemaNodeClient.Execute: boolean;
begin
  Result := false;
  if FClient.IsConnected then
  begin
    if Buffer.Size > 0 then
    begin
      TMonitor.Enter(Self);
      try
        if FClient.SendData(Buffer.Memory, Buffer.Size) then
            Result := true;
        Buffer.Size := 0;
      finally
        TMonitor.Exit(Self);
      end;
    //end else
    //if TThread.GetTickCount64 - FLastSendData > 3000 then
    //begin
    //  FLastSendData := TThread.GetTickCount64;
    //  TMonitor.Enter(Self);
    //  try
    //    PutTlvIsAlive(Name);
    //    if FClient.SendData(Buffer.Memory, Buffer.Size) then
    //        Result := true;
    //  finally
    //    TMonitor.Exit(Self);
    //  end;
    end;
  end else
  if TThread.GetTickCount64 - FLastTryConnected > 3000 then
  begin
    FLastTryConnected := TThread.GetTickCount64;
    FClient.Connect;
  end;
end;

function TCinemaNodeClient.GetCountOwners: int32;
begin
  Result := FOwners.Count;
end;

function TCinemaNodeClient.GetOwners(AIndex: int32): THypervisorCinemaNode;
begin
  Result := FOwners.Items[AIndex];
end;

procedure TCinemaNodeClient.OnClientConnect(AClient: TCinemaSocketClient);
begin
  FLastSendData := TThread.GetTickCount64;
  PutTlvHostName(THypervisor.LocalHostName);
  FHypervisor.OnConnectCinemaNodeClient(Self);
end;

procedure TCinemaNodeClient.OnClientDisconnect(AClient: TCinemaSocketClient);
begin
  FHypervisor.OnDisconnectCinemaNodeClient(Self);
end;

procedure TCinemaNodeClient.OnProtocolError(ATlv: PTlvHeaderBase; const AMsg: UnicodeString);
begin
  FHypervisor.SendEvent(AMsg, MESSAGE_ERROR);
end;

procedure TCinemaNodeClient.RemoveOwner(AOwner: THypervisorCinemaNode);
begin
  FOwners.Remove(AOwner);
end;

{ TDefaultCinemaNodeClient }

constructor TDefaultCinemaNodeClient.Create(AOwner: THypervisor; ARemouteEndPoint: TNetEndpoint);
begin
  inherited Create;
  FOwner := AOwner;
  FRemouteEndPoint := ARemouteEndPoint;
end;

procedure TDefaultCinemaNodeClient.DoHostName(const AName: UnicodeString);
begin
  inherited;
  if FHostName.IsEmpty then
  begin
    FHostName := AName;
    FOwner.OnAcceptDefaultCinemaNode(Self);
  end;

end;

{ THypervisor.TWorkThread }

constructor THypervisor.TWorkThread.Create(AOwner: THypervisor);
begin
  FOwner := AOwner;
  inherited Create(false, 4096*4096);
end;

procedure THypervisor.TWorkThread.Execute;
begin
  NameThreadForDebugging(FOwner.Name, ThreadID);
  repeat
    if not FOwner.Ask then
      Sleep(10);
  until Terminated;
end;

initialization

  TThreading.PoolSize := 32;

finalization

end.
