﻿unit Cinema.Interfaces;

interface

uses
  Cinema.Types
  ;

type

  /// <summary> описание входного/выходного канала плагина, иными словами именованный канал (необходимо для дизайн-тайм систем) </summary>

  TNamedConnection = record
    Name: PWideChar;
    Number: Integer;
  end;

  TNamedOutputConnection = TNamedConnection;
  PNamedOutputConnection = ^TNamedOutputConnection;

  TNamedInputConnection = TNamedConnection;
  PNamedInputConnection = ^TNamedInputConnection;

  /// <summary> краткая информация о плагине </summary>

  PPluginDescription = ^TPluginDescription;
  TPluginDescription = record
    // имя
    Name: PWideChar;
    // автор
    Autor: PWideChar;
    // дата создания (dd.MM.yyyy)
    Created: PWideChar;
    // версия
    Version: PWideChar;
    // краткое описание
    Description: PWideChar;
    // автор последних модификаций
    AutorLastModified: PWideChar;
    // количество именованных входных соединений; 0 - кол-во не определено, может быть сколь угодно
    CountNamedInputConnections: Integer;
    // именованные входные соединения плагина
    NamedInputConnection: PNamedInputConnection;
    // колличество именнованных выходных соединений; -1 - кол-во не определено, может быть сколь угодно
    CountNamedConnections: Integer;
    // именованные выходные соединения плагина
    NamedOutputConnection: PNamedOutputConnection;
    // схема конфигурации; описывается если модуль может иметь простые настройки;
    // если nil - либо плагин не настраиваемый либо разработчик предусмотрел
    // собственный конфигуратор в виде "exe".
    // TODO: спецификации для полей ввода: текст, число, логическое, и т.д.
    SchemeConfig: PWideChar;
  end;

  TPluginDescriptions = array of TPluginDescription;

  /// <summary> поток данных; тип передаваемых данных может быть любым </summary>

  TDataStream = Pointer;

  /// <summary> атрибут канала; колл-во не ограничено, складываются в односвязный список (через Next) </summary>

  PDataStreamAttribute = ^TDataStreamAttribute;
  TDataStreamAttribute = record
    // значение атрибута; тип данных может быть любым, зависит от его ID
    Attribute: PAnsiChar;
    // размер данных в Attribute
    Size: Integer;
    // идентификатор атрибута; все типы будут описаны в справочнике атрибутов
    ID: Integer;
    // следующий атрибут
    Next: PDataStreamAttribute;
    // принадлежность атрибута
    DataStreamOwner: TDataStream;
  end;

  IValueHolder = interface
  ['{C081EBF9-0F68-43AF-8925-ACEAC0B1D9B3}']
    function GetValue: PByte; stdcall;
    function GetSize: Integer; stdcall;
    property Value: PByte read GetValue;
    property Size: Integer read GetSize;
  end;

  /// <summary> Таблица вывода статистики работы плагина </summary>

  IPluginTable = interface
  ['{833B28D0-7D2F-4041-BA33-CCCABFF7901A}']
    function GetName: PWideChar; stdcall;
    function GetCaption: PWideChar; stdcall;
    function AddColumn(AName, ACaption: PWideChar; AType: TCinemaFieldType): Integer; stdcall;
    function AddRow: Integer; stdcall;
    function GetCountColumn: Integer; stdcall;
    function GetCountRows: Integer; stdcall;
    procedure SetCell(AColumn: Integer; ARow: Integer; AValue: PWideChar); stdcall;
    function GetCell(AColumn: Integer; ARow: Integer): PWideChar; stdcall;
    function GetText: PWideChar; stdcall;
    procedure SetText(AText: PWideChar); stdcall;
    function GetJson: PWideChar; stdcall;
    function GetColumnCaption(AIndex: Integer): PWideChar; stdcall;
    procedure SetNeedReload(AValue: Boolean); stdcall;

    property Name: PWideChar read GetName;
    property Caption: PWideChar read GetCaption;
    property CountColumn: Integer read GetCountColumn;
    property CountRows: Integer read GetCountRows;
    property Text: PWideChar read GetText write SetText;
    property Json: PWideChar read GetJson;
    /// <summary> если плагин выставит в true тогда ядро вызовет RedrawView у плагина при первой возможности где плагин
    /// может обновить свои графические данные, в частности и таблицы;
    /// использовать по мере острой необходимости, т.к. при запросе таблицы от GUI эта опция выставляется в автомате <summary>
    property NeedReload: Boolean write SetNeedReload;
  end;

  /// <summary> интерфейс ядра; используется плагином для отправки результата выполнения работы </summary>

  ICinema = interface
  ['{4F6E8DD5-FB09-4831-9188-FEB505A3A98D}']
    ///  <summary> метод для журналирования событий от плагина </summary>
    procedure SendMessage(AMsg: PWideChar; AMessageType: TMessageType); stdcall;
    ///  <summary> Установка текущего состояния качества выполняемой работы плагином </summary>
    ///  TODO: логирование изменений статуса
    procedure SetHealthState(AMsg: PWideChar; AHealthDegree: Byte); stdcall;
    ///  <summary> создать таблицу для вывода статистики работы плагина;
    ///    <return> возвращает эксземпляр с интерфейсом IPluginTable; тип определён как Pointer
    ///    для кросскомпиляторности (Delphi и C++), поэтому необходимо самостоятельное приведение к IPluginTable </return>
    ///  </summary>
    function CreateTable(AName: PWideChar; ACaption: PWideChar): Pointer; stdcall;
    ///  <summary> метод создания потока для отправки данных плагином </summary>
    ///  <param Name="AConnectionNumber"> Номер соединения на котором следующий плагин в сценарии
    ///  будет получать данные потока </param>
    function CreateStream(AConnectionNumber: Integer; AParentDataStream: TDataStream): TDataStream; stdcall;
    ///  <summary> отправить данные по ранее созданному потоку </summary>
    procedure SendData(ADataStream: TDataStream; AData: PByte; ASize: Int64); stdcall;
    ///  <summary> метод уничтожения потока </summary>
    procedure DestroyStream(ADataStream: TDataStream); stdcall;
    ///  <summary> добавить атрибут потока </summary>
    function AddDataStreamAttribute(ADataStream: TDataStream; AID: Integer; AData: Pointer; ADataSize: Integer): PDataStreamAttribute; stdcall;
    ///  <summary> найти атрибут по идентификатору </summary>
    function FindAttribute(ADataStream: TDataStream; AID: Integer): PDataStreamAttribute; stdcall;
    ///  <summary> найти атрибут по идентификатору </summary>
    function FindAttributeValuePtr(ADataStream: TDataStream; AID: Integer; ADefaultValuePtr: Pointer): Pointer; stdcall;
    ///  <summary> найти атрибут типа NativeInt по идентификатору </summary>
    function FindAttributeNativeInt(ADataStream: TDataStream; AID: Integer; ADefault: NativeInt): NativeInt; stdcall;
    ///  <summary> упаковка всех атрибутов потока в буфер </summary>
    function PackAttributes(ADataStreamSource: TDataStream; AFilters: PInteger; AHighIdInFilters: Int32; ALength: PInteger): PByte; stdcall;
    ///  <summary> распаковка всех атрибутов из буфера в поток </summary>
    procedure UnPackAttributes(ADataStreamDest: TDataStream; AAttributes: PByte; ALength: UInt32); stdcall;
    ///  <summary> получить ссылку на все атрибуты </summary>
    function GetAttributes(ADataStream: TDataStream): PDataStreamAttribute; stdcall;
    ///  <summary> выставить атрибуты из списка </summary>
    procedure SetAttributes(ADataStream: TDataStream; AAtributes: PDataStreamAttribute); stdcall;
    ///  <summary> если плагин может выполнять работу без длительного возврата захваченного потока слоя,
    ///  тогда разработчик по возможности должен предусмотреть прерывание текущей работы плагина для того
    ///  что бы при желании пользователя остановить сценарий, можно было корректно завершть его выполнение;
    ///  данный метод позволяет узнать плагину что поток был прерван и необходимо завершить работу </summary>
    function IsThreadTerminated: boolean; stdcall;
    ///  <summary> имя и путь выполняемого сценария </summary>
    function GetScenarioName: PWideChar; stdcall;
    ///  <summary> поучить путь до модуля плагина </summary>
    function GetPluginPath: PWideChar; stdcall;
    ///  <summary> поучить путь до приложения в котором выполняется сценарий </summary>
    function GetAppPath: PWideChar; stdcall;
    ///  <summary> TODO </summary>
    procedure SetForceAsStart(AValue: boolean); stdcall;
    /// <summary> Получить полный путь до каталога со сценариями </summary>
    function GetRtlScenariosPath: PWideChar; stdcall;
    /// <summary> Добавить глобальный атрибут, заменяет на новое значение если такой уже есть </summary>
    procedure AddGlobalAttribute(AID: Integer; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalAttribute(AID: Integer): IValueHolder; stdcall;
    procedure RemoveGlobalAttribute(AID: Integer); stdcall;
    procedure AddGlobalNamedAttribute(AName: PWideChar; AData: Pointer; ADataSize: Integer); stdcall;
    function FindGloabalNamedAttribute(AName: PWideChar): IValueHolder; stdcall;
    procedure RemoveGlobalNamedAttribute(AName: PWideChar); stdcall;
    procedure ClearGlobalAttributes; stdcall;
    function GetVersionScenario: PWideChar; stdcall;
  end;

  /// <summary> интерфейс плагина, реализуется модулем; ядро может создать несколько экземпляров,
  ///  в зависимости от необходимости при выполнении сценария </summary>

  ICinemaPlugin = interface
  ['{54475BE8-128D-4161-B58B-4AC32E9617FC}']
    /// <summary> создать обработчик данных передаваемых во входном потоке ADataStream;
    /// возвращаемый результат может быть любым, на усмотрение разработчика плагина;
    /// как правило, это класс который будет выполнять обработку
    /// входных данных получаемых в методе ProcessDataStream </summary>
    function CreateDataStreamProcessor(AInputDataStream: TDataStream; ANumberInput: Integer): Pointer; stdcall;
    procedure ProcessDataStream(AProcessor: Pointer; AData: PByte; ASize: Int64); stdcall;
    procedure DestroyDataStreamProcessor(AProcessor: Pointer); stdcall;
    /// <summary> обновить графику плагина (таблицы и прочее что в будущем ещё может понадобится вывести) </summary>
    procedure RedrawView; stdcall;
  end;

  ICinemaPluginFactory = interface
    ['{28C95087-0CE1-4BA9-AC59-27CEF5C54FD4}']
    function GetPluginDescriptions: TPluginDescriptions;
    function GetPluginCount: Integer;
    function CreatePlugin(const AName: string; AConfiguration: PWideChar; const ACinema: ICinema): ICinemaPlugin;
  end;

implementation

end.
