﻿unit Cinema.Net.Protocol;

{$ifdef fpc}
	{$mode Delphi}
	{.$define UseUTF16}
{$endif}

{ Протокол взаимодействия между компонентами Cinema }

interface

uses
    Classes
  , SysUtils
  , Cinema.Types
  , Cinema.Hash
  ;

const
  csTlvHashKey               =   0;
  csTlvRunScenario           =   1;
  csTlvStopScenario          =   2;
  csTlvTerminateScenario     =   3;
  csTlvKillScenario          =   4;
  csTlvHealthStateNode       =   5;
  csTlvEvent                 =   6;
  csTlvFile                  =   7;
  csTlvAll                   =   8;
  csTlvIsAlive               =   9;
  csTlvState                 =   10;

  // добавление/обновление имени и конфигурации сценария в список доступных на узле на выполнение
  csTlvAddOrUpdateScenario   =   11;
  csTlvDeleteScenario        =   12;
  csTlvHostName              =   13;
  csTlvRequestStats          =   14;
  csTlvStats                 =   15;
  csTlvScenStat              =   16;
  csTlvNodeStat              =   17;
  csTlvReqNodeStat           =   18;
  csTlvVersion               =   19;
  csTlvOs                    =   20;
  csTlvStats_v2              =   21;


  csFileTypeScenario         =   0;
  csFileTypeLibraryPlugins   =   1;
  csFileCinemaNode           =   2;
  csFile                     =   3;


  {$ifdef MSWINDOWS}
    csActorName: UnicodeString = 'Actor.exe';
    csCineamNodeName: UnicodeString = 'CinemaNode.exe';
  {$endif}
  {$ifdef posix}
    csActorName: UnicodeString = 'Actor';
    csCineamNodeName: UnicodeString = 'CinemaNode';
  {$endif}

  csPluginTablesDelimeter: UnicodeString = 'CinemaTable';

type

  // Базовый заголовок блока данных. В конце блока после данных размещается хэш
  // Если вложенные записи будут иметь поля не кратные 4 байтам, необходимо подумать о выравнивании
  // и возможно проще отказаться от таких полей
  TTlvHeaderBase = packed record
    TypeItem: Int32;   // тип данных
    Length: Int32;     // длина данных, включая хэш в конце блока
    //HashKey: Int32;    // хэш ключа вида /group/host/scenario/
  end;

  // Запись для упрощения доступа к вложенным данным (без вычисления смещения)
  TTlvHeaderData = packed record
    Base: TTlvHeaderBase;
    Data: byte; // начало данных
  end;

  PTlvHeaderBase = ^TTlvHeaderBase;
  PTlvHeaderData = ^TTlvHeaderData;

  // Ключ процесса - источника событий
  TTLVHashKey = packed record
    Base: TTlvHeaderBase;
    Number: int32;
    Pid: uint32;
  end;

  TTLVHashKeyData = packed record
    HashKey: TTLVHashKey;
    Data: Char;
  end;

  PTLVHashKey = ^TTLVHashKey;
  PTLVHashKeyData = ^TTLVHashKeyData;

  // Файл
  TTLVFile = packed record
    Base: TTlvHeaderBase;
    FileType: int32;        // тип файла
    LenFileNameDest: int32;
    Attributes: int32;
    FileCreate: TDateTime;
    FileLastAccess: TDateTime;
    FileLastModified: TDateTime;
  end;

  TTLVFileData = packed record
    TLVFile: TTLVFile;
    Data: Char;  // первый символ имени файла, за ним в памяти следуют остальные
  end;

  PTLVFile = ^TTLVFile;
  PTLVFileData = ^TTLVFileData;

  // сотояние элемента сценария
  TTLVState = packed record
    Base: TTlvHeaderBase;
    State: int32;
    SourceId: TGuid;
  end;

  PTLVState = ^TTLVState;

  // Качество выполняемой работы элемента сценария
  TTLVHealthState = packed record
    Base: TTlvHeaderBase;
    State: int32;
    HealthState: int32;
    SourceId: TGuid;
    LenKey: int32;
    LenMsg: int32;
    // AHealthStateMsg: string;
  end;

  TTLVHealthStateData = packed record
    HealthState: TTLVHealthState;
    Data: Char;
  end;

  PTLVHealthState = ^TTLVHealthState;
  PTLVHealthStateData = ^TTLVHealthStateData;

  // Событие
  TTLVEvent = packed record
    Base: TTlvHeaderBase;
    SourceId: TGuid;
    DateTime: TDateTime;
    EventType: int32;
    LenMsg: int32;
    // AEventMsg: string; - через ';' следуют Source, WhoIs, Msg
  end;

  TTLVEventData = packed record
    TLVEvent: TTLVEvent;
    Data: Char;
  end;

  PTLVEvent = ^TTLVEvent;
  PTLVEventData = ^TTLVEventData;

  // Имя сценария, тип TLV определяет назначение - добовление или обновление, запуск, остановка, прерывание, удалить сценарий из списка на узле
  TTLVScenarioAction = packed record
    Base: TTlvHeaderBase;
    Param: int32;   // номер, кол-во экземпляров, удалять(1)/не удалять(0) файл сценария - в зависимости от типа действия
  end;

  TTLVScenarioActionData = packed record
    TLVScenarioAction: TTLVScenarioAction;
    Data: Char;     // первый символ имени в буфере, за ним в памяти следуют остальные
  end;

  PTLVScenarioAction = ^TTLVScenarioAction;
  PTLVScenarioActionData = ^TTLVScenarioActionData;

  TTlvProcessor = procedure(ATlv: PTlvHeaderBase) of object;

  TProcessErrorNotify = procedure (ATlv: PTlvHeaderBase; const AMsg: UnicodeString) of object;

  // запрос статистики
  TTLVRequestStats = packed record
    Base: TTlvHeaderBase;
    // если ГУИД пустой - запрос общей статистики по сценарию
    Guid: TGuid;
  end;

  TTLVRequestStatsData = packed record
    StatsHeader: TTLVRequestStats;
    Data: char;
  end;

  PTLVRequestStats = ^TTLVRequestStats;
  PTLVRequestStatsData = ^TTLVRequestStatsData;

  // статистика элемента сценария
  TTLVScenarioProcessorStat = packed record
    Base: TTlvHeaderBase;
    Guid: TGuid;
    InputSpeed: uint64;
    OutputSpeed: uint64;
    CountInputDataStreams: uint32;
    CountOutputDataStreams: uint32;
    CountTables: int32;
    LenDataTables: int32;
  end;

  // статистика элемента сценария
  TTLVScenarioProcessorStat_v2 = packed record
    Base: TTlvHeaderBase;
    Guid: TGuid;
    InputSpeed: uint64;
    OutputSpeed: uint64;
    CountInputDataStreams: uint32;
    CountOutputDataStreams: uint32;
    QueuesItemsCount: uint32;
    QueuesSize: uint32;
    CountTables: int32;
    LenDataTables: int32;
  end;

  TTLVScenarioProcessorStatData = packed record
    Stats: TTLVScenarioProcessorStat;
    Data: char;
  end;

  PTLVScenarioProcessorStat = ^TTLVScenarioProcessorStat;
  PTLVScenarioProcessorStatData = ^TTLVScenarioProcessorStatData;

  TTLVScenarioProcessorStatData_v2 = packed record
    Stats: TTLVScenarioProcessorStat_v2;
    Data: char;
  end;

  PTLVScenarioProcessorStat_v2 = ^TTLVScenarioProcessorStat_v2;
  PTLVScenarioProcessorStatData_v2 = ^TTLVScenarioProcessorStatData_v2;

  // общая статистика по сценарию
  TTlvScenarioStat = packed record
    Base: TTlvHeaderBase;
    // общий объём памяти
    TotalMem: int64;
    // доступно памяти
    AvailableMemory: int64;
    // память выделенная под процесс выполнения сценария
    MemoryUsage: int64;
    CPUUsage: int32;
  end;

  TTlvScenarioStatData = packed record
    ScenarioStat: TTlvScenarioStat;
    Data: char;
  end;

  PTlvScenarioStat = ^TTlvScenarioStat;
  PTlvScenarioStatData = ^TTlvScenarioStatData;

  // общая статистика по хосту
  TTlvNodeStat = packed record
    Base: TTlvHeaderBase;
    TotalMem: int64;
    MemoryUsage: int64;
    CPUUsage: int32;
  end;

  TTlvNodeStatData = packed record
    ScenarioStat: TTlvNodeStat;
    Data: char;
  end;

  PTlvNodeStat = ^TTlvNodeStat;
  PTlvNodeStatData = ^TTlvNodeStatData;

  // запрос статистики
  TTLVVersion = packed record
    Base: TTlvHeaderBase;
  end;

  TTLVVersionData = packed record
    VersionHeader: TTLVVersion;
    Data: char;
  end;

  PTLVVersion = ^TTLVVersion;
  PTLVVersionData = ^TTLVVersionData;

  { TCinemaNetProtocol }

  TCinemaNetProtocol = class
  private
    class var
      FWorkDirectory: UnicodeString;
    class var
      FWorkSubDirScens: UnicodeString;
    class constructor Create;
  public
    const
      csServerForHypervisorPort  = 23356;
      csServerForCinemaNodePort  = 23358;
      csProxyServerForActorsPort = 23357;
  public
    class property csWorkDirectory: UnicodeString read FWorkDirectory;
    class property csWorkSubDirScens: UnicodeString read FWorkSubDirScens;
  private
    FBuffer: TMemoryStream;
    FReadBuffer: TMemoryStream;
    FTlvProcessors: array of TTlvProcessor;
    FOnProcessError: TProcessErrorNotify;
    FHash: TCinemaHash;
    FHashRecieve: TCinemaHash;
    // ключ источника данных
    FKey: UnicodeString;

    procedure SetKey(const Value: UnicodeString);

    function GetString(ATlv: PTlvHeaderBase): UnicodeString;
    function ProcessScenarioAction(ATLVScenarioActionData: PTLVScenarioActionData): UnicodeString;

    // методы распаковки
    procedure ProcessTlvHashKey(ATlv: PTlvHeaderBase);
    procedure ProcessTlvHealthStateNode(ATlv: PTlvHeaderBase);
    procedure ProcessTlvEvent(ATlv: PTlvHeaderBase);
    procedure ProcessTlvFile(ATlv: PTlvHeaderBase);
    procedure ProcessTlvRunScenario(ATlv: PTlvHeaderBase);
    procedure ProcessTlvStopScenario(ATlv: PTlvHeaderBase);
    procedure ProcessTlvTerminateScenario(ATlv: PTlvHeaderBase);
    procedure ProcessTlvKillScenario(ATlv: PTlvHeaderBase);
    procedure ProcessTlvAll(ATlv: PTlvHeaderBase);
    procedure ProcessTlvIsAlive(ATlv: PTlvHeaderBase);
    procedure ProcessTlvState(ATlv: PTlvHeaderBase);
    procedure ProcessTlvAddOrUpdateScenario(ATlv: PTlvHeaderBase);
    procedure ProcessTlvDeleteScenario(ATlv: PTlvHeaderBase);
    procedure ProcessTlvHostName(ATlv: PTlvHeaderBase);
    procedure ProcessTlvReqStats(ATlv: PTlvHeaderBase);
    procedure ProcessTlvProcessorStats(ATlv: PTlvHeaderBase);
    procedure ProcessTlvProcessorStats_v2(ATlv: PTlvHeaderBase);
    procedure ProcessTlvScenStat(ATlv: PTlvHeaderBase);
    procedure ProcessTlvNodeStat(ATlv: PTlvHeaderBase);
    procedure ProcessTlvReqNodeStat(ATlv: PTlvHeaderBase);
    procedure ProcessTlvVersion(ATlv: PTlvHeaderBase);
    procedure ProcessTlvOs(ATlv: PTlvHeaderBase);

  protected
    procedure PutTlvScenarioAction(AType: Int32; const AKey: UnicodeString; AParam: int32);

    procedure RegisterTlvProcessor(ATlvType: Int32; AProcessor: TTlvProcessor);

    procedure PutTlv(AType: Int32; ASize: Int32; AData: PByte); overload; virtual;
    procedure PutTlv(AType: Int32; const AData: UnicodeString); overload;
    procedure PutTlv(AType: Int32; const AData: AnsiString); overload;
    procedure PutTlv(AType: Int32; AData: Integer); overload;
    procedure PutTlv(AType: Int32; AData: Int64); overload;

    //
    procedure DoKey(const AKey: UnicodeString; APid: uint32; ANumber: int32); virtual;
    procedure DoHealthStateNode(const AKey: UnicodeString; AState: byte; AHealthState: byte; const AHealthStateMsg: UnicodeString; const ASourceId: TGuid); virtual;
    procedure DoEvent(const ASource, AWhoIs, AMsg: UnicodeString; ADateTime: TDateTime; AEventType: int32; const ASourceId: TGuid); virtual;
    procedure DoRunScenario(const AName: UnicodeString; ACountInstances: int32); virtual;
    procedure DoStopScenario(const AName: UnicodeString); virtual;
    procedure DoTerminateScenario(const AName: UnicodeString); virtual;
    procedure DoKillScenario(const AName: UnicodeString); virtual;
    procedure DoFile(const AFileName: UnicodeString; AFileType: int32); virtual;
    procedure DoAll(const AKey: UnicodeString); virtual;
    procedure DoIsAlive(const AKey: UnicodeString); virtual;
    procedure DoState(const ASource: TGuid; AState: int32); virtual;
    procedure DoAddOrUpdateScenario(const AKey, AName: UnicodeString; ACountInstances: int32); virtual;
    procedure DoDeleteScenario(const AKey: UnicodeString; ADeleteFile: boolean); virtual;
    procedure DoHostName(const AName: UnicodeString); virtual;
    procedure DoReqStat(const AKey, AHostName: UnicodeString; const AGuid: TGuid); virtual;
    procedure DoReqNodeStat(const AName: UnicodeString); virtual;
    procedure DoProcessorStat(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat; const ATables: UnicodeString); virtual;
    procedure DoProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString); virtual;
    procedure DoScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte); virtual;
    procedure DoNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte); virtual;
    procedure DoVersion(const AKey: UnicodeString; const AVersion: UnicodeString); virtual;
    procedure DoOs(const AOS: UnicodeString); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure PutTlvKey(ANumber: int32);
    procedure PutTlvFile(const AFileName, ADestFileName: UnicodeString; AFileType: int32);
    procedure PutTlvState(AState: byte; const ASourceId: TGuid);
    procedure PutTlvHealthState(const AKey: UnicodeString; AState: byte; AHealthState: byte; const AHealthStateMsg: UnicodeString; const ASourceId: TGuid);
    procedure PutTlvEvent(const ASource, AWhoIs, AMsg: UnicodeString; ADateTime: TDateTime; AEventType: int32; const ASourceId: TGuid);
    procedure PutTlvRunScenario(const AKey, AName: UnicodeString; ACountInstances: int32);
    procedure PutTlvStopScenario(const AKey: UnicodeString); virtual;
    procedure PutTlvTerminateScenario(const AKey: UnicodeString); virtual;
    procedure PutTlvKillScenario(const AKey: UnicodeString); virtual;
    procedure PutTlvAll(const AKey: UnicodeString);
    procedure PutTlvIsAlive(const AKey: UnicodeString);
    procedure PutTlvAddOrUpdateScenario(const AKey, AName: UnicodeString; ACountInstances: int32);
    procedure PutTlvDeleteScenario(const AKey: UnicodeString; ADeleteFile: boolean);
    procedure PutTlvHostName(const AName: UnicodeString);
    procedure PutTlvGetStatTables(const AKey, AHostName: UnicodeString; const AGuid: TGuid);
    procedure PutTlvGetNodeStat(const AKey: UnicodeString);
    procedure PutTlvScenarioProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStatsHeader: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString);
    procedure PutTlvScenarioProcessorStat(const AKey, AHostName: UnicodeString; var AStatsHeader: TTLVScenarioProcessorStat; const ATables: UnicodeString);
    procedure PutTlvScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte);
    procedure PutTlvNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte);
    procedure PutTlvVersion(const AKey: UnicodeString; const AVersion: UnicodeString);
    procedure PutTlvCinemaNode(const ASourcePath, ADestName: UnicodeString);
    procedure PutTlvOs(const AOs: UnicodeString);

    procedure ProcessTlv(AData: PByte; ASize: Int32); virtual;

    property Buffer: TMemoryStream read FBuffer;

    // ключ данных
    property Key: UnicodeString read FKey write SetKey;

    property OnProcessError: TProcessErrorNotify read FOnProcessError write FOnProcessError;

  end;

function ScenarioNameToKey(var ASncenarioName: String; const AFullWorkDir: String): UnicodeString;
function ExtractParentKey(const AChildKey: UnicodeString): UnicodeString;
function TrancParentKey(const AKey: UnicodeString): UnicodeString;

implementation

uses
  {$ifdef fpc}
  //LazFileUtils,
  {$else}
  System.IOUtils,
  {$endif}
    Math
  , Cinema.Utils
  ;

function ScenarioNameToKey(var ASncenarioName: String; const AFullWorkDir: String): UnicodeString;
var
  splitted: TArray<string>;
begin
  // возьмём в качесве ключа, по возможности, структуру рабочего каталога
  splitted := ASncenarioName.Split(AFullWorkDir);
  if Length(splitted) = 2 then
  begin
    ASncenarioName := AFullWorkDir + splitted[1];
    Result := StringToWide(splitted[1]);
  end else
  begin
    // на случай, если указан относительный путь
    splitted := ASncenarioName.Split([WideToString(TCinemaNetProtocol.csWorkDirectory) + '/']);
    if (Length(splitted) = 2) and (splitted[0] = '') then
    begin
      ASncenarioName := AFullWorkDir + splitted[1];
      Result := StringToWide(splitted[1]);
    end else
      Result := StringToWide(ExtractFileName(ASncenarioName));
  end;
  Result := UnicodeString(ChangeFileExt(Result, UnicodeString('')));
end;

function ExtractParentKey(const AChildKey: UnicodeString): UnicodeString;
var
  i: Integer;
begin
  Result := '';
  for i := length(AChildKey) downto 2 do
    if CharInSet(AChildKey[i], ['/', '\']) then
    begin
      SetLength(Result, i - 1);
      move(AChildKey[1], Result[1], (i - 1) shl 1);
    end;
end;

function TrancParentKey(const AKey: UnicodeString): UnicodeString;
var
  i: Integer;
begin
  Result := '';
  for i := 1 to length(AKey) - 1 do
    if CharInSet(AKey[i], ['/', '\']) then
    begin
      SetLength(Result, length(AKey) - i);
      move(AKey[i+1], Result[1], (length(AKey) - i) shl 1);
    end;
end;

{ TCinemaNetProtocol }

constructor TCinemaNetProtocol.Create;
begin
  FBuffer := TMemoryStream.Create;
  FReadBuffer := TMemoryStream.Create;

  RegisterTlvProcessor(csTlvHashKey, ProcessTlvHashKey);
  RegisterTlvProcessor(csTlvRunScenario, ProcessTlvRunScenario);
  RegisterTlvProcessor(csTlvStopScenario, ProcessTlvStopScenario);
  RegisterTlvProcessor(csTlvTerminateScenario, ProcessTlvTerminateScenario);
  RegisterTlvProcessor(csTlvKillScenario, ProcessTlvKillScenario);
  RegisterTlvProcessor(csTlvHealthStateNode, ProcessTlvHealthStateNode);
  RegisterTlvProcessor(csTlvEvent, ProcessTlvEvent);
  RegisterTlvProcessor(csTlvFile, ProcessTlvFile);
  RegisterTlvProcessor(csTlvAll, ProcessTlvAll);
  RegisterTlvProcessor(csTlvIsAlive, ProcessTlvIsAlive);
  RegisterTlvProcessor(csTlvState, ProcessTlvState);
  RegisterTlvProcessor(csTlvAddOrUpdateScenario, ProcessTlvAddOrUpdateScenario);
  RegisterTlvProcessor(csTlvDeleteScenario, ProcessTlvDeleteScenario);
  RegisterTlvProcessor(csTlvHostName, ProcessTlvHostName);
  RegisterTlvProcessor(csTlvAddOrUpdateScenario, ProcessTlvAddOrUpdateScenario);
  RegisterTlvProcessor(csTlvRequestStats, ProcessTlvReqStats);
  RegisterTlvProcessor(csTlvStats, ProcessTlvProcessorStats);
  RegisterTlvProcessor(csTlvStats_v2, ProcessTlvProcessorStats_v2);
  RegisterTlvProcessor(csTlvScenStat, ProcessTlvScenStat);
  RegisterTlvProcessor(csTlvNodeStat, ProcessTlvNodeStat);
  RegisterTlvProcessor(csTlvReqNodeStat, ProcessTlvReqNodeStat);
  RegisterTlvProcessor(csTlvVersion, ProcessTlvVersion);
  RegisterTlvProcessor(csTlvOs, ProcessTlvOs);
end;

destructor TCinemaNetProtocol.Destroy;
begin
  FreeAndNil(FBuffer);
  FreeAndNil(FReadBuffer);
  inherited;
end;

procedure TCinemaNetProtocol.DoAddOrUpdateScenario(const AKey, AName: UnicodeString; ACountInstances: int32);
begin

end;

procedure TCinemaNetProtocol.DoAll(const AKey: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoDeleteScenario(const AKey: UnicodeString; ADeleteFile: boolean);
begin

end;

procedure TCinemaNetProtocol.DoEvent(const ASource, AWhoIs, AMsg: UnicodeString; ADateTime: TDateTime; AEventType: int32; const ASourceId: TGuid);
begin

end;

procedure TCinemaNetProtocol.DoFile(const AFileName: UnicodeString; AFileType: int32);
begin

end;

procedure TCinemaNetProtocol.DoKey(const AKey: UnicodeString; APid: uint32; ANumber: int32);
begin
  FKey := AKey;
end;

procedure TCinemaNetProtocol.DoHealthStateNode(const AKey: UnicodeString; AState: byte; AHealthState: byte; const AHealthStateMsg: UnicodeString; const ASourceId: TGuid);
begin

end;

procedure TCinemaNetProtocol.DoHostName(const AName: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoIsAlive(const AKey: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoKillScenario(const AName: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte);
begin

end;

procedure TCinemaNetProtocol.DoOs(const AOS: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoReqNodeStat(const AName: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoReqStat(const AKey, AHostName: UnicodeString; const AGuid: TGuid);
begin

end;

procedure TCinemaNetProtocol.DoRunScenario(const AName: UnicodeString; ACountInstances: int32);
begin

end;

procedure TCinemaNetProtocol.DoScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte);
begin

end;

procedure TCinemaNetProtocol.DoState(const ASource: TGuid; AState: int32);
begin

end;

procedure TCinemaNetProtocol.DoProcessorStat(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat; const ATables: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStats: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoStopScenario(const AName: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoTerminateScenario(const AName: UnicodeString);
begin

end;

procedure TCinemaNetProtocol.DoVersion(const AKey: UnicodeString; const AVersion: UnicodeString);
begin

end;

function TCinemaNetProtocol.GetString(ATlv: PTlvHeaderBase): UnicodeString;
begin
  SetLength(Result, (ATlv.Length - 4) shr 1);
  move(PTlvHeaderData(ATlv).Data, Result[1], ATlv.Length - 4);
end;

function TCinemaNetProtocol.ProcessScenarioAction(ATLVScenarioActionData: PTLVScenarioActionData): UnicodeString;
var
  len: int32;
begin
  if (ATLVScenarioActionData.TLVScenarioAction.Base.Length = 0) or (ATLVScenarioActionData.TLVScenarioAction.Base.Length - (SizeOf(TTLVScenarioAction) - SizeOf(TTlvHeaderBase)) - 4 <= 0) then
    exit('');
  len := ATLVScenarioActionData.TLVScenarioAction.Base.Length - (SizeOf(TTLVScenarioAction) - SizeOf(TTlvHeaderBase)) - 4;
  SetLength(Result, len shr 1);
  move(ATLVScenarioActionData.Data, Result[1], len);
end;

procedure TCinemaNetProtocol.ProcessTlv(AData: PByte; ASize: Int32);
var
  data: PByte;
  size: int32;
  tlvSize: int32;
  tlv: PTlvHeaderBase;
begin
  if FReadBuffer.Size > 0 then
  begin
    FReadBuffer.Position := FReadBuffer.Size;
    FReadBuffer.Write(AData^, ASize);
    data := FReadBuffer.Memory;
    size := FReadBuffer.Size;
  end else
  begin
    data := AData;
    size := ASize;
    // хэш
    FHashRecieve.Reset;
  end;

  while size >= SizeOf(TTlvHeaderBase) + 4 do
  begin
    tlv := PTlvHeaderBase(data);
    tlvSize := tlv.Length + SizeOf(TTlvHeaderBase);
    if (tlvSize > size) or (tlv.Length < 0) or (tlv.TypeItem >= length(FTlvProcessors)) then
    begin
      if (tlv.Length < 0) or (tlv.TypeItem >= length(FTlvProcessors)) then
      begin
        size := 0;
        if Assigned(FOnProcessError) then
        begin
          if (tlv.TypeItem >= length(FTlvProcessors)) then
            FOnProcessError(tlv, 'Error in TLV (' + WideString(IntToStr(tlv.TypeItem)) + ') protocol: incorrect Type Item')
          else
            FOnProcessError(tlv, 'Error in TLV (' + WideString(IntToStr(tlv.TypeItem)) + ') protocol: incorrect length');
        end;
      end;
      break;
    end;

    dec(size, tlvSize);
    inc(data, tlvSize);
    FHashRecieve.Update(tlv^, tlvSize - 4);

    if (FHashRecieve.Hash = PCardinal(data-4)^) and Assigned(FTlvProcessors[tlv.TypeItem]) then
    begin
      FTlvProcessors[tlv.TypeItem](tlv);
      // хэш
      FHashRecieve.Reset;
    end else
    begin
      size := 0;
      if Assigned(FOnProcessError) then
      begin
        FOnProcessError(tlv, 'Error in TLV (' + WideString(IntToStr(tlv.TypeItem)) + ') protocol: incorrect hash');
      end;
      break;
    end;
  end;

  // запишем остаток
  if size > 0 then
  begin
    FReadBuffer.Position := 0;
    FReadBuffer.Write(data^, size);
    FReadBuffer.Size := size;
  end else
  begin
    // очистим чтоб сбросить память, на случай если принимались большие данные
    FReadBuffer.Clear;
  end;
end;

procedure TCinemaNetProtocol.ProcessTlvAddOrUpdateScenario(ATlv: PTlvHeaderBase);
var
  data: TArray<UnicodeString>;
begin
  data := ProcessScenarioAction(PTLVScenarioActionData(ATlv)).Split([';']);
  if length(data) = 2 then
    DoAddOrUpdateScenario(data[0], data[1].Replace(ContrPathDelim, PathDelim), PTLVScenarioAction(ATlv).Param);
end;

procedure TCinemaNetProtocol.ProcessTlvAll(ATlv: PTlvHeaderBase);
begin
  DoAll(GetString(ATlv));
end;

procedure TCinemaNetProtocol.ProcessTlvDeleteScenario(ATlv: PTlvHeaderBase);
var
  key: UnicodeString;
begin
  key := ProcessScenarioAction(PTLVScenarioActionData(ATlv));
  DoDeleteScenario(key, PTLVScenarioAction(ATlv).Param > 0);
end;

procedure TCinemaNetProtocol.ProcessTlvHashKey(ATlv: PTlvHeaderBase);
var
  tlvHashKey: PTLVHashKeyData;
  key: UnicodeString;
  lenKey: int32;
begin
  tlvHashKey := PTLVHashKeyData(ATlv);
  if tlvHashKey.HashKey.Base.Length < 13 then
    exit;
  lenKey := tlvHashKey.HashKey.Base.Length - (SizeOf(TTLVHashKey) - SizeOf(TTlvHeaderBase)) - 4;
  SetLength(key, lenKey shr 1);
  move(tlvHashKey.Data, key[1], lenKey);
  DoKey(key, tlvHashKey.HashKey.Pid, tlvHashKey.HashKey.Number);
end;

procedure TCinemaNetProtocol.ProcessTlvEvent(ATlv: PTlvHeaderBase);
var
  tlvEventData: PTLVEventData;
  eventData: UnicodeString;
  event: TArray<UnicodeString>;
begin
  tlvEventData := PTLVEventData(ATlv);
  if tlvEventData.TLVEvent.LenMsg <= 0 then
    exit;
  SetLength(eventData, tlvEventData.TLVEvent.LenMsg shr 1);
  move(tlvEventData.Data, eventData[1], tlvEventData.TLVEvent.LenMsg);
  event := eventData.Split([';']);
  if length(event) = 3 then
    DoEvent(event[0], event[1], event[2], tlvEventData.TLVEvent.DateTime, tlvEventData.TLVEvent.EventType, tlvEventData.TLVEvent.SourceId);
end;

procedure TCinemaNetProtocol.ProcessTlvFile(ATlv: PTlvHeaderBase);
var
  fileData: PTLVFileData;
  fileName, directory: UnicodeString;
  fileStream: TFileStream;
  path: TArray<UnicodeString>;
  i: Integer;
begin
  fileData := PTLVFileData(ATlv);
  SetLength(fileName, fileData.TLVFile.LenFileNameDest shr 1);
  move(fileData.Data, fileName[1], fileData.TLVFile.LenFileNameDest);

  path := fileName.Split(['/', '\']);
  directory := WideString(ExtractFilePath(ParamStr(0)));

  for i := 0 to length(path) - 2 do
  begin
    directory := directory + path[i] + PathDelim;
    if not DirectoryExists(directory) then
      CreateDir(directory);
  end;

  fileName := directory + path[length(path)-1];

  try
    fileStream := TFileStream.Create(WideToString(fileName), fmCreate);
  except
    on e: Exception do
    begin
      if Assigned(OnProcessError) then
        OnProcessError(ATlv, 'Ошибка создания файла: ' + e{%H-}.Message{%H-});
      exit;
    end;
  end;

  try
    try
      fileStream.Write((PByte(@fileData.Data) + fileData.TLVFile.LenFileNameDest)^, fileData.TLVFile.Base.Length - (SizeOf(TTLVFile) - SizeOf(TTlvHeaderBase)) - fileData.TLVFile.LenFileNameDest - 4);
    finally
      fileStream.Free;
    end;
  except
    on e: Exception do
    begin
      if Assigned(OnProcessError) then
        OnProcessError(ATlv, 'Ошибка записи в файл: ' + e{%H-}.Message{%H-});
      exit;
    end;
  end;
  // установка атрибутов
  //TFile.SetAttributes(fileName, TFile.IntegerToFileAttributes(fileData.TLVFile.Attributes));
  {$ifdef fpc}
  	// todo or not todo
  {$else}
    TFile.SetCreationTime(fileName, fileData.TLVFile.FileCreate);
    TFile.SetLastWriteTime(fileName, fileData.TLVFile.FileLastModified);
    TFile.SetLastAccessTime(fileName, fileData.TLVFile.FileLastAccess);
  {$endif}
  DoFile(fileName, fileData.TLVFile.FileType);
end;

procedure TCinemaNetProtocol.ProcessTlvReqNodeStat(ATlv: PTlvHeaderBase);
begin
  DoReqNodeStat(GetString(ATlv));
end;

procedure TCinemaNetProtocol.ProcessTlvReqStats(ATlv: PTlvHeaderBase);
var
  statsReq: PTLVRequestStatsData;
  len: int32;
  keyAndHostName: UnicodeString;
  splitted: TArray<String>;
begin
  statsReq := PTLVRequestStatsData(ATlv);
  len := statsReq.StatsHeader.Base.Length - (SizeOf(TTLVRequestStats) - SizeOf(TTlvHeaderBase)) - 4;
  SetLength(keyAndHostName, len shr 1);
  move(statsReq.Data, keyAndHostName[1], len);
  splitted := string(keyAndHostName).Split([';']);
  DoReqStat(UnicodeString(splitted[0]), UnicodeString(splitted[1]), statsReq.StatsHeader.Guid);
end;

procedure TCinemaNetProtocol.ProcessTlvRunScenario(ATlv: PTlvHeaderBase);
var
  name: UnicodeString;
begin
  name := ProcessScenarioAction(PTLVScenarioActionData(ATlv));
  if not name.IsEmpty then
    DoRunScenario(name, PTLVScenarioAction(ATlv).Param);
end;

procedure TCinemaNetProtocol.ProcessTlvHealthStateNode(ATlv: PTlvHeaderBase);
var
  tlvHealthState: PTLVHealthStateData;
  key: UnicodeString;
  stateMessage: UnicodeString;
begin
  tlvHealthState := PTLVHealthStateData(ATlv);
  if tlvHealthState.HealthState.LenKey > 0 then
  begin
    SetLength(key, tlvHealthState.HealthState.LenKey shr 1);
    move(tlvHealthState.Data, key[1], tlvHealthState.HealthState.LenKey);
  end else
    key := '';

  if tlvHealthState.HealthState.LenMsg > 0 then
  begin
    SetLength(stateMessage, tlvHealthState.HealthState.LenMsg shr 1);
    move((PByte(@(tlvHealthState.Data)) + tlvHealthState.HealthState.LenKey)^, stateMessage[1], tlvHealthState.HealthState.LenMsg);
  end else
    stateMessage := '';

  DoHealthStateNode(key, tlvHealthState.HealthState.State, tlvHealthState.HealthState.HealthState, stateMessage, tlvHealthState.HealthState.SourceId);
end;

procedure TCinemaNetProtocol.ProcessTlvHostName(ATlv: PTlvHeaderBase);
begin
  DoHostName(GetString(ATlv));
end;

procedure TCinemaNetProtocol.ProcessTlvIsAlive(ATlv: PTlvHeaderBase);
begin
  DoIsAlive(GetString(ATlv));
end;

procedure TCinemaNetProtocol.ProcessTlvKillScenario(ATlv: PTlvHeaderBase);
var
  name: UnicodeString;
begin
  name := ProcessScenarioAction(PTLVScenarioActionData(ATlv));
  if not name.IsEmpty then
    DoKillScenario(name);
end;

procedure TCinemaNetProtocol.ProcessTlvNodeStat(ATlv: PTlvHeaderBase);
var
  stats: PTlvNodeStatData;
  len: int32;
  data: UnicodeString;
begin
  stats := PTlvNodeStatData(ATlv);
  len := stats.ScenarioStat.Base.Length - (SizeOf(TTlvNodeStat) - SizeOf(TTlvHeaderBase)) - 4;
  SetLength(data, len shr 1);
  move(stats.Data, data[1], len);

  DoNodeStat(data, stats.ScenarioStat.TotalMem, stats.ScenarioStat.MemoryUsage, stats.ScenarioStat.CPUUsage);
end;

procedure TCinemaNetProtocol.ProcessTlvOs(ATlv: PTlvHeaderBase);
begin
  DoOs(GetString(ATlv));
end;

procedure TCinemaNetProtocol.ProcessTlvScenStat(ATlv: PTlvHeaderBase);
var
  stats: PTlvScenarioStatData;
  len: int32;
  data: UnicodeString;
  splitted: TArray<UnicodeString>;
begin
  stats := PTlvScenarioStatData(ATlv);
  len := stats.ScenarioStat.Base.Length - (SizeOf(TTlvScenarioStat) - SizeOf(TTlvHeaderBase)) - 4;
  SetLength(data, len shr 1);
  move(stats.Data, data[1], len);
  splitted := data.Split([';']);
  DoScenarioStat(splitted[0], splitted[1], stats.ScenarioStat.TotalMem, stats.ScenarioStat.AvailableMemory, stats.ScenarioStat.MemoryUsage, stats.ScenarioStat.CPUUsage);
end;

procedure TCinemaNetProtocol.ProcessTlvState(ATlv: PTlvHeaderBase);
var
  state: PTLVState;
begin
  state := PTLVState(ATlv);
  DoState(state.SourceId, state.State);
end;

procedure TCinemaNetProtocol.ProcessTlvProcessorStats(ATlv: PTlvHeaderBase);
var
  statsReq: PTLVScenarioProcessorStatData;
  len: int32;
  data: UnicodeString;
  splitted: TArray<UnicodeString>;
begin
  statsReq := PTLVScenarioProcessorStatData(ATlv);
  len := statsReq.Stats.Base.Length - (SizeOf(TTLVScenarioProcessorStat) - SizeOf(TTlvHeaderBase)) - statsReq.Stats.LenDataTables - 4;
  SetLength(data, len shr 1);
  move(statsReq.Data, data[1], len);
  splitted := data.Split([';']);

  if statsReq.Stats.LenDataTables > 1 then
  begin
    SetLength(data, statsReq.Stats.LenDataTables shr 1);
    move((PByte(@statsReq.Data) + len)^, data[1], statsReq.Stats.LenDataTables);
  end else
  begin
    data := '';
  end;

  DoProcessorStat(splitted[0], splitted[1], statsReq.Stats, data);
end;

procedure TCinemaNetProtocol.ProcessTlvProcessorStats_v2(ATlv: PTlvHeaderBase);
var
  statsReq: PTLVScenarioProcessorStatData_v2;
  len: int32;
  data: UnicodeString;
  splitted: TArray<UnicodeString>;
begin
  statsReq := PTLVScenarioProcessorStatData_v2(ATlv);
  len := statsReq.Stats.Base.Length - (SizeOf(TTLVScenarioProcessorStat_v2) - SizeOf(TTlvHeaderBase)) - statsReq.Stats.LenDataTables - 4;
  SetLength(data, len shr 1);
  move(statsReq.Data, data[1], len);
  splitted := data.Split([';']);

  if statsReq.Stats.LenDataTables > 1 then
  begin
    SetLength(data, statsReq.Stats.LenDataTables shr 1);
    move((PByte(@statsReq.Data) + len)^, data[1], statsReq.Stats.LenDataTables);
  end else
  begin
    data := '';
  end;

  DoProcessorStat_v2(splitted[0], splitted[1], statsReq.Stats, data);
end;

procedure TCinemaNetProtocol.ProcessTlvStopScenario(ATlv: PTlvHeaderBase);
var
  name: UnicodeString;
begin
  name := ProcessScenarioAction(PTLVScenarioActionData(ATlv));
  if not name.IsEmpty then
    DoStopScenario(name);
end;

procedure TCinemaNetProtocol.ProcessTlvTerminateScenario(ATlv: PTlvHeaderBase);
var
  name: UnicodeString;
begin
  name := ProcessScenarioAction(PTLVScenarioActionData(ATlv));
  if not name.IsEmpty then
    DoTerminateScenario(name);
end;

procedure TCinemaNetProtocol.ProcessTlvVersion(ATlv: PTlvHeaderBase);
var
  statsReq: PTLVVersionData;
  len: int32;
  keyAndVersion: UnicodeString;
  splitted: TArray<String>;
begin
  statsReq := PTLVVersionData(ATlv);
  len := statsReq.VersionHeader.Base.Length - (SizeOf(TTLVVersion) - SizeOf(TTlvHeaderBase)) - 4;
  SetLength(keyAndVersion, len shr 1);
  move(statsReq.Data, keyAndVersion[1], len);
  splitted := string(keyAndVersion).Split([';']);
  DoVersion(UnicodeString(splitted[0]), UnicodeString(splitted[1]));
end;

procedure TCinemaNetProtocol.PutTlv(AType: Int32; const AData: UnicodeString);
begin
  if Length(AData) > 0 then
    PutTlv(AType, Length(AData) shl 1, @AData[1]);
end;

procedure TCinemaNetProtocol.PutTlv(AType: Int32; const AData: AnsiString);
begin
  if Length(AData) > 0 then
    PutTlv(AType, Length(AData), @AData[1]);
end;

procedure TCinemaNetProtocol.PutTlv(AType: Int32; AData: Integer);
begin
  PutTlv(AType, SizeOf(AData), @AData);
end;

procedure TCinemaNetProtocol.PutTlvFile(const AFileName, ADestFileName: UnicodeString; AFileType: int32);
var
  tlvFile: TTLVFile;
  f: TFileStream;
  hash: uint32;
  pos: int32;
begin
  if not FileExists(AFileName) then
    exit;
  f := TFileStream.Create(WideToString(AFileName), fmOpenRead);
  try
    FillChar(tlvFile, SizeOf(tlvFile), 0);
    tlvFile.LenFileNameDest := length(ADestFileName) shl 1;
    tlvFile.Base.TypeItem := csTlvFile;
    tlvFile.Base.Length := sizeof(TTLVFile) - sizeOf(TTlvHeaderBase) + tlvFile.LenFileNameDest + f.Size + 4;
    tlvFile.FileType := AFileType;
    {$ifdef fpc}
    // todo or not todo
    {$else}
    tlvFile.Attributes := TFile.FileAttributesToInteger(TFile.GetAttributes(AFileName));
    tlvFile.FileCreate := TFile.GetCreationTime(AFileName);
    tlvFile.FileLastAccess := TFile.GetLastAccessTime(AFileName);
    tlvFile.FileLastModified := TFile.GetLastWriteTime(AFileName);
    {$endif}

    FHash.Reset;
    FHash.Update(tlvFile, SizeOf(tlvFile));
    // В символах???
    FHash.Update(ADestFileName[1], tlvFile.LenFileNameDest);
    // заголовок
    FBuffer.Write(tlvFile, SizeOf(tlvFile));
    // данные
    // В символах???
    FBuffer.Write(ADestFileName[1], tlvFile.LenFileNameDest);

    pos := FBuffer.Position;
    FBuffer.Size := pos + f.Size;

    f.Read((pByte(FBuffer.Memory) + pos)^, f.Size);

    FHash.Update((pByte(FBuffer.Memory) + pos)^, f.Size);

    FBuffer.Position := FBuffer.Size;
    // хэш
    hash := FHash.Hash;
    FBuffer.Write(hash, SizeOf(hash));
  finally
    f.Free;
  end;

end;

procedure TCinemaNetProtocol.PutTlvGetNodeStat(const AKey: UnicodeString);
begin
  PutTlv(csTlvReqNodeStat, AKey);
end;

procedure TCinemaNetProtocol.PutTlvGetStatTables(const AKey, AHostName: UnicodeString; const AGuid: TGuid);
var
  statsReq: TTLVRequestStats;
  hash: uint32;
  keyAndHostName: UnicodeString;
begin
  keyAndHostName := AKey + ';' + AHostName;
  statsReq.Base.Length := SizeOf(TTLVRequestStats) - SizeOf(TTlvHeaderBase) + keyAndHostName.Length shl 1 + 4;
  statsReq.Base.TypeItem := csTlvRequestStats;
  statsReq.Guid := AGuid;

  FHash.Reset();
  FHash.Update(statsReq, SizeOf(statsReq));
  FHash.Update(keyAndHostName[1], keyAndHostName.Length shl 1);

  // заголовок
  FBuffer.Write(statsReq, SizeOf(statsReq));

  // данные
  FBuffer.Write(keyAndHostName[1], keyAndHostName.Length shl 1);

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvKey(ANumber: int32);
var
  header: TTLVHashKey;
  hash: uint32;
  lenKey: int32;
begin
  lenKey := FKey.Length shl 1;
  header.Base.TypeItem := csTlvHashKey;
  header.Base.Length := SizeOf(TTLVHashKey) - SizeOf(TTlvHeaderBase) + lenKey + 4;

  header.Pid := GetCurrentProcessId;
  header.Number := ANumber;

  FHash.Reset();
  FHash.Update(header, SizeOf(header));
  // заголовок
  FBuffer.Write(header, SizeOf(header));
 // данные
  if FKey.Length > 0 then
  begin
    FHash.Update(FKey[1], lenKey);
    FBuffer.Write(FKey[1], lenKey);
  end;

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvHealthState(const AKey: UnicodeString; AState: byte; AHealthState: byte; const AHealthStateMsg: UnicodeString; const ASourceId: TGuid);
var
  healthState: TTLVHealthState;
  hash: uint32;
begin
  healthState.LenKey := length(AKey) shl 1;
  healthState.LenMsg := length(AHealthStateMsg) shl 1;
  healthState.Base.TypeItem := csTlvHealthStateNode;
  healthState.Base.Length := sizeof(TTLVHealthState) - sizeOf(TTlvHeaderBase) + healthState.LenMsg + healthState.LenKey + 4;
  healthState.State := AState;
  healthState.HealthState := AHealthState;
  healthState.SourceId := ASourceId;
  FHash.Reset();
  FHash.Update(healthState, SizeOf(healthState));
  if healthState.LenKey > 0 then
    FHash.Update(AKey[1], healthState.LenKey);
  if healthState.LenMsg > 0 then
    FHash.Update(AHealthStateMsg[1], healthState.LenMsg);
  hash := FHash.Hash;
  // заголовок
  FBuffer.Write(healthState, SizeOf(healthState));
  // данные
  if healthState.LenKey > 0 then
    FBuffer.Write(AKey[1], healthState.LenKey);
  if healthState.LenMsg > 0 then
    FBuffer.Write(AHealthStateMsg[1], healthState.LenMsg);
  // хэш
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvHostName(const AName: UnicodeString);
begin
  PutTlv(csTlvHostName, AName);
end;

procedure TCinemaNetProtocol.PutTlvIsAlive(const AKey: UnicodeString);
begin
  PutTlv(csTlvIsAlive, AKey);
end;

procedure TCinemaNetProtocol.PutTlvKillScenario(const AKey: UnicodeString);
begin
  PutTlvScenarioAction(csTlvKillScenario, AKey, 0);
end;

procedure TCinemaNetProtocol.PutTlvNodeStat(const AKey: UnicodeString; ATotalMem, AMemUsage: int64; ACPUUsage: byte);
var
  scenStat: TTlvNodeStat;
  hash: uint32;
  lenHost: int32;
begin
  lenHost := AKey.Length shl 1;
  scenStat.Base.Length := SizeOf(TTlvNodeStat) - SizeOf(TTlvHeaderBase) + lenHost + 4;
  scenStat.Base.TypeItem := csTlvNodeStat;
  scenStat.TotalMem := ATotalMem;
  scenStat.MemoryUsage := AMemUsage;
  scenStat.CPUUsage := ACPUUsage;

  FHash.Reset();
  FHash.Update(scenStat, SizeOf(scenStat));
  FHash.Update(AKey[1], lenHost);

  // заголовок
  FBuffer.Write(scenStat, SizeOf(scenStat));

  // данные
  FBuffer.Write(AKey[1], lenHost);

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvOs(const AOs: UnicodeString);
begin
  PutTlv(csTlvOs, AOs);
end;

procedure TCinemaNetProtocol.PutTlvRunScenario(const AKey, AName: UnicodeString; ACountInstances: int32);
begin
  // сначало положим сценарий
  PutTlvFile(AName, csWorkDirectory + csWorkSubDirScens + ExtractFileName(AName), csFileTypeScenario);
  // теперь команда на его запуск
  PutTlvScenarioAction(csTlvRunScenario, AKey, ACountInstances);
end;

procedure TCinemaNetProtocol.PutTlvScenarioAction(AType: Int32; const AKey: UnicodeString; AParam: int32);
var
  header: TTLVScenarioAction;
  hash: uint32;
  len: int32;
begin
  len := AKey.Length shl 1;
  header.Base.TypeItem := AType;
  header.Base.Length := SizeOf(TTLVScenarioAction) - SizeOf(TTlvHeaderBase) + len + 4;
  header.Param := AParam;

  FHash.Reset();
  FHash.Update(header, SizeOf(header));
  if len > 0 then
    FHash.Update(AKey[1], len);

  // заголовок
  FBuffer.Write(header, SizeOf(header));

  // данные
  if len > 0 then
    FBuffer.Write(AKey[1], len);

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvScenarioProcessorStat_v2(const AKey, AHostName: UnicodeString; var AStatsHeader: TTLVScenarioProcessorStat_v2; const ATables: UnicodeString);
var
  hash: uint32;
  lenHost: int32;
  keyAndHostName: UnicodeString;
begin
  keyAndHostName := AKey + ';' + AHostName;
  AStatsHeader.LenDataTables := ATables.Length shl 1;
  lenHost := keyAndHostName.Length shl 1;
  AStatsHeader.Base.Length := SizeOf(TTLVScenarioProcessorStat_v2) - SizeOf(TTlvHeaderBase) + lenHost + AStatsHeader.LenDataTables + 4;
  AStatsHeader.Base.TypeItem := csTlvStats_v2;

  FHash.Reset();
  FHash.Update(AStatsHeader, SizeOf(AStatsHeader));
  FHash.Update(keyAndHostName[1], lenHost);
  if AStatsHeader.LenDataTables > 0 then
    FHash.Update(ATables[1], AStatsHeader.LenDataTables);

  // заголовок
  FBuffer.Write(AStatsHeader, SizeOf(AStatsHeader));

  // данные
  FBuffer.Write(keyAndHostName[1], lenHost);
  if AStatsHeader.LenDataTables > 0 then
    FBuffer.Write(ATables[1], AStatsHeader.LenDataTables);

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvScenarioProcessorStat(const AKey, AHostName: UnicodeString; var AStatsHeader: TTLVScenarioProcessorStat; const ATables: UnicodeString);
var
  hash: uint32;
  lenHost: int32;
  keyAndHostName: UnicodeString;
begin
  keyAndHostName := AKey + ';' + AHostName;
  AStatsHeader.LenDataTables := ATables.Length shl 1;
  lenHost := keyAndHostName.Length shl 1;
  AStatsHeader.Base.Length := SizeOf(TTLVScenarioProcessorStat) - SizeOf(TTlvHeaderBase) + lenHost + AStatsHeader.LenDataTables + 4;
  AStatsHeader.Base.TypeItem := csTlvStats;

  FHash.Reset();
  FHash.Update(AStatsHeader, SizeOf(AStatsHeader));
  FHash.Update(keyAndHostName[1], lenHost);
  if AStatsHeader.LenDataTables > 0 then
    FHash.Update(ATables[1], AStatsHeader.LenDataTables);

  // заголовок
  FBuffer.Write(AStatsHeader, SizeOf(AStatsHeader));

  // данные
  FBuffer.Write(keyAndHostName[1], lenHost);
  if AStatsHeader.LenDataTables > 0 then
    FBuffer.Write(ATables[1], AStatsHeader.LenDataTables);

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvScenarioStat(const AKey, AHostName: UnicodeString; ATotalMem, AAvailableMem, AMemUsage: int64; ACPUUsage: byte);
var
  scenStat: TTlvScenarioStat;
  hash: uint32;
  lenHost: int32;
  keyAndHostName: UnicodeString;
begin
  keyAndHostName := AKey + ';' + AHostName;
  lenHost := keyAndHostName.Length shl 1;
  scenStat.Base.Length := SizeOf(TTlvScenarioStat) - SizeOf(TTlvHeaderBase) + lenHost + 4;
  scenStat.Base.TypeItem := csTlvScenStat;
  scenStat.TotalMem := ATotalMem;
  scenStat.AvailableMemory := AAvailableMem;
  scenStat.MemoryUsage := AMemUsage;
  scenStat.CPUUsage := ACPUUsage;

  FHash.Reset();
  FHash.Update(scenStat, SizeOf(scenStat));
  FHash.Update(keyAndHostName[1], lenHost);

  // заголовок
  FBuffer.Write(scenStat, SizeOf(scenStat));

  // данные
  FBuffer.Write(keyAndHostName[1], lenHost);

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));

end;

procedure TCinemaNetProtocol.PutTlvState(AState: byte; const ASourceId: TGuid);
var
  state: TTLVState;
  hash: uint32;
begin
  state.Base.TypeItem := csTlvState;
  state.Base.Length := SizeOf(TTLVState) - SizeOf(TTlvHeaderBase) + 4;
  state.State := AState;
  state.SourceId := ASourceId;
  FHash.Reset();
  FHash.Update(state, SizeOf(state));
  hash := FHash.Hash;
  // заголовок
  FBuffer.Write(state, SizeOf(state));
  // хэш
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvStopScenario(const AKey: UnicodeString);
begin
  PutTlvScenarioAction(csTlvStopScenario, AKey, 0);
end;

procedure TCinemaNetProtocol.PutTlvTerminateScenario(const AKey: UnicodeString);
begin
  PutTlvScenarioAction(csTlvTerminateScenario, AKey, 0);
end;

procedure TCinemaNetProtocol.PutTlvVersion(const AKey: UnicodeString; const AVersion: UnicodeString);
var
  version: TTLVVersion;
  hash: uint32;
  keyAndVersion: UnicodeString;
begin
  keyAndVersion := AKey + ';' + AVersion;
  version.Base.Length := SizeOf(TTLVVersion) - SizeOf(TTlvHeaderBase) + keyAndVersion.Length shl 1 + 4;
  version.Base.TypeItem := csTlvVersion;

  FHash.Reset();
  FHash.Update(version, SizeOf(version));
  FHash.Update(keyAndVersion[1], keyAndVersion.Length shl 1);

  // заголовок
  FBuffer.Write(version, SizeOf(version));

  // данные
  FBuffer.Write(keyAndVersion[1], keyAndVersion.Length shl 1);

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlvEvent(const ASource, AWhoIs, AMsg: UnicodeString; ADateTime: TDateTime; AEventType: int32; const ASourceId: TGuid);
var
  eventHdr: TTLVEvent;
  hash: uint32;
  event: UnicodeString;
begin
  event := ASource + ';' + AWhoIs + ';' + AMsg;
  eventHdr.LenMsg := length(event) shl 1;
  eventHdr.Base.TypeItem := csTlvEvent;
  eventHdr.Base.Length := sizeof(TTLVEvent) - sizeOf(TTlvHeaderBase) + eventHdr.LenMsg + 4;
  eventHdr.DateTime := ADateTime;
  eventHdr.EventType := AEventType;
  eventHdr.SourceId := ASourceId;
  FHash.Reset();
  FHash.Update(eventHdr, SizeOf(eventHdr));
  FHash.Update(event[1], eventHdr.LenMsg);
  hash := FHash.Hash;
  // заголовок
  FBuffer.Write(eventHdr, SizeOf(eventHdr));
  // данные
  FBuffer.Write(event[1], eventHdr.LenMsg);
  // хэш
  FBuffer.Write(hash, SizeOf(hash));
end;

procedure TCinemaNetProtocol.PutTlv(AType: Int32; AData: Int64);
begin
  PutTlv(AType, SizeOf(AData), @AData);
end;

procedure TCinemaNetProtocol.PutTlvAddOrUpdateScenario(const AKey, AName: UnicodeString; ACountInstances: int32);
begin
  PutTlvScenarioAction(csTlvAddOrUpdateScenario, AKey + ';' + AName, ACountInstances);
end;

procedure TCinemaNetProtocol.PutTlvAll(const AKey: UnicodeString);
begin
  PutTlv(csTlvAll, AKey);
end;

procedure TCinemaNetProtocol.PutTlvCinemaNode(const ASourcePath, ADestName: UnicodeString);
begin
  PutTlvFile(ASourcePath, ADestName, csFileCinemaNode);
end;

procedure TCinemaNetProtocol.PutTlvDeleteScenario(const AKey: UnicodeString; ADeleteFile: boolean);
begin
  PutTlvScenarioAction(csTlvDeleteScenario, AKey, IfThen(ADeleteFile, 1, 0));
end;

procedure TCinemaNetProtocol.RegisterTlvProcessor(ATlvType: Int32; AProcessor: TTlvProcessor);
begin
  if ATlvType >= Length(FTlvProcessors) then
    SetLength(FTlvProcessors, ATlvType + 1);

  FTlvProcessors[ATlvType] := AProcessor;
end;

class constructor TCinemaNetProtocol.Create;
begin
  FWorkDirectory := Cinema.Types.csWorkDir;
  FWorkSubDirScens := Cinema.Types.csWorkSubDirScens;
end;

procedure TCinemaNetProtocol.SetKey(const Value: UnicodeString);
begin
  if FKey = Value then
    exit;
  FKey := Value;
end;

procedure TCinemaNetProtocol.PutTlv(AType: Int32; ASize: Int32; AData: PByte);
var
  header: TTlvHeaderBase;
  hash: uint32;
begin
  header.TypeItem := AType;
  header.Length := ASize + 4;

  FHash.Reset();

  // заголовок
  FHash.Update(header, SizeOf(header));
  FBuffer.Write(header, SizeOf(header));

  // данные
  if ASize > 0 then
  begin
    FHash.Update(AData^, ASize);
    FBuffer.Write(AData^, ASize);
  end;

  // хэш
  hash  := FHash.Hash;
  FBuffer.Write(hash, SizeOf(hash));
end;

end.
