﻿unit Cinema.Net.SocketClient;

{$ifdef FPC}
	{$mode delphi}
	{$ifdef posix}
  	{$define fpc_and_posix}
	{$endif}
{$endif}

interface

uses
  Cinema.Net.SocketCommon,
  Cinema.Collections
  ;


type
  TCinemaSocketClient = class;

  TOnClientEventNotify = procedure(AClient: TCinemaSocketClient) of object;

  TCinemaSocketClient = class
  private
    FSocketType: TSocketType;
    FEndPoint: TNetEndpoint;
    FDSetWrite: TCinemaFDSet;
    FPort: Word;
    //FDSetError: TCinemaFDSet;

    FHandle: TSocketHandle;
    FOwnHandle: boolean;
    FIsConnected: boolean;
    FLastStateMessage: string;
    FOnConnect: TOnClientEventNotify;
    FOnDisconnect: TOnClientEventNotify;
    FIsBroadcast: boolean;
    FTimeoutOnConnect: UInt32;

    // этот метод выполняется таской в отдельном потоке
    procedure DoConnect;
    procedure Disconnect;
    procedure SetIsConnected(const Value: boolean);
    procedure SetPort(const Value: Word);
    procedure SetHandle(const Value: TSocketHandle);
  protected
    procedure DoOnConnectEvent(ANewHandle: TSocketHandle); virtual;
    procedure DoOnDisconnectEvent; virtual;
  public
    constructor Create(const AEndPoint: TNetEndpoint; ASocketType: TSocketType; AHandleClient: TSocketHandle = 0);
    destructor Destroy; override;
    procedure Connect;
    function SendData(AData: PByte; ASize: Int64): boolean;
    property Handle: TSocketHandle read FHandle write SetHandle;
    property OwnHandle: boolean read FOwnHandle;
    property IsConnected: boolean read FIsConnected write SetIsConnected;
    property LastStateMessage: string read FLastStateMessage;
    property EndPoint: TNetEndpoint read FEndPoint;
    property Port: Word read FPort write SetPort;
    property TimeoutOnConnect: UInt32 read FTimeoutOnConnect write FTimeoutOnConnect;
    property OnConnect: TOnClientEventNotify read FOnConnect write FOnConnect;
    property OnDisconnect: TOnClientEventNotify read FOnDisconnect write FOnDisconnect;
  end;

implementation

uses
  Classes,
  SysUtils
  ;

const
  csIsConnectedMsg = 'Соединение с сервером установлено';
  csIsDisconnectedMsg = 'Нет подключения';

procedure TCinemaSocketClient.DoConnect;
var
  count: int32;
  handle: TSocketHandle;
begin

  if FHandle = 0 then
  begin
    FOwnHandle := true;
    handle := CreateSocketHandle(FSocketType);
    if handle = InvalidSocket then
      exit;

    SetNonBlockingMode(handle, true);

    SetReuseAddrSocketOption(handle);
    if FIsBroadcast then
    begin
      SetAsBroadcastSocket(handle);
    end;

  end else
    handle := FHandle;

  if not FIsBroadcast then
  begin
    // в неблокирующем режиме SocketConnect и SocketBind всегда возвращают InvalidSocket.
    // Это значит, что процесс коннекта начался успешно. Ниже успех подключения
    // проверяется через SocketsSelect
    if FSocketType in [TSocketType.TCP, TSocketType.RAW, TSocketType.RDM, TSocketType.SEQ] then
      SocketConnect(handle, FEndpoint.Sockaddr, SizeOf(FEndpoint.Sockaddr))
    else
      SocketBind(handle, FEndpoint.Sockaddr, SizeOf(FEndpoint.Sockaddr));
  end;

  SetSendTimeout(handle, FTimeoutOnConnect);
  // оставляем не блокирующий режим, т.к. заданный выше таймаут на send не отработает, только в неблокируюм режиме
  // иначе сокет может уйти в бесконечную блокировку при нестабильной связи
  // SetNonBlockingMode(handle, false);

  FDSetWrite.Clear;
  FDSetWrite.SetFds(@handle, 1);
  count := SocketsSelect(FDSetWrite.MaxHandle + 1, nil, FDSetWrite.FdSetPtr, nil, FTimeOutOnConnect);
  if count <= 0 then
    CloseSocketHandle(handle, FSocketType, nil, 0)
  else
  begin
    DoOnConnectEvent(handle);
  end;
end;

procedure TCinemaSocketClient.DoOnConnectEvent(ANewHandle: TSocketHandle);
var
  buf: array[0..255] of byte;
begin
  if FOwnHandle and (FHandle > 0) then
    CloseSocketHandle(FHandle, FSocketType, @buf[0], SizeOf(buf));

  FHandle := ANewHandle;
  FIsConnected := true;
  FLastStateMessage := csIsConnectedMsg;

  if Assigned(FOnConnect) then
    FOnConnect(Self);
end;

procedure TCinemaSocketClient.DoOnDisconnectEvent;
var
  buf: array[0..255] of byte;
begin
  FIsConnected := false;

  if Assigned(FOnDisconnect) then
    FOnDisconnect(Self);

  if FOwnHandle and (FHandle > 0) then
  begin
    CloseSocketHandle(FHandle, FSocketType, @buf[0], SizeOf(buf));
    FHandle := 0;
  end;

  FLastStateMessage := csIsDisconnectedMsg;
end;

constructor TCinemaSocketClient.Create(const AEndPoint: TNetEndpoint; ASocketType: TSocketType; AHandleClient: TSocketHandle = 0);
begin
  FTimeoutOnConnect := 10000;
  FDSetWrite := TCinemaFDSet.Create(1);
  //FDSetError := TCinemaFDSet.Create(1);
  FHandle := AHandleClient;
  FEndPoint := AEndPoint;
  FPort := FEndPoint.Port;
  FSocketType := ASocketType;
  FOwnHandle := FHandle = 0;
  FIsConnected := FHandle > 0;
  if FIsConnected then
    FLastStateMessage := csIsConnectedMsg
  else
    FLastStateMessage := csIsDisconnectedMsg;

  FIsBroadcast := (FEndPoint.Address.Address.Length > 3) and (Copy(FEndPoint.Address.Address, FEndPoint.Address.Address.Length - 2, 3) = '255');

end;

destructor TCinemaSocketClient.Destroy;
begin

  if FIsConnected then
    Disconnect;

  FDSetWrite.Free;
  //FDSetError.Free;
  inherited;
end;

procedure TCinemaSocketClient.Disconnect;
begin
  if not FIsConnected and (FHandle = 0) then
    exit;

  DoOnDisconnectEvent;
end;

procedure TCinemaSocketClient.Connect;
begin
  DoConnect;
end;

function TCinemaSocketClient.SendData(AData: PByte; ASize: Int64): boolean;
var
  count: Int64;
begin
  if not FIsConnected then
    Connect;

  if (FHandle = 0) or not FIsConnected or (ASize <= 0) then
    exit(false);

// Не делай больше так никогда
//  FDSetError.Clear;
//  FDSetWrite.Clear;
//  FDSetWrite.SetFds(@FHandle, 1);
//  count := SocketsSelect(FDSetWrite.MaxHandle + 1, nil, FDSetWrite.FdSetPtr, FDSetError.FdSetPtr, 0);
//
//  if (count < 0) {$ifdef posix} or FDSetError.FdIsSet(FHandle) {$else} or (FDSetError.CountEvent > 0) {$endif} then
//  begin
//    Disconnect;
//    exit(false);
//  end;
//
//  if count = 0 then
//    exit(false);

  if FSocketType = UDP then
    count := SocketSendTo(FHandle, FEndPoint, AData, ASize, 0)
  else
    count := SocketSend(FHandle, AData, ASize);

  if count <= 0 then
  begin
    Result := false;
    DoOnDisconnectEvent;
  end else
  begin
    Result := count = ASize;
  end;

end;

procedure TCinemaSocketClient.SetHandle(const Value: TSocketHandle);
begin
  if FHandle = Value then
    exit;

  if FIsConnected then
    Disconnect;

  FHandle := Value;
  FOwnHandle := FHandle = 0;
end;

procedure TCinemaSocketClient.SetIsConnected(const Value: boolean);
begin
  if FIsConnected = Value then
    exit;
  if Value then
    Connect
  else
    Disconnect;
end;

procedure TCinemaSocketClient.SetPort(const Value: Word);
begin
  if FPort = Value then
    exit;

  if FIsConnected then
    Disconnect;

  FPort := Value;
  FEndPoint.Port := Value;

end;

end.
