unit Cinema.Serialization;

{ Original sources is fpjsonrtti }

{$mode Delphi}

interface

uses
  Classes, SysUtils, contnrs, typinfo, fpjson, rttiutils, jsonparser, rtti, Generics.Collections;

Const
  RFC3339DateTimeFormat = 'yyyy"-"mm"-"dd"T"hh":"nn":"ss';
  RFC3339DateTimeFormatMsec = RFC3339DateTimeFormat+'.zzz';


type
  TJSONStreamOption = (jsoStreamChildren,         // If set, children will be streamed in 'Children' Property
                       jsoEnumeratedAsInteger,    // Write enumerated as integer. Default is string.
                       jsoSetAsString,            // Write Set as a string. Default is an array.
                       jsoSetEnumeratedAsInteger, // Write enumerateds in set array as integers.
                       jsoSetBrackets,            // Use brackets when creating set as array
                       jsoComponentsInline,       // Always stream components inline. Default is to stream name, unless csSubcomponent in ComponentStyle
                       jsoTStringsAsArray,        // Stream TStrings as an array of strings. Associated objects are not streamed.
                       jsoTStringsAsObject,       // Stream TStrings as an object : string = { object }
                       jsoDateTimeAsString,       // Format a TDateTime value as a string
                       jsoUseFormatString,        // Use FormatString when creating JSON strings.
                       jsoCheckEmptyDateTime,     // If TDateTime value is empty and jsoDateTimeAsString is used, 0 date returns empty string
                       jsoLegacyDateTime,         // Set this to enable old date/time formatting. Current behaviour is to save date/time as a ISO 9601 value.
                       jsoLowerPropertyNames,     // Set this to force lowercase names when streaming to JSON.
                       jsoStreamTList             // Set this to assume that TList contains a list of TObjects. Use with care!
                       );
  TJSONStreamOptions = set of TJSONStreamOption;


  { TJSONStreamer }

  TJSONStreamer = class
  private
    FDateTimeFormat: String;
    FOptions: TJSONStreamOptions;
    function ObjectFromString(const JSON: TJSONStringType): TJSONData;
    function WriteMember(AProperty: TRttiProperty; const AValue: TValue): TJSONData;
    procedure WriteMembers(AObject: TObject; AJsonObject: TJsonObject);
    procedure ReadMember(AObject: TObject; AProperty: TRttiProperty; APropData: TJSONData);
    procedure ReadMembers(AObject: TObject; AJsonObject: TJsonObject);
    function GetObjectPropsDic(ATypeInfo: PTypeInfo): TDictionary<string, TRttiProperty>;
  protected
    function StreamClassProperty(AObject: TObject): TJSONData; virtual;
    //Function StreamProperty(const AObject : TObject; const PropertyName : string) : TJSONData; overload;
    //function StreamProperty(const AObject: TObject; APropertyInfo: PPropInfo): TJSONData; overload;
    Function FormatDateProp(DateTime : TDateTime) : TJSONString;
  public
    Constructor Create;
    Destructor Destroy;override;
    function ObjectToJSON(AObject: TObject) : TJSONObject;
    procedure JSONToObject(AObject: TObject; AJsonObject: TJSONObject); overload;
    procedure JSONToObject(AObject: TObject; const AJsonData: WideString); overload;
    function Deserialize<T: class, constructor>(const AJsonDto: WideString): T;
    function Serialize<T: TObject>(AObject: T; AWrapIntoClassNameField: boolean = false): WideString;
    // Stream a collection - always returns an array
    function StreamCollection(ACollection: TCollection): TJSONArray;
    // Stream a collection - always returns an array
    //function StreamCollection(const AArray: Variant): TJSONArray;
    // Stream an objectlist - always returns an array
    function StreamObjectList(AnObjectList: TObjectList): TJSONArray;
    // Stream a List - always returns an array
    function StreamTList(AList: TList): TJSONArray;
    // Stream a TStrings instance as an array
    function StreamTStringsArray(AStrings: TStrings): TJSONArray;
    // Stream a TStrings instance as an object
    function StreamTStringsObject(AStrings: TStrings): TJSONObject;
    // Stream a TStrings instance. Takes into account Options.
    function StreamTStrings(AStrings: TStrings): TJSONData;
    //
    // Some utility functions.
    //
    // Call ObjectToJSON and convert result to JSON String.
    Function ObjectToJSONString(AObject : TObject) : TJSONStringType;
    // Convert TSTrings to JSON string with array or Object.
    Function StringsToJSON(Const Strings : TStrings; AsObject : Boolean = False) : TJSONStringType;
    // Convert collection to JSON string
    Function CollectionToJSON(Const ACollection : TCollection) : TJSONStringType;
  published
    // Format used when formatting DateTime values. Only used in conjunction with jsoDateTimeToString
    Property DateTimeFormat : String Read FDateTimeFormat Write FDateTimeFormat;
    // Options to use when streaming
    Property Options : TJSONStreamOptions Read FOptions Write FOptions;
  end;

  EJSONRTTI = Class(Exception);

implementation

uses dateutils, variants, rtlconsts;

ResourceString
  SErrUnknownPropertyKind     = 'Unknown property kind for property : "%s"';
  SErrUnsupportedPropertyKind = 'Unsupported property kind for property: "%s"';
  SErrUnsupportedVariantType  = 'Unsupported variant type : %d';
  SErrUnsupportedArrayType    = 'JSON array cannot be streamed to object of class "%s"';
  SErrUnsupportedJSONType     = 'Cannot destream object from JSON data of type "%s"';
  SErrUnsupportedCollectionType = 'Unsupported JSON type for collections: "%s"';
  SErrUnsupportedCollectionItemType = 'Array element %d is not a valid type for a collection item: "%s"';
  SErrUnsupportedStringsItemType = 'Array element %d is not a valid type for a stringlist item: "%s"';
  SErrUnsupportedStringsType = 'Unsupported JSON type for stringlists: "%s"';
  SErrUnsupportedStringsObjectType = 'Object Element %s is not a valid type for a stringlist object: "%s"';
  SErrUnSupportedEnumDataType = 'Unsupported JSON type for enumerated property "%s" : "%s"';
  SErrUnsupportedVariantJSONType = 'Unsupported JSON type for variant value : "%s"';
  SErrUnsupportedObjectData = 'Unsupported JSON type for object property: "%s"';

type
  TSet = set of 0..31; // Used to (de)stream set properties.

{ TJSONStreamer }

constructor TJSONStreamer.Create;
begin
  Inherited;
end;

destructor TJSONStreamer.Destroy;
begin
  Inherited;
end;

function TJSONStreamer.WriteMember(AProperty: TRttiProperty; const AValue: TValue): TJSONData;
var
  arr: TJSONArray;
  i: Integer;
  dt: TDateTime;
  sval: string;
  svalu: Unicodestring;
begin
  Result:=Nil;
  If AValue.IsEmpty then
    Exit(TJSONNull.Create);

  case AValue.Kind of
    tkAString,
    tkUString,
    tkWString,
    tkSString:
      Result := TJSONString.Create(AValue.AsUnicodeString);
    tkInteger:
      Result := TJSONIntegerNumber.Create(AValue.AsInteger);
    tkInt64,
    tkQWord:
      Result := TJSONInt64Number.Create(AValue.AsInt64);
    tkBool:
      Result := TJSONBoolean.Create(AValue.AsBoolean);
    tkFLoat:
      begin
       if (AValue.TypeInfo = System.TypeInfo(TDateTime)) or
          (AValue.TypeInfo = System.TypeInfo(TDate)) or
          (AValue.TypeInfo = System.TypeInfo(TTime)) then
          begin
            dt := AValue.AsType<TDateTime>;
            if dt <> 0 then
            begin
              //var LAttr: DateSerializeFormatAttribute;
              //if AProperty.TryGetCustomAttribute<DateSerializeFormatAttribute>(LAttr) then
              //  Result := TJSONString.Create(LAttr.DateTimeToStr(dt))
              //else
                Result := TJSONString.Create(DateToISO8601(dt, False));
            end;
          end
       else
         Result := TJSONFloatNumber.Create(AValue.AsExtended);
      end;

    tkClass:
      if AValue.IsObject then
      begin
        Result := TJSONObject.Create;
        WriteMembers(AValue.AsObject, TJsonObject(Result));
      end;

    tkArray, tkDynArray: begin
      Result := TJSONArray.Create;
      for i := 0 to AValue.GetArrayLength - 1 do
        TJSONArray(Result).Add(WriteMember(AProperty, AValue.GetArrayElement(i)));
    end else
      Raise EJSONRTTI.CreateFmt(SErrUnsupportedVariantType,[AValue.Kind]);
  end;
end;

procedure TJSONStreamer.WriteMembers(AObject: TObject; AJsonObject: TJsonObject);
var
  i: int32;
  LType: TRttiType;
  LPropValue: TValue;
  LProps: TRttiPropertyArray;
  LProp: TRttiProperty;
  LJsonData: TJsonData;
begin
  LType := TRTTIContext.Create.GetType(AObject.ClassType);
  LProps := LType.AsInstance.GetProperties;
  for i := 0 to length(LProps) - 1 do
  begin
    LProp := LProps[i];
    LPropValue := LProp.GetValue(AObject);
    LJsonData := WriteMember(LProp, LPropValue);
   if Assigned(LJsonData) then
     AJsonObject.Add(LProp.Name, LJsonData);
  end;
end;

procedure TJSONStreamer.ReadMember(AObject: TObject; AProperty: TRttiProperty; APropData: TJSONData);
var
  PI: PPropInfo;
  TI: PTypeInfo;
  TD: PTypeData;
  I,J,S : Integer;
  arr: TJSONArray;
  JS : TJSONStringType;
  LDateTime: TDateTime;
  LObject: TObject;
  LArray: TArray<TValue>;
begin
  PI := AProperty.Handle;
  TI := AProperty.PropertyType.Handle;
  case TI^.Kind of
    tkUnknown:
      Raise EJSONRTTI.CreateFmt(SErrUnknownPropertyKind,[PI^.Name]);

    tkInteger:
      SetOrdProp(AObject,PI,APropData.AsInteger);

    tkInt64:
      SetOrdProp(AObject,PI,APropData.AsInt64);

    tkEnumeration: begin
      if (APropData.JSONType=jtNumber) then
        I := APropData.AsInteger
      else if APropData.JSONType=jtString then
        I := GetEnumValue(TI,APropData.AsString)
      else
        raise EJSONRTTI.CreateFmt(SErrUnSupportedEnumDataType,[PI^.Name,GetEnumName(TypeInfo(TJSONType),Ord(APropData.JSONType))]);
      SetOrdProp(AObject,PI,I);
    end;

    tkFloat: begin
        if (TI = System.TypeInfo(TDateTime)) or (TI = System.TypeInfo(TDate)) or (TI = System.TypeInfo(TTime)) then
        begin
          //var LDateTime: TDateTime;
          //var LAttr: DateSerializeFormatAttribute;
          //if LProperty.TryGetCustomAttribute<DateSerializeFormatAttribute>(LAttr) then
          //  LDateTime := LAttr.StrToDateTime(LJSONValue.AsType<string>)
          //else
            LDateTime := ISO8601ToDate(APropData.AsString, False);
          AProperty.SetValue(AObject, LDateTime);
        end
        else
          AProperty.SetValue(AObject, APropData.AsFloat);
      //if (TI=TypeInfo(TDateTime)) and (APropData.JSONType=jtString) then
      //  SetFloatProp(AObject,PI,ExtractDateTime(APropData.AsString))
      //else
      //  SetFloatProp(AObject,PI,APropData.AsFloat)
      end;
    tkSet :
      If APropData.JSONType=jtString then
        SetSetProp(AObject,PI,APropData.AsString)
      else
      if (APropData.JSONType=jtArray) then
      begin
        arr := APropData as TJSONArray;
        TI := GetTypeData(TI)^.CompType;
        S := 0;
        for I := 0 to arr.Count-1 do
        begin
          if arr.types[i] = jtNumber then
            J := arr.Integers[i]
          else
            J:= GetEnumValue(TI, arr.strings[i]);
          TSet(S) := TSet(S)+[j];
        end;
        SetOrdProp(AObject, PI, S);
      end;
    tkChar: begin
      JS:=APropData.AsString;
      If (JS<>'') then
        SetOrdProp(AObject,PI,Ord(JS[1]));
    end;
    tkSString,
    tkLString,
    tkAString:
      SetStrProp(AObject,PI,APropData.AsString);
    tkWString :
      SetWideStrProp(AObject,PI,APropData.AsUnicodeString);
    //tkVariant:
    //  SetVariantProp(AObject, PI, JSONToVariant(APropData));

    tkClass: begin
      if not AProperty.GetValue(AObject).IsEmpty then   // Пропускаем не созданные объекты
        ReadMembers(AProperty.GetValue(AObject).AsObject, TJsonObject(APropData));
    end;

    tkWChar: begin
      JS:=APropData.asString;
      If (JS<>'') then
        SetOrdProp(AObject,PI,Ord(JS[1]));
    end;

    tkBool:
      SetOrdProp(AObject,PI,Ord(APropData.AsBoolean));

    tkQWord:
      SetPropValue(AObject,PI, APropData.AsQWord);

    tkDynArray: begin
      arr := APropData as TJSONArray;
      TD := GetTypeData(TI);
      SetLength(LArray, arr.Count);

      for i := 0 to arr.Count-1 do
      begin
        if (((TD)^.ParentInfoRef)^)^.Kind = TTypeKind.tkClass then
        begin
          LObject := GetTypeData((((TD)^.ParentInfoRef)^)).ClassType.Create;
          LArray[i] := LObject;
          ReadMembers(LObject, TJsonObject(APropData.Items[i]));
        end
        else
        begin
          //LValVar := APropData.Items[i].Value;
          //TValue.Make(@LValVar, ((TD)^.ParentInfoRef)^, LArray[i]);
        end;
        //ReadMember
      end;
      if Length(LArray) > 0 then
      begin
        AProperty.SetValue(AObject, TValue.FromArray(TI, LArray));
      end;
    end;
    tkObject: begin

    end;
    tkArray,
    tkRecord,
    tkInterface,
    tkInterfaceRaw,
    tkProcVar,
    tkMethod :
      Raise EJSONRTTI.CreateFmt(SErrUnsupportedPropertyKind,[PI^.Name]);
    tkUString :
      SetUnicodeStrProp(AObject,PI,APropData.AsUnicodeString);
    tkUChar:
      begin
      JS:=APropData.asString;
      If (JS<>'') then
        SetOrdProp(AObject,PI,Ord(JS[1]));
      end;
  end;
end;

procedure TJSONStreamer.ReadMembers(AObject: TObject; AJsonObject: TJsonObject);
var
  i: int32;
  LType: TRttiType;
  LProps: TRttiPropertyArray;
  LProp: TRttiProperty;
  LJsnData: TJSONData;
  LPropsDic: TDictionary<string, TRttiProperty>;
  LFieldName: string;
begin
  LPropsDic := TDictionary<string, TRttiProperty>.Create;
  try
    LType := TRTTIContext.Create.GetType(AObject.ClassType);
    LProps := LType.AsInstance.GetProperties;
    for i := 0 to length(LProps) - 1 do
    begin
      LProp := LProps[i];
      LPropsDic.Add(LProp.Name, LProp);
    end;
    for i := 0 to AJsonObject.Count - 1 do
    begin
      LFieldName := AJsonObject.Names[i];
      if not LPropsDic.TryGetValue(LFieldName, LProp) then
        continue;
      LJsnData := AJsonObject.Items[i];
      ReadMember(AObject, LProp, LJsnData);
    end;
  finally
    LPropsDic.Free;
  end;
end;

function TJSONStreamer.GetObjectPropsDic(ATypeInfo: PTypeInfo): TDictionary<string, TRttiProperty>;
var
  i: int32;
  LType: TRttiType;
  LPropValue: TValue;
  LProps: TRttiPropertyArray;
  LProp: TRttiProperty;
begin
  Result := TDictionary<string, TRttiProperty>.Create;
  LType := TRTTIContext.Create.GetType(ATypeInfo);
  LProps := LType.AsInstance.GetProperties;
  for i := 0 to length(LProps) - 1 do
  begin
    LProp := LProps[i];
    Result.Add(LProp.Name, LProp);
  end;
end;

function TJSONStreamer.ObjectFromString(const JSON: TJSONStringType): TJSONData;
begin
  With TJSONParser.Create(JSON,[]) do
    try
      Result := Parse;
    finally
      Free;
    end;
end;

function TJSONStreamer.ObjectToJSON(AObject: TObject): TJSONObject;
begin

  if AObject = nil then
    Exit;

  Result := TJSONObject.Create;
  try
    WriteMembers(AObject, Result);
  except
    FreeAndNil(Result);
    Raise;
  end;
end;

procedure TJSONStreamer.JSONToObject(AObject: TObject; AJsonObject: TJSONObject);
begin
  ReadMembers(AObject, AJsonObject);
end;

procedure TJSONStreamer.JSONToObject(AObject: TObject; const AJsonData: WideString);
var
  data: TJSONData;
begin
  data := ObjectFromString(UTF8Encode(AJsonData));
  try
    if Assigned(data) and (data.InheritsFrom(TJSONObject)) then
      JSONToObject(AObject, TJSONObject(data));
  finally
    data.Free;
  end;
end;

function TJSONStreamer.Deserialize<T>(const AJsonDto: WideString): T;
begin
  Result := T.Create;
  JSONToObject(Result, AJsonDto);
end;

function TJSONStreamer.Serialize<T>(AObject: T; AWrapIntoClassNameField: boolean): WideString;
begin
  Result := UTF8Decode(ObjectToJSONString(AObject));
end;

function TJSONStreamer.ObjectToJSONString(AObject: TObject): TJSONStringType;
var
  O: TJSONData;
begin
  O := ObjectToJSON(AObject);
  try
    if (jsoUseFormatString in Options) then
      Result:=O.FormatJSON()
    else
      Result:=O.AsJSON;
  finally
    FreeAndNil(O);
  end;
end;

function TJSONStreamer.StringsToJSON(const Strings: TStrings; AsObject: Boolean): TJSONStringType;
var
  D: TJSONData;
begin
  If ASObject then
    D :=StreamTSTringsObject(Strings)
  else
    D :=StreamTStringsArray(Strings);
  try
    if (jsoUseFormatString in Options) then
      Result := D.FormatJSON
    else
      Result := D.AsJSON;
  finally
    FreeAndNil(D);
  end;
end;

function TJSONStreamer.CollectionToJSON(const ACollection: TCollection): TJSONStringType;
var
  D: TJSONArray;
begin
  D := StreamCollection(ACollection);
  try
    if (jsoUseFormatString in Options) then
      Result := D.FormatJSON()
    else
    Result := D.AsJSON;
  finally
    FreeAndNil(D);
  end;
end;

function TJSONStreamer.StreamTList(AList: TList): TJSONArray;
var
  i: Integer;
  o: TJSONObject;
begin
  Result := TJSONArray.Create;
  try
    for i := 0 to AList.Count - 1 do begin
      o := ObjectToJSON(TObject(AList.Items[i]));
      if Assigned(o) then
        Result.Add(o);
    end;
  except
    FreeAndNil(Result);
    Raise;
  end;
end;

function TJSONStreamer.StreamTStringsArray(AStrings: TStrings): TJSONArray;
var
  i: Integer;
begin
  Result := TJSONArray.Create;
  try
    For i := 0 to AStrings.Count - 1 do
      Result.Add(AStrings[i]);
  except
    FreeAndNil(Result);
    Raise;
  end;
end;

function TJSONStreamer.StreamTStringsObject(AStrings: TStrings): TJSONObject;
var
  I: Integer;
  O: TJSONData;
begin
  Result:=TJSONObject.Create;
  try
    For I:=0 to AStrings.Count-1 do
      begin
      O:=ObjectToJSON(AStrings.Objects[i]);
      If O=Nil then
        O:=TJSONNull.Create;
      Result.Add(AStrings[i],O);
      end;
  except
    FreeAndNil(Result);
    Raise;
  end;
end;

function TJSONStreamer.StreamTStrings(AStrings: TStrings): TJSONData;
begin
  If jsoTStringsAsArray in Options then
    Result:=StreamTStringsArray(AStrings)
  else If jsoTStringsAsObject in Options then
    Result:=StreamTStringsObject(AStrings)
  else
    Result:=TJSONString.Create(AStrings.Text);
end;


function TJSONStreamer.StreamCollection(ACollection: TCollection): TJSONArray;
var
  I: Integer;
begin
  Result:=TJSONArray.Create;
  try
    For I:=0 to ACollection.Count-1 do
      Result.Add(ObjectToJSON(ACollection.Items[i]));
  except
    FreeAndNil(Result);
    Raise;
  end;
end;

function TJSONStreamer.StreamObjectList(AnObjectList: TObjectList): TJSONArray;
var
  i: Integer;
begin
  if not Assigned(AnObjectList) then
    Result := nil;
  Result:=TJSONArray.Create;
  try
    For i:=0 to AnObjectList.Count-1 do
      Result.Add(ObjectToJSON(AnObjectList.Items[i]));
  except
    FreeAndNil(Result);
    Raise;
  end;
end;

function TJSONStreamer.StreamClassProperty(AObject: TObject): TJSONData;
begin
  Result:=Nil;
  If (AObject=Nil) then
    Result:=TJSONNull.Create()
  else if (AObject is TComponent) then
    begin
    if (csSubComponent in TComponent(AObject).ComponentStyle) or (jsoComponentsInline in Options) then
      Result:=ObjectToJSON(AObject)
    else
      Result:=TJSONString.Create(TComponent(AObject).Name);
    end
  else if (AObject is TStrings) then
    Result:=StreamTStrings(TStrings(AObject))
  else if (AObject is TCollection) then
    Result:=StreamCollection(TCollection(Aobject))
  else If AObject is TObjectList then
    Result:=StreamObjectList(TObjectList(AObject))
  else // Normally, this is only TPersistent.
    Result:=ObjectToJSON(AObject);
end;

//function TJSONStreamer.StreamProperty(const AObject: TObject; APropertyInfo: PPropInfo): TJSONData;
//type
//  TArrStr = TArray<string>;
//var
//  pt: PTypeInfo;
//  td: PTypeData;
//  s, i: Integer;
//  v: Variant;
//  val: TValue;
//  data: Pointer;
//begin
//  Result := nil;
//  case APropertyInfo^.PropType^.Kind of
//    tkUnknown :
//      Error(SErrUnknownPropertyKind,[APropertyInfo^.PropType^.Name]);
//    tkInteger :
//      Result:=TJSONIntegerNumber.Create(GetOrdProp(AObject,APropertyInfo));
//    tkEnumeration :
//      if jsoEnumeratedAsInteger in Options then
//        Result:=TJSONIntegerNumber.Create(GetOrdProp(AObject, APropertyInfo))
//      else
//        Result:=TJSONString.Create(GetEnumName(APropertyInfo^.PropType,GetOrdProp(AObject,APropertyInfo)));
//    tkFloat :
//      if (APropertyInfo^.PropType = TypeInfo(TDateTime)) and (jsoDateTimeAsString in Options) then
//        Result:=FormatDateProp(GetFloatProp(AObject,APropertyInfo))
//      else
//        Result:=TJSONFloatNumber.Create(GetFloatProp(AObject, APropertyInfo));
//    tkSet :
//      If jsoSetAsString in Options then
//        Result:=TJSONString.Create(GetSetProp(AObject, APropertyInfo, jsoSetBrackets in Options))
//      else
//      begin
//        pt := GetTypeData(APropertyInfo.PropType)^.CompType;
//        s := GetOrdProp(AObject, APropertyInfo);
//        Result := TJSONArray.Create;
//        try
//          for i := 0 to 31 do
//            if (i in TSet(s)) then
//              if jsoSetEnumeratedAsInteger in Options then
//                TJSONArray(Result).Add(i)
//              else
//                TJSONArray(Result).Add(GetEnumName(pt, i));
//        except
//          FreeAndNil(Result);
//          Raise;
//        end;
//      end;
//    tkChar:
//      Result:=TJSONString.Create(AnsiChar(GetOrdProp(AObject, APropertyInfo)));
//    tkSString,
//    tkLString,
//    tkAString:
//      Result := TJSONString.Create(GetStrProp(AObject, APropertyInfo));
//    tkWString:
//      Result := TJSONString.Create(GetWideStrProp(AObject, APropertyInfo));
//    //tkVariant:
//    //  Result := StreamVariant(GetVariantProp(AObject, APropertyInfo));
//    tkClass:
//      Result := StreamClassProperty(GetObjectProp(AObject, APropertyInfo));
//    tkWChar:
//      Result := TJSONString.Create(WideChar(GetOrdProp(AObject, APropertyInfo)));
//    tkBool:
//      Result := TJSONBoolean.Create(GetOrdProp(AObject, APropertyInfo)<>0);
//    tkInt64:
//      Result :=TJSONInt64Number.Create(GetOrdProp(AObject, APropertyInfo));
//    tkQWord:
//      Result := TJSONFloatNumber.Create(GetOrdProp(AObject, APropertyInfo));
//    tkObject:
//      Result := ObjectToJSON(GetObjectProp(AObject, APropertyInfo));
//    //tkArray:  //GetObjectProp
//    //  Result := StreamVariant(GetVariantProp(AObject, APropertyInfo));
//    tkDynArray: begin
//      td := GetTypeData(APropertyInfo.PropType);
//      //TValue.MakeOpenArray(GetDynArrayProp(AObject, APropertyInfo), td.ArrayData.Size, APropertyInfo.PropType, val);
//      data := GetDynArrayProp(AObject, APropertyInfo);
//      TValue.Make<TArrStr>(TArrStr(data), val);
//
//
//      DynArrayToVariant(v, GetDynArrayProp(AObject, APropertyInfo), (((APropertyInfo)^.PropTypeRef)^));
//      Result := StreamVariant(v);
//    end;
//      //Result := TJSONArray.Create;
//    tkRecord,
//    tkInterface,
//    tkInterfaceRaw,
//    tkProcVar,
//    tkMethod :
//      Error(Format(SErrUnsupportedPropertyKind,[APropertyInfo^.PropType.Name]));
//    tkUString :
//      Result:=TJSONString.Create(GetWideStrProp(AObject,APropertyInfo));
//    tkUChar:
//      Result:=TJSONString.Create(UnicodeChar(GetOrdProp(AObject,APropertyInfo)))
//    else
//      Error(SErrUnsupportedPropertyKind,[APropertyInfo^.PropType.Name]);
//  end;
//  If Assigned(FOnStreamProperty) then
//    FOnStreamProperty(Self, AObject, APropertyInfo, Result);
//end;

function TJSONStreamer.FormatDateProp(DateTime: TDateTime): TJSONString;
var
  S: String;
begin
  if (jsoCheckEmptyDateTime in Options) and (DateTime=0) then
    S:=''
  else if (DateTimeFormat<>'') then
    S:=FormatDateTime(DateTimeFormat,DateTime)
  else if (jsoLegacyDateTime in options) then
    begin
    if Frac(DateTime)=0 then
      S:=DateToStr(DateTime)
    else if Trunc(DateTime)=0 then
      S:=TimeToStr(DateTime)
    else
      S:=DateTimeToStr(DateTime);
    end
  else
    S:=FormatDateTime(RFC3339DateTimeFormat,DateTime);

  Result:=TJSONString.Create(S);
end;

end.

