﻿unit Cinema.FPC.Threading;

{$ifdef fpc}
  {$mode Delphi}{$H+}
{$endif}

interface

uses
  Classes, SysUtils, Generics.Collections, Cinema.Collections, syncobjs;

type

  TTaskBody = procedure of object;

  TTaskStatus = (Created, WaitingToRun, Running, WaitingForChildren, Canceled, Completed, Exception);

  { ITask }

 ITask = interface
  ['{5F66D93A-6247-47A5-8727-D873F2F53A52}']
		function GetTaskBody: TTaskBody;
    function GetStatus: TTaskStatus;
  	procedure SetStatus(AValue: TTaskStatus);
    procedure Start;
    procedure Cancel;
    procedure Wait(ATimeOut: uint32);
  	property TaskBody: TTaskBody read GetTaskBody;
    property Status: TTaskStatus read GetStatus write SetStatus;
  end;

  { TTask }

  TTask = class(TInterfacedObject, ITask)
  private
    FTaskBody: TTaskBody;
    FStatus: TTaskStatus;
  	function GetTaskBody: TTaskBody;
    function GetStatus: TTaskStatus;
    procedure SetStatus(AValue: TTaskStatus);
    procedure Cancel;
  public
    constructor Create(ATaskBody: TTaskBody);
    procedure Start;
    procedure Wait(ATimeOut: uint32);
    class function Run(ATaskBody: TTaskBody): ITask;
    property TaskBody: TTaskBody read GetTaskBody;
    property Status: TTaskStatus read GetStatus write SetStatus;
  end;

  { TThreading }

  TThreading = class
  public
    type

  		TQueueTasks = TQueueFIFO<ITask>;

      { TWorker }

    	TWorker = class(TThread)
      private
      	FQueue: TQueueTasks;
        FCriticalSection: TCriticalSection;
        FEvent: TSimpleEvent;
        procedure SetWaiting(AValue: Boolean);
        procedure OnWrite(ASource: TQueueTasks);
      protected
      	procedure Execute; override;
        procedure TerminatedSet; override;
      public
    		constructor Create;
      	destructor Destroy; override;
      	procedure PutTask(const ATask: ITask);
        function RunTask(ATaskBody: TTaskBody): ITask;
      end;

  private
  	class var FPoolSize: int32;
  	class var FPull: TList<TWorker>;
  	class var FWorkerIndex: int32;

    class function GetWorker: TWorker;

  	class constructor Create;
    class destructor Destroy;
  public
  	class function RunTask(ATaskBody: TTaskBody): ITask; overload;
    class procedure RunTask(const ATask: ITask); overload;
    class function CreateWorker: TWorker;
    class property PoolSize: int32 read FPoolSize write FPoolSize;
  end;

implementation

{ TTask }

function TTask.GetTaskBody: TTaskBody;
begin
  Result := FTaskBody;
end;

class function TTask.Run(ATaskBody: TTaskBody): ITask;
begin
  Result := TThreading.RunTask(ATaskBody);
end;

function TTask.GetStatus: TTaskStatus;
begin
	Result := FStatus;
end;

procedure TTask.SetStatus(AValue: TTaskStatus);
begin
  FStatus := AValue;
end;

procedure TTask.Cancel;
begin
  if (integer(FStatus) < integer(TTaskStatus.Running)) then
  	FStatus := TTaskStatus.Canceled;
end;

constructor TTask.Create(ATaskBody: TTaskBody);
begin
	FTaskBody := ATaskBody;
end;

procedure TTask.Start;
begin
  FStatus := TTaskStatus.WaitingToRun;
	TThreading.RunTask(Self);
end;

procedure TTask.Wait(ATimeOut: uint32);
var
	timeStart: uint64;
begin
  timeStart := TThread.GetTickCount64;
  while (FStatus < TTaskStatus.Completed) and (TThread.GetTickCount64 - timeStart < ATimeOut) do
  begin
    // TODO: через ивенты?
  	Sleep(1);
  end;
end;

{ TThreading.TWorker }

procedure TThreading.TWorker.SetWaiting(AValue: Boolean);
begin
  if AValue then
  begin
    { рабочий поток сервиса засыпает }

    if not FCriticalSection.TryEnter then
      Exit;
    try
      FEvent.ResetEvent;
    finally
      FCriticalSection.Leave;
    end;

    FEvent.WaitFor(INFINITE);

  end else
  begin
    { рабочий поток сервиса пробуждается }

    FCriticalSection.Enter;
    try
      FEvent.SetEvent;
    finally
      FCriticalSection.Leave;
    end;

  end;
end;

procedure TThreading.TWorker.OnWrite(ASource: TQueueTasks);
begin
  SetWaiting(False);
end;

procedure TThreading.TWorker.Execute;
var
  task: ITask;
begin
	repeat
    while FQueue.Count > 0 do
    begin
			if FQueue.Pop(task) then
      begin

        if task.Status = TTaskStatus.Canceled then
        begin
          task.Status := TTaskStatus.Completed;
          task := nil;
        	continue;
      	end;

        try
        	task.Status := TTaskStatus.Running;
      		task.TaskBody();
          task.Status := TTaskStatus.Completed;
        except
          task.Status := TTaskStatus.Exception;
        end;

        task := nil;
      end else
      	break;
    end;

		if (FQueue.Count = 0) and not Terminated then
    	SetWaiting(true);

  until Terminated;
end;

procedure TThreading.TWorker.TerminatedSet;
begin
	inherited TerminatedSet;
  SetWaiting(false);
end;

constructor TThreading.TWorker.Create;
begin
  inherited Create(false);
  FQueue := TQueueTasks.Create;
  FQueue.OnWrite := OnWrite;
  FCriticalSection := TCriticalSection.Create;
  FEvent := TSimpleEvent.Create;
end;

destructor TThreading.TWorker.Destroy;
begin
  inherited Destroy;
  FQueue.Free;
  FCriticalSection.Free;
  FEvent.Free;
end;

procedure TThreading.TWorker.PutTask(const ATask: ITask);
begin
  // на случай если сервисом будут пользоваться из множества потоков
  FCriticalSection.Enter;
  try
	  FQueue.Push(ATask);
  finally
    FCriticalSection.Leave;
  end;
end;

function TThreading.TWorker.RunTask(ATaskBody: TTaskBody): ITask;
begin
  Result := TTask.Create(ATaskBody);
  PutTask(Result);
end;

{ TThreading }

class function TThreading.GetWorker: TWorker;
begin
  if FWorkerIndex >= FPull.Count then
  begin
    Result := TWorker.Create;
    FPull.Add(Result);
  end else
  begin
  	Result := FPull.Items[FWorkerIndex];
  end;
  inc(FWorkerIndex);
  FWorkerIndex := FWorkerIndex mod FPoolSize;
end;

class constructor TThreading.Create;
begin
  FPull := TList<TWorker>.Create;
  FPoolSize := TThread.ProcessorCount * 2;
end;

class function TThreading.CreateWorker: TWorker;
begin
  Result := TWorker.Create;
end;

class destructor TThreading.Destroy;
var
  i: int32;
  worker: TWorker;
begin
	for i	:= 0 to FPull.Count - 1 do
  begin
  	worker := FPull.Items[i];
    worker.Terminate;
    worker.WaitFor;
    worker.Free;
  end;
  FPull.Free;
end;

class function TThreading.RunTask(ATaskBody: TTaskBody): ITask;
begin
  Result := TTask.Create(ATaskBody);
  RunTask(Result);
end;

class procedure TThreading.RunTask(const ATask: ITask);
var
	worker: TWorker;
begin
  worker := GetWorker;
  worker.PutTask(ATask);
end;

end.

